/**
* @file AbstractURIRequest.js
*
* Defines burst.io.AbstractURIRequest
*
* @author Copyright 2003 Mark D. Anderson (mda@discerning.com)
* @author Licensed under the Academic Free License 1.2 http://www.opensource.org/licenses/academic.php
*/

//=java package burst.io;

/**
* Abstract base class for fetching text content or an (XML) Document object given a URI.
*
* Generally, a programmer need not use this directly, because the <code>bu_Runtime</code> object has
* convenience functions such as <code>readEval</code> , <code>getTextAsync</code> , <code>getDocumentAsync</code> .
* If you want to get an instance of this abstract class that is supported by the current
* environment, us <code>bu_Runtime.getURIRequest</code> .
*
* To load and eval a script at a url from the same origin as this library, use instead
* <code>bu_Runtime.readEval</code> or <code>bu_require</code> .
*
* Several different implementation mechanisms are potentially available, according to the environment:
* - a global load() function
* - an object representing the local file system
* - an XMLHTTP object
* - an XMLDOM object
* - dynamically created script element
* - dynamically created iframe element
*
* Regardless of mechanism, it is generally limited to loading data or script from
* the same origin as the library itself.
*
* Some of the parameters that affect whether functionality is available include:
* - local file vs. url
* - text vs. binary resource
* - eval vs. just get contents
* - sync vs. async
*/
//=java public abstract class AbstractURIRequest {

function BU_AbstractURIRequest() {
}
burst.io.AbstractURIRequest = BU_AbstractURIRequest;

// class variable to hold cache.
BU_AbstractURIRequest.text_url_cache = {};

/**
Request the text content of the specified url, and call the handler when it arrives.

If HttpRequest is available in the environment, that is used.
Otherwise an iframe mechanism is used.

There is no return value. The handler is called with a first argument containing
the text content. The second argument is the implementation-specific request object.  

This function does not provide detailed control over such things as authentication,
progress callback, and so on. For that, you would need to explicitly create an HttpRequest
and use its methods directly.

@param url The url to load
@param handler The callback, which will be passed in the content text.
@param nocache Optional. If true, it disables a global by-url cache.

@todo make sure we never cache if the request failed.
*/
//=java public void getTextAsync(String url, Function handler, Boolean nocache);

BU_AbstractURIRequest.prototype.getTextAsync = function(url, handler, nocache) {
    if (arguments.length < 3) nocache = false;

  if (!nocache && bu_in(url, BU_AbstractURIRequest.text_url_cache)) {
      handler(BU_AbstractURIRequest.text_url_cache[url]);
  }

  this.getTextAsync_(url, function(text) {
			 if (!nocache) BU_AbstractURIRequest.text_url_cache[url] = text;
			 handler(text, this);
		     });
}

/** The method the subclass must implement (AbstractURIRequest takes care of caching). */
//=java protected abstract void getTextAsync_(String url, Function handler);
BU_AbstractURIRequest.prototype.getTextAsync_ = function() {bu_unimplemented('burst.io.AbstractURIRequest.getTextAsync_');}

/**
Get a Document object from the specified url asynchronously.

This is implemented with HttpRequest if available, otherwise via a hidden iframe.

There is no return value. The handler function is called with the Document object
as its firstu argument.

@todo when will responseXML be null? When is parseError set?

@param url the url to load.
@param handler a function called with the Document object
*/
//=java public abstract void getDocumentAsync(String url, Function handler);
BU_AbstractURIRequest.prototype.getDocumentAsync = function() {bu_unimplemented('burst.io.AbstractURIRequest.getDocumentAsync');}

//=java } // class AbstractURIRequest

bu_loaded('burst.io.AbstractURIRequest', null, ['burst.Lang','burst.io.IframeURIRequest', 'burst.xml.DomUtil','burst.io.XmlHttpURIRequest']);
