/**
* @file IframeURIRequest.js
*
* Defines burst.io.IframeURIRequest
*
* @author Copyright 2003 Mark D. Anderson (mda@discerning.com)
* @author Licensed under the Academic Free License 1.2 http://www.opensource.org/licenses/academic.php
*/

//=java package burst.io;

/**
Subclass of AbstractURIRequest based on a hidden iframe.
This is not needed in environments supporting XMLHttpRequest;
in that case use XmlHttpURIRequest .

<h3>Implementation Issues</h3>

This implementation relies on the primitive (and stateless) utility functions for iframes
in burst.xml.HtmlUtil .

There are several considerations:

<ul>
<li>"display: none" is mildly preferable over "visibility: hidden; width: 0; height: 0"), as 
"display: none" is definitively not going to affect layout.
But NS7 and Moz 1.0 will not submit form data or load an iframe src if it has "display: none":
  <ul>
  <li>Grant Wagner http://groups.google.com/groups?selm=3CDC2D02.66C9E67D%40agricoreunited.com
  <li>Martin Honnen http://groups.google.com/groups?selm=3D9C7879.4050708%40t-online.de
   </ul>
This is fixed in Moz 1.1.
In my experience, Opera 7 has the same bug, not yet fixed.

<li>IE5.0 PC does not support createElement("iframe"). Instead it must be done via document.write or innerHTML.

<li>Mozilla does not support modification of iframe content if the iframe is made by createElement: 
http://bugzilla.mozilla.org/show_bug.cgi?id=22681

<li>Load completion is difficult to detect.
In theory, an onload handler could be attached to:
  <ul>
  <li>The body of the parent document, above the iframe element.<br>
I do not believe this is reliable for iframe children created after the parent document has completed loading, 
and regardless it is messy to try to successfully wrap any existing body onload handlers.
  <li>The iframe element itself (in the parent document).<br>
In IE5.0 PC, the iframe element onload handler is not called:
   http://support.microsoft.com/default.aspx?scid=KB;en-us;q239638
According to Martin Honnen (http://groups.google.com/groups?selm=3C5BC731.9070803%40t-online.de )
onload will work on the iframe element in IE5.5 and NS6.2 (at least for 'http:' protocol, but perhaps not 'file:')

  <li>The body element inside the iframe's contents.<br>
Any example might be <code>onload='parent.done()'</code>. This would require cooperation from the 
response being returned (unless the Martin Honnen nested iframe hack is used, about which more below).
  </ul>

<li>In HTTPS, IE requires an initial src attribute or it will give a security warning:
http://groups.google.com/groups?selm=Xns91EB70C0BC81Fjocke4uhotmailcom%40130.133.1.4

<li>Opera 6 (perhaps others), does not allow later change of an iframe src if it isn't set originally:
Grant Wagner, http://groups.google.com/groups?selm=3D2DC0BD.8BFEFFCC%40agricoreunited.com

<li>It is important to use <code>location.replace()</code> rather than other means of setting/changing 
the iframe's source, to avoid putting the url on the browser's history (and hence affecting the back button
and refresh).

<li>A single iframe can't be used to achieve more than one request at a time. Solutions include
queuing through a single iframe serially, maintaining a pool of iframes, or creating
a new iframe for each request.

</ul>


<h3>Antecedents</h3>

The basic principle is straightforward and has been done many times before.
However, the tricky implementation issues above are often not all dealt
with, or are not dealt with well.

Furthermore, rampant confusion abounds which conflates iframe element objects, iframe window objects,
and iframe document objects.


Some implementations that bear examination are listed below.
I note here only some salient distinguishing characteristics, without detailing
any criticisms.
<ul>
<li>Brent Ashley, JSRS, http://www.ashleyit.com/rs/jsrs/test.htm <br>
  <ul>
  <li>Puts invisible iframe inside span inside body.
  <li>Relies on returned html to have a body onload which calls a callback function window.parent.jsrsLoaded
  <li>does pooling.
  </ul>

<li>Alex Russell, http://cvs.sourceforge.net/cgi-bin/viewcvs.cgi/netwindows/netWindows/winScripts/environ_core.js?rev=1.182&content-type=text/vnd.viewcvs-markup <br>
  <ul>
  <li>relies on returned content to use load_page_callback.js
  <li>deals with createElement("iframe") not working in IE5.0 PC
  <li>maintains a queue of loading requests to go through a single iframe channel.
  </ul>


<li>Martin Honnen, http://www.faqts.com/knowledge_base/view.phtml/aid/1272/fid/128/lang/en
               and http://www.faqts.com/knowledge_base/view.phtml/aid/6773/fid/128/lang/ 
               and http://groups.google.com/groups?selm=38F0C265.E877706A%40t-online.de
<br>
  <ul>
  <li>First creates an iframe inside the body. Then cleverly writes html into that iframe which itself
  has an iframe. The body element sandwich'd in between has an onload event to communicate up.
  </ul>

<li>Eric Costello, http://www.oreillynet.com/pub/a/javascript/2002/02/08/iframe.html
and (with corrections) http://developer.apple.com/internet/javascript/iframe.html#revisionnotes
and http://www.oreillynet.com/cs/user/view/cs_msg/8092 <br>
   <ul>
   <li>corrections deal with IE5 PC.
   <li>discusses the importance of location.replace()
   </ul>

<li>Steve Thames, http://www.faqts.com/knowledge_base/view.phtml/aid/7418 <br>
  <ul>
  <li>On NN6, adds load event handler to iframe element.
  <li>On IE5, polls with setTimeout until readyState is set.
  </ul>
</ul>


<h3>Our Approach</h3>

Our approach is:
<ul>
<li>In environments that support it, we attach an onload handler to the iframe element (IE5.5+ and NS6.2+).

<li>In environments that do not support iframe element onload, we use the Martin Honnen nested iframe trick.
(An alternative would be polling the readyState, but that is IE-specific and is not as well-performing.
Another alternative would be relying on the response to contain javascript to make a callback, but we
want to avoid that.)

<li>We maintain a pool of available hidden iframe children.
</ul>




<h3>iframe element, window, and document</h3>

This is a cross-platform morass. 

For getting an iframe window from the id or name attribute value of the iframe element: 
<ul>
<li>both Moz and IE can retrieve a frame by its name attribute: 
<pre>
   var iframe_win = window.frames[nameval]
</pre>
Note that IE has both document.frames and window.frames; Moz has only window.frames.

<li>IE but not Moz can retrieve a frame by its id attribute:
<pre>
   var iframe_win = window.frames[idval]
</pre>

<li>both Moz and IE can retrieve a frame by numerical index:
<pre>
   var iframe_win = window.frames[index]
</pre>
A person could just assume that the last frame is the desired one:
<pre>
   var iframe_win = window.frames[window.frames.length - 1]
</pre>

<li>If the iframe element is available, it is possible to retrieve the iframe window object via burst.xml.HtmlUtil.iframeContentWindow(iframe_el).

</ul>


<p>

To get the element given the id or name attribute value:
<ul>
<li>Moz and IE 5.5 PC (not IE 5.0 or IE Mac) can retrieve the iframe element via getElementById
<pre>
    var iframe_el = document.getElementById(idval);
</pre>
(IE4+ has document.all[idval] and document.all[nameval], with document.all[nameval]
being an Array if there are multiple ones.) 


<li>Moz and IE 5.? can retrieve by getElementsByName:
<pre>
    var iframe_el = document.getElementsByName(nameval)[0]
</pre>
(recall that IE considers both ID and NAME for getElementsByName,
and according to PPK http://www.xs4all.nl/~ppk/js/version5oud.html it doesn't
cover all elements. Moz and W3 cover all elements, and consider only name.)

<li>Moz and IE 5.? can retrieve by getElementsByTagName and numerical index:
<pre>
    var iframe_el = document.getElementsByTagName('IFRAME')[index]
</pre>
PPK says getElementsByTagName does not work on IE5 PC: http://www.xs4all.nl/~ppk/js/version5oud.html
</ul>


<p>
For going from iframe element to iframe window, see burst.xml.HtmlUtil.iframeContentWindow(iframe_el).

<p>
For going from iframe element to iframe document, see burst.xml.HtmlUtil.iframeContentDocument(iframe_el).

*/
//=java public class IframeURIRequest extends AbstractURIRequest {


/**
Construct a wrapper object around an iframe Node channel.

You generally want to call the class method getInstance() instead of
the constructor directly, so that you can take advantage of instance pooling.

If you happen to already have an
iframe element on the page that is intended to be used for communication,
that may be provided.

The actual iframe element is created only on demand; it may already have
one if this instance came from the pool.
 
Any constructed object is automatically added to the pool whenever
it is done being used.

All instances have a state variable which is one of:
- 'uninitialized' (after constructor call)
- 'loading' (just after entry to load())
- 'complete' (when load() callback first comes back)
- 'deleted' (when all load() callbacks have finished, and it is put back into the pool)

(This mimics microsoft readyState: http://msdn.microsoft.com/workshop/author/dhtml/reference/properties/readystate_1.asp
except for the addition of 'deleted').


@param iframe_el Optional. An iframe element for use. It must have a (unique) name attribute.
*/

//=java public IframeURIRequest(Node iframe_el) {}
function BU_IframeURIRequest(iframe_el) {
  if (iframe_el) {
    if (!iframe_el.nodeName) bu_throw("iframe argument not a node: " + iframe_el);
    if (iframe_el.nodeName.toUpperCase() !== 'IFRAME') bu_throw("iframe argument not an iframe but a " + iframe_el.nodeName);
    this.element_ = iframe_el;
    this.id_ = iframe_el.getAttribute('name') || bu_throw("no name attribute in provided iframe element");
  }
  else {
    // note, could do this as window.frames.length instead of a counter
    this.id_ = 'if' + BU_IframeURIRequest.id_counter_++;
    this.element_ = null;
  }

  BU_IframeURIRequest.register(this);
  this.state_ = 'uninitialized';
  this.loadCount_ = 0; // maintained throughout lifetime
  return this;
}
burst.io.IframeURIRequest = BU_IframeURIRequest;
bu_inherits(BU_IframeURIRequest, BU_AbstractURIRequest);

// class variables used for pooling and unique id generation.
BU_IframeURIRequest.id_counter_ = 0;
BU_IframeURIRequest.all_instances_ = {};
BU_IframeURIRequest.loading_instances_ = {};
BU_IframeURIRequest.pool_ = [];
BU_IframeURIRequest.DO_POOLING = true;

/**
Class method to get an instance. This is preferable to calling the constructor directly.
 */
//=java public static IframeURIRequest getInstance() {}
BU_IframeURIRequest.getInstance = function() {
    var instance = BU_IframeURIRequest.acquire();
    return instance ? instance : new BU_IframeURIRequest();
}

/**
* Registers the just-created BU_IframeURIRequest instance in the global registry.
*/
//=java protected static void register(BU_IframeURIRequest instance) {}
BU_IframeURIRequest.register = function(instance) {
  var iframe_id = instance.id_;
  if (bu_in(iframe_id, BU_IframeURIRequest.all_instances_)) bu_throw("attempt to re-register iframe instance with id '" + iframe_id + "'");
  BU_IframeURIRequest.all_instances_[iframe_id] = instance;
}

/**
* Reconstitute an instance from the pool. If none are available returns a false value.
* Sets state from 'deleted' to 'uninitialized'.
*/
//=java protected static BU_IframeURIRequest acquire() {}
BU_IframeURIRequest.acquire = function() {
  var instance = BU_IframeURIRequest.pool_.pop();
  if (instance) {
    instance.state_ = 'uninitialized';
    bu_debug("(IframeURIRequest.js) returning instance from pool");
    return instance;
  }
  bu_debug("(IframeURIRequest.js) no instance in pool");
  return instance;
}

/**
* Class method called when load callbacks are done. Sets state to 'deleted'.
*/
//=java protected static void release(BU_IframeURIRequest instance) {}
BU_IframeURIRequest.release = function(instance) {
  instance.state_ = 'deleted';
  if (typeof instance.innerDocument_ != 'undefined') delete instance[innerDocument_];
  
  delete BU_IframeURIRequest.loading_instances_[instance.id_];

  if (BU_IframeURIRequest.DO_POOLING) {
    BU_IframeURIRequest.pool_.push(instance);
  }
  else {
    // since we'll never re-use them, don't keep a pointer
    delete BU_IframeURIRequest.all_instances_[instance.id_];
    instance.destroy();
  }
}

/**
* remove the iframe Element from its parent Node, and eliminate any pointers we have.
* used if we are doing pooling.
*/
//=java protected void destroy() {}
BU_IframeURIRequest.prototype.destroy = function() {
  this.element_.parentNode.removeChild(this.element_);
  this.element_ = null;
}

/**
Class constant determining the src value for new iframe elements.
Some platforms require a src attribute to exist when creating an iframe element.

We currently use a value of '' for the default value.
Apparently Opera has a bug where use of 'about:blank' introduces cross-site security problems:
    Richard Cornford, http://groups.google.com/groups?selm=akgspb%24git%241%40mail.opera.no
*/
//=java public static final String NO_SRC = ''
BU_IframeURIRequest.NO_SRC = '';
//BU_IframeURIRequest.NO_SRC = 'about:blank';



/**
Implementation of mandatory base class method.
 */
//=java protected void getTextAsync_(String url, Function handler) {}
BU_IframeURIRequest.prototype.getTextAsync_ = function(url, handler) {
     this.load(url, function(iframeobj) {
        var text = iframeobj.getContents();
        handler(text, this);
     	});
}

/**
Implementation of mandatory base class method.
 */
//=java public void getDocumentAsync(String url, Function handler) {}
BU_IframeURIRequest.prototype.getDocumentAsync = function(url, handler, progress_handler) {
     this.load(url, function(iframeobj) {
        var doc = iframeobj.getDocument();
        handler(doc);
     });
}



// Return a Function suitable for use as an onload handler.
// This is attached to the iframe element, and calls BU_IframeURIRequest.callHandler.
function bu_make_onload_function(iframe_id) {
  return function() {
    // Moz does. IE and Opera do not.
    //alert("(IframeURIRequest.js) in iframe element onload, set as function");
    /*
    for(var i=0, w=window; w ; w=(w===w.parent?null:w.parent), i++) {
      alert('i=' + i + ' w.BU_IframeURIRequest=' + w.BU_IframeURIRequest + ' w.location=' + w.location);
      if (w.BU_IframeURIRequest) w.BU_IframeURIRequest.callHandler(iframe_id);
    } 
    */
    // 'BU_IframeURIRequest' instead of 'parent.BU_IframeURIRequest' should just work, since BU_IframeURIRequest is lexically bound.
    // But sometimes???? in Moz 1.3, it doesn't seem to be.
    // We do have to have a lexically bound id rather than 'this.id_' because this will be dynamic.
    if (typeof BU_IframeURIRequest == 'undefined') alert("(IframeURIRequest.js) huh? no BU_IframeURIRequest. parent.BU_IframeURIRequest=" + parent.BU_IframeURIRequest);
    else BU_IframeURIRequest.callHandler(iframe_id);
    return false;
  };
}

// Return a String whose text is suitable for use as an onload handler.
// if is_nested, then this is for use in the iframe loaded contents, not
// as the iframe element's onload.
// if inner_id, then we are doing the nested iframe hack, and should pass
// that into the handler.
function bu_make_onload_string(iframe_id, is_nested, inner_id) {
  var functest = is_nested ? 'parent && (typeof parent.BU_IframeURIRequest != "undefined")' : 'typeof BU_IframeURIRequest != "undefined"';
  var calljs = (is_nested ? 'parent.' : '') +
               "BU_IframeURIRequest.callHandler('" + iframe_id + "'" +
               (inner_id ? ", window, '" + inner_id + "'" : '') +
               ")";
  var onload_string;
  if (!BU_DBG_DOM) {
     onload_string = calljs;
  }
  else {
     onload_string = "alert('(IframeURIRequest.js) in onload handler for iframe " + 
          iframe_id + 
          ", from String onload, is_nested=" + 
          (is_nested ? true : false) + 
   "');" +
   "alert(" +
      ["'(IframeURIRequest.js) parent? '",
       "( (parent && parent !== window) ? true : false)", 
       "' BU_IframeURIRequest?'", 
       "(typeof BU_IframeURIRequest != 'undefined')",
       "' parent.BU_IframeURIRequest?'",
       "(typeof parent.BU_IframeURIRequest != 'undefined')",
       "' parent.parent.BU_IframeURIRequest?'",
       "(typeof parent.parent.BU_IframeURIRequest != 'undefined')"].join(' +\n ') +
    ");" +
    "if (" + functest + ") {" +
        calljs + ";" +
    "} else {" +
       "alert('(IframeURIRequest.js) test failed: " + functest + "');" +
       "for(var w=window; w ; w=(w===w.parent?null:w.parent)) {" +
          "alert(' w.BU_IframeURIRequest=' + (typeof w.BU_IframeURIRequest) + ' w.location=' + w.location + ' w.fallbackBU_IframeURIRequest=' + (typeof w.fallbackBU_IframeURIRequest));" +
       "}" +
//doesn't work in safari
//            "for(var i=0, w=window; w ; w=(w===w.parent?null:w.parent), i++) {" +
//               "alert('i=' + i + ' w.BU_IframeURIRequest=' + w.BU_IframeURIRequest + ' w.location=' + w.location + ' w.parent=' + (w===w.parent?null:w.parent));" +
//            "}" +
    "} return false;"
    }
    return onload_string;
}

// Return a String providing the entire html contents which will replace the document object of the iframe element.
// Note that:
//    IE6 calls both the body onload and the inner iframe onload.
//    Opera calls only the body onload.
function bu_nested_iframe_html(iframe_id, url) {
    var inner_id = iframe_id + 'inner';
    var onload_string = bu_make_onload_string(iframe_id, true, inner_id);
    if (typeof onload_string != 'string') bu_throw("(IframeURIRequest.js) bad onload_string: " + onload_string);
/*
At least in the case of a dynamically created iframe element (created via innerHTML
or createElement), with either an initial src value of '' or 'about:blank',
when that iframe's contents is later written, Opera does not correctly
resolve urls within that content to the iframe element's src value (or its
parent window if the src is relative).

One might make a case that this is a grey area, because the src value of
'' or 'about:blank' might be taken to be invalid or bizarre absolute urls.
However, what Opera actually does is interpret relative urls in the iframe
contents to be relative to the base url of the top window (not
the parent).

This problem can be dealt with either by resolving urls ourselves, as by
<code>url = burst.xml.HtmlUtil.resolveWindowUrl(url, window)</code>, or by inserting
an html 'base' element in the 'head' section.

See Richard Cornford http://groups.google.com/groups?selm=akgspb%24git%241%40mail.opera.no
*/
    var base_el = bu_UA.bug_iframe_relative_url() ? 
	burst.xml.DomUtil.el('BASE', {HREF: self.document.location.href}) : null;

    var html = burst.xml.DomUtil.el('HTML', null, 
                 burst.xml.DomUtil.el('HEAD', null, 
			burst.xml.DomUtil.el('TITLE'),
			base_el
		 ),
                 burst.xml.DomUtil.el('BODY', {
                    ONLOAD: onload_string
                   },
                   burst.xml.DomUtil.el('IFRAME', {
                     NAME: inner_id,
                     // IE6 calls. Opera does not.
                     ONLOAD: (BU_DBG_DOM ? "alert('(IframeURIRequest.js) inner iframe onload called, for inner iframe id " + inner_id + "');" : null),
                     SRC: url
                   }, null)));

    /*
    // could also do it wih a frameset instead of an iframe
    var html2 = '<HTML><HEAD><TITLE></TITLE></HEAD><FRAMESET ONLOAD="' +
       "for(var i=0, w=window; w ; w=(w===w.parent?null:w.parent), i++) {" +
          "alert('i=' + i + ' w.BU_IframeURIRequest=' + w.BU_IframeURIRequest + ' w.location=' + w.location);" +
          "if (w.BU_IframeURIRequest) w.BU_IframeURIRequest.callHandler('" + iframe_id + "');" +
       "} return false;" +
       '">' +     
       '<FRAME NAME="' + inner_id + '" SRC="' + url + '"></FRAMESET></HTML>';
    */
    return html;
}


/**
Set the iframe's visibility. Useful for debugging.
*/
//:CLIMETHOD void setVisible(Boolean isvis)
BU_IframeURIRequest.prototype.setVisible = function(isvis) {burst.xml.HtmlUtil.iframeSetVisible(this.element_, isvis)};

/** 
* class method to call the registered iframe's handler object. 
* Called from the nested iframe.
*
* The optional inner_id is for use when the iframe contents has a nested iframe
* within it, which is the real one we want.
*/
//:CLCMETHOD void callHandler(String iframe_id, Window caller_window, String inner_id)
BU_IframeURIRequest.callHandler = function(iframe_id, caller_window, inner_id) {
  if (!iframe_id) 
     bu_throw("(IframeURIRequest.js) got callback with " + arguments.length + " arguments and no frame id: " + iframe_id);

  bu_debug_Dom("(IframeURIRequest.js) got callback for frame id '" + iframe_id + "'" + 
        (inner_id ? ' (inner_id=' + inner_id + ')' : '') + 
        ''); 

  var inner_window;
  if (inner_id) {
    inner_window = caller_window.frames[inner_id] || bu_throw("no inner window for inner_id='" + inner_id + "'");
    //bu_debug_Dom("(IframeURIRequest.js) have inner window for inner_id='" + inner_id + "'");
  }

  var instance = BU_IframeURIRequest.loading_instances_[iframe_id];

  if (instance) bu_debug_Dom("(IframeURIRequest.js) loadcount=" + instance.loadCount_ + " state=" + instance.state_);

  if (!instance) {
    instance = BU_IframeURIRequest.all_instances_[iframe_id];
    if (!instance) bu_throw("(IframeURIRequest.js) got callback for unknown frame id '" + iframe_id + "'");

    // sigh, maybe have to ignore an obsolete element onload if we already got a nested iframe onload.	
    // see comment within the load() function.
    var can_dyn = bu_UA.can_iframe_onload_dyn();
    var can_static = bu_UA.can_iframe_onload_static();
    if (!inner_id && !can_dyn && can_static && instance.loadCount_ > 1) {
      bu_debug_Dom('(IframeURIRequest.js) ignoring redundant onload event');
      return;
    }
    bu_throw("(IframeURIRequest.js) got callback for frame id '" + iframe_id + "'" + 
        (inner_id ? ' (inner_id=' + inner_id + ')' : '') +
        " that exists but is in state '" + instance.state_ + "', loadCount=" + instance.loadCount_);
  }

  // maybe the iframe element was created with an onload, and onload got fired before
  // element_ could be set....
  if (!instance.element_) {
    instance.element_ = caller_window ? caller_window.parent.getElementById(iframe_id) : window.getElementByid(iframe_id);
    bu_alert("(IframeURIRequest.js) no element_ in instance for id " + iframe_id + "; " + (instance.element_ ? "was" : "was not") + " able to set");
    if (!instance.element_) 
        bu_throw("(IframeURIRequest.js) BU_IframeURIRequest instance for id " + iframe_id +
            " has no element_ and was not able to find it " + 
            (caller_window ? ' (has caller_window)' : ' (no caller_window)'));
  }

  instance.state_ = 'complete';
  if (inner_id) {
     // merely refering to window.document may get a "access denied" error in IE5.0 if it thinks it violates same origin.
     instance.contentDocument_ = inner_window.document || bu_throw("no document object in inner_window");
  }
  else {
     instance.contentDocument_ = burst.xml.HtmlUtil.iframeContentDocument(instance.element_);
  }

  // TEMPORARY - call handler without try-catch
  //instance.handler_(instance);
  //BU_IframeURIRequest.release(instance);
  try {
    instance.handler_(instance);
  }
  catch(e) {
    bu_alert("(IframeURIRequest.js) handler threw an exception: " + e + "\nrethrowing...");
    // rethrow
    throw e;
  }
  finally {
    BU_IframeURIRequest.release(instance);
  }
}

/*
Methods to get any of:
- the document (and use DOM methods)
- the innerText
- the outerHTML of the whole documentElement
- the innerHTML of the document body
*/

/**
* Get the Document object for the fetched url.
*
* Only legal after load().
*
* This is suitable whether the Content-Type is 'text/plain', 'text/html', or 'text/xml'.
*/
BU_IframeURIRequest.prototype.getDocument = function() {
  bu_assertState(this.state_, 'complete');
  return (this.contentDocument_ || bu_throw('no contentDocument_'));
}

/*
text/plain
text/html
text/xml               favored over application/xml when readable/viewable (text/plain fallback)
text/xhtml             deprecated in favor of application/xhtml+xml
application/xml
application/xhtml      deprecated in favor of application/xhtml+xml
application/xhtml+xml

RFC3023.has:
  text/xml
  application/xml
  suggests application/xhtml+xml

According to http://www.opera.com/docs/specs/#xml Opera will treat 'text/xml'
and 'application/xml' content as XML.

what about these:
text/javascript     unregistered, but IE doesn't use type="application/javascript" without language
application/javascript
application/x-javascript
application/ecmascript
text/ecmascript
text/css
*/

/**
* Get the contents as String, making a best effort at guessing how to do that.
*/
BU_IframeURIRequest.prototype.getContents = function() {
  switch(burst.URI.guessMimeType(this.url_)) {
  case 'text/plain': return this.getText();
  case 'text/html': return this.getHTML();
  case 'text/xml': return this.getXML();
  default: return this.getHTML();
  }
}

/**
* Get the contents as a String of XML.
*
* Only legal after load().
*
* This uses burst.xml.XmlDoc.toXmlString(getDocument()), so this function should
* generally be used only when you know the iframe has Content-Type of 'text/xml'
* (or 'text/xhtml' or 'application/xml' or 'application/xhtml+xml'?).
* Note that burst.xml.XmlDoc.toXmlString may not be supported in some browsers
* even then. 
*/
//:CLIMETHOD String getXML()
BU_IframeURIRequest.prototype.getXML = function() {
  var doc = this.getDocument();
  return burst.xml.XmlDoc.toXmlString(doc);
}

/** 
* Get the contents as text. 
*
* Only legal after load().
*
* This uses innerText (native or emulation), so this function should
* generally be used only when you know the iframe has Content-Type of 'text/plain'.
*/

//:CLIMETHOD String getText()
BU_IframeURIRequest.prototype.getText = function() {
  var doc = this.getDocument();

  // bu_dbg_doc(doc);

  return burst.xml.DomUtil.getInnerText(doc);
}

/** 
* Get the html contents as String.
*
* Only legal after load().
*
* This uses outerHTML (native or emulation), so this function should
* generally be used only when you know the iframe has Content-Type of 'text/html'.
*/
//:CLIMETHOD String getHTML()
BU_IframeURIRequest.prototype.getHTML = function() {
  var doc = this.getDocument();
         
  // bu_dbg_doc(doc);

  return burst.xml.DomUtil.getOuterHTML(doc.documentElement);
}

var BU_ONLOAD_AS_FUNCTION = false;

/**
Use the BU_IframeURIRequest instance to load the contents from the specified url.

Legal only in the state 'uninitialized'.

Upon load completion, the handler is called with a single argument which is the BU_IframeURIRequest instance.
From that, it is possible to call one of the BU_IframeURIRequest.get* methods

@todo For curiosity, determine when this can be implemented by setting this.element_.src (even though replace is preferable).

@todo error handling

@todo consider exposing readystatechange

@todo Need to set base element for opera? Richard Cornford, http://groups.google.com/groups?selm=akgspb%24git%241%40mail.opera.no

* @param url
* @param handler The handler function to be called when the contents is loaded. 

*/
//:CLIMETHOD void load(String url, Function handler)
BU_IframeURIRequest.prototype.load = function(url, handler) {
  bu_assertState(this.state_, 'uninitialized');

  // update state
  this.state_ = 'loading';
  this.loadCount_++;
  BU_IframeURIRequest.loading_instances_[this.id_] = this;
  //loadPage(url, 1); return;
  this.handler_ = handler;
  this.url_ = url;
  if (false) { // if changing url
     // ...change url...
     this.actual_url_ = url;
  }

  bu_debug("(IframeURIRequest.js) url='", url, "' window.location.href=", window.location.href);
  bu_debug("(IframeURIRequest.js) base=", burst.xml.HtmlUtil.getWindowBaseUrl(window));

  var iframe_el = this.element_;
  var iframe_id = this.id_; // serves for lexical binding of id within function defiition

  // if it supports an onload attribute for iframe elements, create and load in one step
  var can_dyn = bu_UA.can_iframe_onload_dyn();
  var can_static = bu_UA.can_iframe_onload_static();

  // we have an iframe element already (perhaps because we got it from the pool).
  // just set its location (or rewrite its innerHTML if setting location won't work).
  if (iframe_el) {
     // set the onload attribute.
     // works in Moz, but not IE6 (onload is never called).
     if (can_dyn) {
        bu_debug("(IframeURIRequest.js) element exists, can_iframe_onload_dyn()=true. Setting onload and replacing location, for id " + iframe_id + " and url " + url);

        // Changing this dynamically only seems to work in Moz.
        iframe_el.onload = bu_make_onload_function(iframe_id);

        var loc = burst.xml.HtmlUtil.iframeLocation(iframe_el, iframe_id);
        // todo: PPK claims IE5 can't set location.href http://www.xs4all.nl/~ppk/js/iframe.html 
        loc.replace(url);
     }

     // Use the Martin Honnen nested iframe hack
     else {
        bu_debug("(IframeURIRequest.js) element exists, can_iframe_onload_dyn()=false, so writing inner content for id " + iframe_id + " and url " + url);

        var html = bu_nested_iframe_html(iframe_id, url);

	// NOTE: in the case of (!can_dyn && can_static) we already created an iframe element with an onload
	// handler, and it will get invoked again whenever we replace the inner contents, so we shouldn't
	// have an additional onload handler do things on the inner iframe onload.
	// One or the other has to go, starting with the second use of the iframe (loadCount == 2).
	// Oddly in IE6, the current value of iframe_el is actually the window's, not the one we set,
	// and setting it now has no effect.
	// Yet that is the one we'd like to get rid of, since it may have state that is now obsolete.
	// Since we can't seem to get rid of it, we have an ugly hack to ignore its callback later on.
	/*
	if (can_static) {
	   bu_alert('(IframeURIRequest.js) need to remove iframe_el.onload=' + iframe_el.onload);
	   iframe_el.onload = function() {};
        }
	*/
        burst.xml.HtmlUtil.iframeReplaceContents(iframe_el, html);
 
        // hack to ensure there is something on the other side to make its way back...
        if (true) {
          var iframe_win = burst.xml.HtmlUtil.iframeContentWindow(iframe_el);
          iframe_win.fallbackBU_IframeURIRequest = BU_IframeURIRequest;
        }
     }
  }

  // else we do not already have an iframe element.
  else {
     if (can_dyn || can_static) { 
        bu_debug_Dom("(IframeURIRequest.js) no element, can_dyn=" + can_dyn + " can_static=" + can_static + ", so creating and loading together, for id " + iframe_id + " and url " + url);
	// note that if innerHTML is used to create the iframe and set the onload at once, 
	// it might happen that the onload will be fired prior to us being able to set this.element_
        var onloadval = (can_dyn || BU_ONLOAD_AS_FUNCTION) ? bu_make_onload_function(iframe_id) : bu_make_onload_string(iframe_id);
        var this_ = this;
        iframe_el = burst.xml.HtmlUtil.iframeCreate(iframe_id, url, onloadval, function(el) {this_.element_ = el});
     }

     // otherwise we have to create the iframe element with no src, then overwrite the
     // contents with nested iframe. 
     // this does work in Safari 60, but it seems not to have any way to communicate back to the parent.
     else {
        bu_debug_Dom("(IframeURIRequest.js) no element, can_iframe_onload_static()=false, so creating and loading separately for id " + iframe_id + " and url " + url);
        iframe_el = burst.xml.HtmlUtil.iframeCreate(iframe_id);
        // set element_ so available in callback and later use.
        this.element_ = iframe_el;

	var created_handler = function() {
          var newhtml = bu_nested_iframe_html(iframe_id, url);
          burst.xml.HtmlUtil.iframeReplaceContents(iframe_el, newhtml);

          // hack to ensure there is something on the other side to make its way back...
          if (true) {
            var iframe_win2 = burst.xml.HtmlUtil.iframeContentWindow(iframe_el);
            iframe_win2.fallbackBU_IframeURIRequest = BU_IframeURIRequest;
          }
        }
	if (bu_UA.bug_iframe_delayed_create()) {window.setTimeout(created_handler, 10)}
	else created_handler();
     }
  }
}

//=java } class IframeURIRequest

bu_loaded('burst.io.IframeURIRequest', ['burst.xml.HtmlUtil', 'burst.xml.DomUtil', 'burst.URI']);

