/**
* @file XmlHttpURIRequest.js
*
* Defines burst.io.XmlHttpURIRequest , an implementation of burst.io.AbstractURIRequest using XMLHTTP object.
*
* @author Copyright 2003 Mark D. Anderson (mda@discerning.com)
* @author Licensed under the Academic Free License 1.2 http://www.opensource.org/licenses/academic.php
*/

//=java package burst.io;

/**
Subclass of AbstractURIRequest using XMLHTTP object support.

This is not supported in all environments: only if <code>bu_Runtime.has_HttpRequest()</code> .

Microsoft calls their implementation object "XMLHTTP", and Mozilla calls theirs "XMLHttpRequest".
Note that despite the name, they are able to to retrieve text/plain content as well as text/xml.
Furthermore, they aren't limited to HTTP either, but also "file:" and "ftp:", depending on
security restrictions.

The browser implementation of <var>XMLHTTP</var> is connected to its implementation of the Xml DOM,
because when retrieving xml, a <var>HttpRequest</var> has a <var>responseXML</var> member which is
an instance of <var>DomDocument</var>.

On IE, the XMLHTTP object (interface IXMLHTTPRequest) is used:
- http://msdn.microsoft.com/library/en-us/xmlsdk30/htm/xmobjpmexmlhttprequest.asp
- http://msdn.microsoft.com/library/en-us/xmlsdk30/htm/xmobjxmlhttprequest.asp

Note that Microsoft's IServerXMLHTTPRequest has additional methods such as waitForResponse:
- http://msdn.microsoft.com/library/en-us/xmlsdk/htm/xml_obj_iserverxmlhttprequest_7zlc.asp

On Mozilla, this is implemented with <code>new XMLHttpRequest()</code> .
See:
- http://www.xulplanet.com/references/elemref/ref_XMLHttpRequest.html
- http://www.mozilla.org/xmlextras/

Mozilla does a pretty good job of mimicking IE.
So after the object is created and returned, essentially the same object methods can be used.
We do provide some additional convenience methods in our wrapper.

When retrieving text/plain, the <var>responseText</var> variable holds the text content
(transformed to a unicode string -- it can't be used to retrieve binary).

When retrieving text/xml, the <var>responseXML</var> variable is a DomDocument instance.
In that case, the usual DOM methods apply to it, or its entire XML serialized string
can be obtained via
<code>burst.xml.XmlDom.toXmlString</code> .

The <var>responseBody</var> variable is an array of bytes of exactly the response body,
which is part of the language-independent interfaces for both IE and Mozilla.
However, it is not exposed to javascript. IE has typeof 'unknown' for httpreq.responseBody,
and Mozilla has typeof 'undefined'.

For an overview of the XMLHTTP object, see:
- Jim Ley, http://jibbering.com/2002/4/httprequest.html

There are still some research areas:
- how does XMLHTTP interact with the browser's cache?
- what happens if you try to intermix sync and async request objects to the same server?
- what happens with regard to persistent http connections?

Note that RFC 2616 recommends a limit of two persistent connections from a client to any
one server:
<pre>
  Clients that use persistent connections SHOULD limit the
  number of simultaneous connections that they maintain to
  a given server. 
  A single-user client SHOULD NOT maintain more than 2 connections
  with any server or proxy.
</pre>
See http://www.w3.org/Protocols/rfc2616/rfc2616.html

Browsers have different conventions (not mandated by any RFC) about how many HTTP 1.0 connections
they may hold to a server, and how many overall threads/connections (combined from all servers) they may
maintain. For IE, see:
http://support.microsoft.com/default.aspx?scid=kb;en-us;183110


Some caveats about IE's XMLHTTP implementation:
<ul>
<li>See http://www.perfectxml.com/msxmlkbs.asp?FilterCat=4 for a summary of MSXML bugs that Microsoft lists.
In particular:
  <ul>
  <li>Up to MSXML 3.0, DomDocument.load will ignore Expires headers and send GET with If-Modified-Since anyway: http://support.microsoft.com/default.aspx?scid=kb;EN-US;q293099
  <li>XMLHTTP won't send cookies: http://support.microsoft.com/default.aspx?scid=kb;EN-US;q290899
and http://support.microsoft.com/default.aspx?scid=kb;EN-US;q234486
  </ul>

<li>If the response's HTTP Content-Type header field is not "text/xml", then 'responseXML' will be empty regardless
of the content.
<li>If the response's HTTP Content-Type header field is not "text/plain", then 'responseText' will be empty regardless of the content.
<li>When 'readyState' is 3 (interactive) the object is still in an undefined state. For example, it is not
possible to display load progress by looking at responseText.length (this works in Mozilla).
I have seen this and Scott Andrew reports it also at http://www.scottandrew.com/weblog/2002_12
</ul>

Some caveats about Mozilla's implementation:
<ul>
<li>
Jeffrey M. Yates claims that Mozilla will not actually carry out multiple requests at once, even in async mode:
 http://www.pbwizard.com/Articles/DownloadDoc.htm#bugs
However, this may just have to do with attempting multiple requests using <em>the same xmlhttp object</em>.
I have not yet verified with Mozilla has such a limitation with multiple
xmlhttp objects (except for any queueing behavior due to a connection limit).

<li>Search Mozilla bugs for 'XMLHttpRequest': http://bugzilla.mozilla.org/buglist.cgi
</ul>

@todo Whether both Mozilla and IE support the same transfer-encoding's as the browser.
@todo Support the xml and http mechanisms available through Flash (XML.sendAndLoad(), etc.)
@todo Consider Konqueror W3 "Load and Save" support.
@todo See if it works in Safari
*/

//=java public class XmlHttpURIRequest extends AbstractURIRequest {

function BU_XmlHttpURIRequest() {
}
burst.io.XmlHttpURIRequest = BU_XmlHttpURIRequest;
bu_inherits(BU_XmlHttpURIRequest, BU_AbstractURIRequest);

BU_XmlHttpURIRequest.prototype.getTextAsync_ = function(url, handler) {
    BU_XmlHttpURIRequest.getAsync(url, function(httpreq, url) {
		      var text = httpreq.responseText;
		      handler(text);
		  });

}

BU_XmlHttpURIRequest.prototype.getDocumentAsync = function(url, handler, progress_handler) {
     BU_XmlHttpURIRequest.getAsync(url, function(httpreq, url) {
       var xml = httpreq.responseXML;
       //bu_alert("(XmlHttpURIRequest.js) getDocumentAsync calling callback with xml object: " + xml + " typeof=" + (typeof xml));
       if (xml) {
         var parse_error = burst.xml.XmlDoc.getParseError(xml);
	 if (parse_error && parse_error != 0) bu_alert("(XmlHttpURIRequest.js) parseError: " + parse_error);
       }
       else
	 bu_alert("(XmlHttpURIRequest.js) no xml object");
       handler(xml, httpreq);
     }, progress_handler);
}

/** ProgID values to try for Internet Explorer XMLHTTP. See burst.xml.XmlDoc.XMLDOM_PROGID for general information about Microsoft ProgIDs. */
//=java public static final Array XMLHTTP_PROGID = ['Msxml2.XMLHTTP.4.0', 'Msxml2.XMLHTTP', 'Microsoft.XMLHTTP']
BU_XmlHttpURIRequest.XMLHTTP_PROGID = ['Msxml2.XMLHTTP.4.0', 'Msxml2.XMLHTTP', 'Microsoft.XMLHTTP'];

// IE and Moz just return undefined from XMLHTTP method calls, so this utility to check results isn't used.
function bu_check_result(result, what) {}
/*
// var BU_S_OK = 0; // microsoft return value
function bu_check_result_unused(result, what) {
  if (result == BU_S_OK) return;
  for(var i=2;i<arguments.length;i++) {what += arguments[i]}
  bu_alert("The call " + what + " returned " + result + " instead of " + BU_S_OK);
}
*/

/**
Class method to create and return a native XMLHttpRequest object.

While creation differs between IE and Mozilla, the methods on the resulting object are
very similar.

You'll want to call those methods to get control over:
- authentication
- POST or GET or other request method
- setting request headers
- progress callbacks
- request abort
- and so on

If you don't need that level of control, you can use the convenience methods in this wrapper.
*/
//=java public static XMLHttpRequest create() {}
BU_XmlHttpURIRequest.create = function() {
  var onerror_ok = true;
  var httpreq;
  if (bu_Runtime.has_moz_XMLHttpRequest()) {
    httpreq = new XMLHttpRequest();
    // httpreq = bu_fix_readystate(httpreq);
  }
  else if (bu_Runtime.has_msxml()) {
    onerror_ok = false;
    httpreq = burst.Lang.createActiveXObject(BU_XmlHttpURIRequest, 'XMLHTTP_PROGID', 'XMLHttpRequest');
  }
  else {
    bu_unsupported("(XmlHttpURIRequest.js) no XMLHttpRequest implementation available");
  }

  // Not part of IE's IXMLHTTPRequest interface, and if you try to set it,
  // IE6 throws TypeError "Object doesn't support this property or method"
  if (onerror_ok)
     httpreq.onerror = function() {
	bu_alert("(XmlHttpURIRequest.js) onerror callback fired from httprequest " + httpreq); 
	for(var i=0;i<arguments.length;++i) bu_alert("argument[" + i + "]='" + arguments[i] + "'");
     }

  return httpreq;
}

/**
A convenience utility to create an HttpRequest object and send a request for a url,
in one step.

This is async. The function returns the created HttpRequest object, but nothing much
can be done with it til the handler is called.

The handler is called with arguments (httpreq, url).
From that, the handler can retrieve responseText or responseXML, as it pleases.

@todo What is the error handling for Mozilla and IE?

@param url
@param handler Called when done.
@param progress_handler Optional. Called every time onreadystatechange is called.
@throws BurstError if any XMLHttpRequest method fails.
@return The HttpRequest object.
*/
//=java public static XMLHttpRequest getAsync(String url, Function handler, Function progress_handler)
BU_XmlHttpURIRequest.getAsync = function(url, handler, progress_handler) {
  var httpreq = BU_XmlHttpURIRequest.create();

  var already_called = false;
  httpreq.onreadystatechange = function() {
    if (progress_handler) progress_handler(httpreq, url);
    //bu_alert("(XmlHttpURIRequest.js) in onreadystatechange callback for " + url + ", readyState=" + httpreq.readyState);
    // IE5.0 with progid "Microsoft.XMLHTTP" calls us 4 times, every time with readyState == 4
    if (httpreq.readyState == 4) {
      if (!already_called) {
        already_called = true;
        //bu_alert("(XmlHttpURIRequest.js) httpRequestGetAsync invoking callback");
        bu_dbg_httpreq(url, httpreq);
        handler(httpreq, url);
      }
      //else {bu_alert("(XmlHttpURIRequest.js) redundant onreadystatechange");}
    }
  };

  bu_httpRequest_open_send(httpreq, url, null, true);

  return httpreq;
}

/**
A convenience utility to create an HttpRequest object and fetch the
text content of the specified url, in one step.

You generally do not want to do this:
in a web environment you don't want to use synchronous requests from remote urls.

@param url The url to GET.
@throws BurstError if any XMLHttpRequest method fails.
*/
//=java public static String getTextSync(String url)
BU_XmlHttpURIRequest.getTextSync = function(url) {
  var httpreq = BU_XmlHttpURIRequest.create();

  bu_httpRequest_open_send(httpreq, url, null, false);

  // bu_dbg_httpreq(url, httpreq);
  // var ct = httpreq.getResponseHeader('Content-Type');

  var text = httpreq.responseText;
  return text;
}

// internal function to actually do the sending
function bu_httpRequest_open_send(httpreq, url, requestBody, async) {
  // the actual url we send in
  var open_url = url;

  // mozilla has a bug (at least up through 1.3 20030312) where it uses top.location.href
  // to relativize non-absolute urls for xmlhttprequest
  // See http://bugzilla.mozilla.org/show_bug.cgi?id=76744 but it was resolved long ago
  // http://bugzilla.mozilla.org/show_bug.cgi?id=141002 recent bug for openWindow being relative to top instead of frameset
  if (bu_UA.family_ === BU_UA_FAMILY_GECKO && 
        (typeof window != 'undefined') && window !== top && 
	!burst.URI.isAbsolute(url)) {
    bu_debug("(XmlHttpURIRequest.js) compensating for HttpRequest relative url bug, for url=" + url);
    open_url = burst.xml.HtmlUtil.resolveWindowUrl(url, window);
 alert("normalized '" + url + "' to '" + open_url + "' for mozilla");
  }
  else {
    bu_debug("(XmlHttpURIRequest.js) not expecting HttpRequest relative url bug, for url=" + url);
  }

  /*
   * open(method, url, async, user, password).
   * default for async is true.
   */
  var result = httpreq.open('GET', open_url, async);
  bu_check_result(result, 'HttpRequest.open(', open_url, ",", async, ')');

  /*
   * send(requestBody)
   *
   * IE6 is ok with no args. Moz (tested 1.3) throws an exception if you pass no args.
   *
   * If not null, IE6 will take a String or an XML Dom 
   * (or a byte array or stream, if have native objects):
   *    var xmldoc = new ActiveXObject("Microsoft.XMLDOM");
   *    xmldoc.loadXML(xmlstr); 
   *    xmlstr2 = xmldoc.xml
   *
   * Moz takes only a Document until 0.9.7 http://www.mozilla.org/xmlextras/
   * Thereafter it also takes a String.
   *    var parser = new DOMParser();
   *    var xmldoc = parser.parseFromString(xmlstr,"text/xml");
   */
  result = httpreq.send(requestBody);
  bu_check_result(result, "HttpRequest.send(", requestBody, ")");

  // assert status if synchronous
  if (!async && httpreq.status != 200) bu_throw("Request for url '" + open_url + "' resulted in http status " + httpreq.status + " (" + httpreq.statusText + ")");
}

// internal function that typically does nothing but may issue debug output
function bu_dbg_httpreq(url, httpreq) {
/*
    // 200 or 404, a number
    bu_alert("(XmlHttpURIRequest.js) send completed. httpreq.status=" + httpreq.status + ' typeof=' + (typeof httpreq.status));
    // 'OK' or 'Not Found'
    bu_alert("(XmlHttpURIRequest.js) send completed. httpreq.statusText=" + httpreq.statusText);

    bu_alert("(XmlHttpURIRequest.js) httpreq.responseText=" + httpreq.responseText);
*/

    // seen: 'text/plain'
    // what about <meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">
    // bu_alert("(XmlHttpURIRequest.js) for url " + url + " got httpreq.getResponseHeader('Content-Type')=" + httpreq.getResponseHeader('Content-Type'));
}









//=java } class XmlHttpURIRequest

bu_loaded('burst.io.XmlHttpURIRequest', null, ['burst.xml.HtmlUtil', 'burst.URI', 'burst.xml.XmlDoc']);