/**
* @file DragDrop.js
*
* Defines burst.web.DragDrop which holds static functions related to drag and drop.
*
* @author Copyright 2003 Mark D. Anderson (mda@discerning.com)
* @author Licensed under the Academic Free License 1.2 http://www.opensource.org/licenses/academic.php
*/

//=java package burst.web;

/**
Scoping class to hold static functions related to drag and drop.
*/
//:NSBEGIN DragDrop

burst.web.DragDrop = {};
var BU_DragDrop = burst.web.DragDrop;

/*
Drag (and drop)

Implementation note:

  if can determine initial value of style.left and style.top when dragging is started,
  then as dragged, need to just add the distance travelled between the two events,
  in mouse units (diff of ev.clientX).

  if can't determine initial style.left/style.top, have to calculate it.
  one way is to use the initial element-relative click location:

   // element-relative offset when first clicked
   // on IE, also have to add an offset to make mouse clicks and element positions work.
   // In IE, offsetX/offsetY is in the element offsetLeft/offsetTop coordinate system:
   //    http://msdn.microsoft.com/workshop/author/dhtml/reference/properties/offsetx.asp
   // For an absolutely positioned object, still the offsetLeft value still needs an adjustment.
   ieborderX = isIE ? burst.xml.HtmlUtil.getDocumentElementIE().clientLeft : 0; // 2px adjust
   ieborderY = isIE ? burst.xml.HtmlUtil.getDocumentElementIE().clientTop : 0; // 2px adjust
   origOffsetX = (isMoz ? event.layerX : event.offsetX) + ieborderX;
   origOffsetY = (isMoz ? event.layerY : event.offsetY) + ieborderY;

   // new value for left = doc relative click location - original offsetX
   n.style.left = event.clientX + (isIE ? document.documentElement.scrollLeft : window.pageXOffset) -
                  origOffsetX
   n.style.top = event.clientY +  (isIE ? document.documentElement.scrollTop : window.pageYOffset) -
                  origOffsetY




IE has specific support for drag and drop
  http://msdn.microsoft.com/workshop/author/dhtml/reference/properties/dropeffect.asp
  http://msdn.microsoft.com/workshop/author/datatransfer/overview.asp
  http://msdn.microsoft.com/workshop/author/dhtml/reference/events/ondragstart.asp
events include:
  ondrag      Fires on the source object continuously during a drag operation. 
  ondragend   Fires on the source object when the user releases the mouse at the close of a drag operation. 
  ondragstart Fires on the source object when the user starts to drag a text selection or selected object. 

  ondragenter Fires on the target element when the user drags the object to a valid drop target. 
  ondragleave Fires on the target object when the user moves the mouse out of a valid drop target during a drag operation. 
  ondragover  Fires on the target element continuously while the user drags the object over a valid drop target. 
  ondrop      Fires on the target object when the mouse button is released during a drag-and-drop operation. 

On a standard browser, need to do drag and drop via mousedown, mousemove and mouseup

IE has Node.setCapture and document.releaseCapture. When a Node has mouse capture, it receives all events, though srcElement
is still set to the Node that would have gotten it.
Starting in IE 5.5, setCapture takes a boolean bContainerCapture as to whether to capture ancestor Node events too.
There is an 'onlosecapture' event sent when the mouse escapes.
  http://msdn.microsoft.com/workshop/author/dhtml/overview/mousecapture.asp
  http://msdn.microsoft.com/workshop/author/dhtml/reference/methods/setcapture.asp
  http://msdn.microsoft.com/workshop/author/dhtml/reference/events/onlosecapture.asp

captureEvents was a NN4 thing.

http://www.walterzorn.com/dragdrop/dragdrop_e.htm
http://brainjar.com/dhtml/drag/ 
http://www.brainjar.com/dhtml/windows/default8.asp
http://webfx.eae.net/dhtml/domgenmove/domgenmove.html 
http://www.bangeye.com/mozilla/dragme.htm 
http://forums.devshed.com/archive/1/2001/07/1/17629
http://www.xml.com/pub/a/2002/02/27/drag.html

If not an absolute positioned element, can still move it around?
What browsers will just drag without our interference?
For drop, need to make cursor change when over something that will accept drop.

Node could be from a getElementById or by just passing in "this" from an event handler.
Be careful not to pass in a text node.
Should be a node that is absolutely positioned.
Typically this is in a 'mousedown' event handler.

TODO: search through parents for one that is movable?
TODO: make a new wire frame gizmo to drag around. which will have to be initialized to dimensions of initial node.
TODO: set node.css.cursor = 'auto' or 'pointer' or 'se-resize' http://www.w3schools.com/css/pr_class_cursor.asp
TODO: set of multiple nodes to drag at once (shift click?)
TODO: need to do anything with alt or title attributes while dragging?
TODO: some other function to allow resizing with some fixed point.
TODO: use dynamic expressions on IE

params:

  srcNode - node that should be offered to a drop destination node. might be ancestor of one with mousedown event.
  mouseNode - default srcNode. node to move around with the mouse. 
    Instead of srcNode, might be wire frame the same size as srcNode, or a clone, or just an img.
    Can be a Function instead of a Node.
  mouseToggle - default false. boolean. whether it is necessary to unhide and hide mouseNode at start end end.
  finalNode - default srcNode. node to leave after a successful move. Can be a Function instead of a Node.
  dragType - no default. a string which a drop destination node can check for
  clipNode - no default. a node that mouseNode should be kept inside of (visibly, regardless of parenting)
  clipRec - no default. an array [top, right, bottom, left] of strings (with units). '0' or 'auto' for no clipping? all relative to srcNode, or absolute document-relative?
  cursor - no default. cursor to use while dragging 
  zIndex - no default. zIndex to set on node while moving. you might get max via burst.xml.HtmlUtil.maxZIndex(node.ownerDocument) + 1.
  dropOnly - default false. boolean, whether must find a drop site (won't just do a move -- will snap back to original place) 
  leaveDuring - default false. boolean, whether source node is visible in original location while moving mouseNode (until mouseup)
  leaveAfter - default false. boolean, whether source node is visible after successful drop or move.
  onlyX - default false. boolean, only allow horizontal movement
  onlyY - default false. boolean, only allow vertical movement
  snapX - no default. number of pixels to round to for horizontal movement
  snapY - no default. number of pixels to round to for vertical movement
  autoscrollX - default false. boolean. whether to window.scrollTo (or ancestor div that scrolls?) as drag approaches horiz edge.
  autoscrollY - default false. boolean. whether to window.scrollTo (or ancestor div that scrolls?) as drag approaches vert edge.
  autoscrollSpace - default 27. if autoscroll=true, number of pixels at which to start moving. 
    or maybe just scroll to always keep drag item in place? or make scrollTo delta equal to screenX/Y offset?

 */
BU_DragDrop.dragStart = function(node, ev, dragZIndex) {
    // save document-relative click location 
    var startPageX = BU_WindowEvent.getMousePos(ev, 'pageX');
    var startPageY = BU_WindowEvent.getMousePos(ev, 'pageY');

    // save document-relative node location
    var startLeft = burst.xml.HtmlBox.getBorderBoxLeft(node);
    var startTop = burst.xml.HtmlBox.getBorderBoxTop(node);

    // save original z-index
    var startZ = node.style.zIndex;

    // put node on top while dragging
    // note that positioned elements always go above non-positioned elements?
    if (!dragZIndex) dragZIndex = burst.xml.HtmlUtil.maxZIndex(node.ownerDocument) + 1;
    node.style.zIndex = dragZIndex;

    // disable normal mousedown handling (might be selecting text or something)
    // also prevent bubbling
    BU_WindowEvent.stopEvent(ev);

    // TODO: on IE, disable onselectstart 
    
    // set cursor
    node.style.cursor = 'move'

    var dragData = {
	startPageX: startPageX,
	startPageY: startPageY,
	startLeft: startLeft,
	startTop: startTop,
	startZ: startZ,
	dragNode: node
    };

    var moveHandler = function(ev) {BU_DragDrop.dragMove(dragData, ev?ev:window.event)};
    var upHandler = function(ev) {BU_DragDrop.dragEnd(dragData, ev?ev:window.event)};
    // so can unregister
    dragData['moveHandler'] = moveHandler;
    dragData['upHandler'] = upHandler;

    // register for mousemove and mouseup, at document level
    // this is at the document level because there is allegedly a chance with fast motion to get the mouse
    // off of the node in question
    if (document.attachEvent) {
	document.attachEvent('onmousemove', moveHandler);
	document.attachEvent('onmouseup', upHandler);
    }
    else {
	document.addEventListener('mousemove', moveHandler, true); // useCapture = true
	document.addEventListener('mouseup', upHandler, true);
    }
}

/*
 * move the node being dragged according to how the mouse has moved.
 * associated with 'mousemove'
 */
BU_DragDrop.dragMove = function(dragData, ev) {
    var mousePageX = BU_WindowEvent.getMousePos(ev, 'pageX');
    var mousePageY = BU_WindowEvent.getMousePos(ev, 'pageY');

    // new left and top are based on adding mouse distance to starting left and top
    var newLeft = dragData.startLeft + (mousePageX - dragData.startPageX);
    var newTop = dragData.startTop + (mousePageY - dragData.startPageY);

    // apply any clipping (note mouse cursor will go past, but we won't drag past that)
    if (dragData.minLeft && dragData.minLeft > newLeft) newLeft = dragData.minLeft;
    if (dragData.minTop && dragData.minTop > newTop) newTop = dragData.minTop;
    if (dragData.maxRight && dragData.maxRight < newLeft + dragData.dragNode.offsetWidth) newLeft = dragData.maxRight - dragData.dragNode.offsetWidth;
    if (dragData.maxBottom && dragData.maxBottom < newTop + dragData.dragNode.offsetHeight) newTop = dragData.maxBottom - dragData.dragNode.offsetHeight;

    BU_DragDrop.moveTo(dragData.dragNode, newLeft, newTop);

    BU_WindowEvent.stopEvent(ev);
}

BU_DragDrop.moveTo = function(el, left, top) {
      // not correct since didn't start with these.
      //burst.xml.HtmlBox.setStylePos(el, left, top);
      burst.xml.HtmlBox.setBorderBoxLeft(el, left);
      burst.xml.HtmlBox.setBorderBoxTop(el, top);
}

/*
 * restore zIndex, and unsubscribe for mouse events
 */
BU_DragDrop.dragEnd = function(dragData, ev) {
   dragData.dragNode.zIndex = dragData.startZ;

   if (document.attachEvent) {
   document.detachEvent('onmousemove', dragData.moveHandler);
   document.detachEvent('onmouseup', dragData.upHandler);
   }		  
   else {
   document.removeEventListener('mousemove', dragData.moveHandler, true);
   document.removeEventListener('mouseup', dragData.upHandler, true);
   }

   BU_WindowEvent.stopEvent(ev);
}

//:NSEND

