/**
* @file TextSelection.js
*
* Defines burst.web.TextSelection which holds static functions related to text selection.
*
* @author Copyright 2003 Mark D. Anderson (mda@discerning.com)
* @author Licensed under the Academic Free License 1.2 http://www.opensource.org/licenses/academic.php
*/

//=java package burst.web;

/**
Scoping class to hold static functions related to text selection.
*/
//:NSBEGIN TextSelection

burst.web.TextSelection = {};

/*
textarea 
  set caret pos
  get caret pos
  get text line count
  replace text
  set selected text

A caret is a collapsed selection.

Text ranges.

See:
   http://www.w3.org/TR/DOM-Level-2-Traversal-Range/

   http://msdn.microsoft.com/workshop/author/dhtml/reference/objects/obj_textrange.asp
   http://msdn.microsoft.com/workshop/author/dhtml/reference/methods/createrange.asp
   http://msdn.microsoft.com/workshop/author/dhtml/reference/objects/obj_selection.asp
   http://msdn.microsoft.com/workshop/samples/browser/webediting/onepageeditor.htm

   http://www.xulplanet.com/references/xpcomref/ifaces/nsIDOMRange.html
   http://www.xulplanet.com/references/xpcomref/ifaces/nsISelection.html

   http://www.faqts.com/knowledge_base/view.phtml/aid/13562/fid/130
   http://www.xs4all.nl/~ppk/js/selected.html
   http://www.pbwizard.com/Articles/Moz_Range_Object_Article.htm
   http://dev.lophty.com/ahoy/article.htm

W3:
  var ok = document.implementation.hasFeature("Range", "2.0")

IE:
  var range = document.selection.createRange(); // IE-specific
  current selected text = range.text

  range.text = whatever; // IE-only property

  orig_range = document.selection.createRange().duplicate(); 
  ...
  orig_range.text += 'more text';

  range = textarea_node.createTextRange();
  range.findText(text, fromPos);
  range.select();

Mozilla 1.3+ support textarea: http://bugzilla.mozilla.org/show_bug.cgi?id=88049 .
That means that Netscape 6 and 7.0 don't implement setSelectionRange() for textarea.


Mozilla
  var sel = window.getSelection(); // document.getSelection() is deprecated
  sel.anchorOffset
  sel.focusOffset

  current selected text = sel.toString();
  
Safari:
  window.getSelection()

*/




// Martin Honnen, http://www.faqts.com/knowledge_base/view.phtml/aid/13562/fid/130
// minor modifications.
function setSelectionRange(input, selectionStart, selectionEnd) {
    // Mozilla
  if (input.setSelectionRange) {
    input.focus();
    input.setSelectionRange(selectionStart, selectionEnd);
  }
  // IE
  else if (input.createTextRange) {
    var range = input.createTextRange();
    range.collapse(true);
    range.moveEnd('character', selectionEnd);
    range.moveStart('character', selectionStart);
    range.select();
  }
}

function setCaretToEnd (input) {
  setSelectionRange(input, input.value.length, input.value.length);
}
function setCaretToBegin (input) {
  setSelectionRange(input, 0, 0);
}
function setCaretToPos (input, pos) {
  setSelectionRange(input, pos, pos);
}

function selectMatchingString (input, string) {
  var match = new RegExp(string, "i").exec(input.value);
  if (match) {
    setSelectionRange (input, match.index, match.index + match
[0].length);
  }
}

function replaceSelection (input, replaceString) {
   // Mozilla
  if (input.setSelectionRange) {
    var selectionStart = input.selectionStart;
    var selectionEnd = input.selectionEnd;
    input.value = input.value.substring(0, selectionStart)
                  + replaceString
                  + input.value.substring(selectionEnd);
    // if there was a selection before, then make replaced text selected
    if (selectionStart != selectionEnd)
      setSelectionRange(input, selectionStart, selectionStart + replaceString.length);
    // otherwise set caret to end of replaced string
    else 
      setCaretToPos(input, selectionStart + replaceString.length);
  }

  // IE
  else if (document.selection) {
    var range = document.selection.createRange();
    // do nothing if we aren't the currently selected node. 
    if (range.parentElement() == input) {
      var isCollapsed = range.text == '';
      range.text = replaceString;
      // there was selected text beforehand, so select the new stuff
      if (!isCollapsed)  { 
        //it appears range.select() should select the newly 
        //inserted text but that fails with IE
        range.moveStart('character', -replaceString.length);
        range.select();
      }
    }
  }
} 

//:NSEND
