/**
* @file WidgetManager.js
* Defines burst.webui.WiddgetManager , a (typically singleton) class to manage widget instances.
*
* @author Copyright 2003 Mark D. Anderson (mda@discerning.com)
* @author Licensed under the Academic Free License 1.2 http://www.opensource.org/licenses/academic.php
*/

//=java package burst.webui;

/** 
* The (typically singleton) class whose instance manages all widget instances.
*
* @todo The choice of what is a class method/variable versus instance method/variable is somewhat ad hoc here, as one might expect for a singleton.
*/
//=java public class WidgetManager {

function BU_WidgetManager() {
// global counter to generate unique instance ids
	 this.instance_counter_ = 0;
// Array of created instances, indexed by counter
	 this.widgets_by_id_ = [];
}

// private class variable. hash of constructors by widget class name.
BU_WidgetManager.classes_by_name_ = {};

burst.webui.WidgetManager = BU_WidgetManager;

BU_WidgetManager.getWidgetManager = function() {return bu_WidgetManager}
var bu_WidgetManager = new BU_WidgetManager();

// protected instance method.
BU_WidgetManager.prototype.nextInstanceId = function() {return this.instance_counter_++;}

BU_WidgetManager.prototype.registerWidget = function(instance, id) {
  this.widgets_by_id_[id] = instance;
}

/**
* Find a widget instance by its id. 
* If none found, will return if missing_ok, otherwise will throw.
* @param widget_id The id. If not a Number, it will be converted to one (and throws if that fails). 
* @param missing_ok optional. If true, will return null rather than throwing if none found.
*/
//:CLIMETHOD AbstractWidget findByWidgetId(Number widget_id, Boolean missing_ok)
BU_WidgetManager.prototype.findByWidgetId = function (widget_id, missing_ok) {
  if (typeof widget_id != 'number') {
    widget_id = parseInt(widget_id);
    if (isNaN(widget_id)) bu_throw("bogus widget_id '" + widget_id + "'");
  }
  var widget = this.widgets_by_id_[widget_id];
  if (widget) return widget;
  if (missing_ok) return null;
  return bu_throw("no such widget with id '" + widget_id + "'");
}

/**
* Find a widget instance which was constructed using the given input Node.
* Note there may conceivably be multiple widgets associated with the same node,
* but only the first is returned.
* Returns undefined if none.
*/
//:CLIMETHOD AbstractWidget findByNode(Node node)
BU_WidgetManager.prototype.findByNode = function(node) {
  return burst.Alg.find_if_value(this.widgets_by_id_, function(w) {return w.inputNode === node});
}

/**
* Find a widget instance whose <var>appId</var> member variable is equal to the provided app_id.
* Returns undefined if none.
* Note that the application may have used the same appId in multiple widgets, but
* if so, only the first is returned.
* Returns undefined if none.
*/
//:CLIMETHOD AbstractWidget findByAppId(String app_id)
BU_WidgetManager.prototype.findByAppId = function(app_id) {
  return burst.Alg.find_if_value(this.widgets_by_id_, function(w) {return w.appId === app_id});
}

/**
* Register a widget constructor by name. 
* Called at load time for each AbstractWidget subclass.
* This is a class method; there is only one class registry per process, but there can be multiple instance managers. 
*
* NOTE: At this time, the name used at registration time must match identically to that used in
* inline constructor marker attribute values. In particular, we do not attempt to have implied packages.
* If you register 'SortTable' as a name, then you can use 'SortTable' in a attribute value,
* regardless of the package the constructor function lives in.
*/
//:CLCMETHOD void registerWidgetClass(String name, Function ctor)
BU_WidgetManager.registerWidgetClass = function(name, ctor) {
  BU_WidgetManager.classes_by_name_[name] = ctor;
}

// retrieve widget constructor function by name.
BU_WidgetManager.getWidgetConstructor = function(name) {
  return bu_get_soft(BU_WidgetManager.classes_by_name_, name, null);
}

/**
* Process all the constructor nodes in the provided Array.
* Notice that the first constructor to thrown an exception will abort all processing.
*/
//:CLIMETHOD void processConstructorNodes(Array nodes)
BU_WidgetManager.prototype.processConstructorNodes = function(nodes, ctor_must_exist) {
//alert("about to process " + nodes.length + " nodes");
  return burst.Alg.for_each_call2(nodes, this, BU_WidgetManager.prototype.processConstructorNode, ctor_must_exist);
}

/*
If it contains no periods, then prepend "burst.widgets.webui.".
If it contains periods, leave it alone.
*/
BU_WidgetManager.wtype_to_module_name = function(wtype) {
    return (wtype.indexOf('.') == -1) ? 'burst.webui.widgets.' + wtype : wtype;
}

/**
* Process the single constructor node.
* It will get the value of the marker attribute to determine the widget type (class name).
* If the script file defining that widget is not yet loaded, and !ctor_must_exist it will attempt to 
* load the appropriate script. In that case, completion is asynchronous (the widget may not be defined
* at the time the function returns, and nothing will be returned synchronously).
*
* @todo Right now, dynamic widget script loading seems not to work in Safari or Opera, just Moz and IE.
*/
//:CLIMETHOD burst.webui.widgets.AbstractWidget processConstructorNode(Node node, Boolean ctor_must_exist)
BU_WidgetManager.prototype.processConstructorNode = function(node, ctor_must_exist) {
  var wtype = BU_WidgetManager.getInlineWidgetType(node);
  var ctor = BU_WidgetManager.getWidgetConstructor(wtype);

  var module_name = BU_WidgetManager.wtype_to_module_name(wtype);

  // no constructor function, either load or throw
  if (!ctor) {
  //alert("no such constructor '" + wtype + "'");
     if (ctor_must_exist) bu_throw("attempt to construct a widget of unknown type '" + wtype + "'");

     // script loading might be asynchronous. we provide a done_handler that does the construction when the script is loaded.
     var this_ = this;
     var done_handler = function() {
        ctor = BU_WidgetManager.getWidgetConstructor(wtype);
        if (!ctor) bu_throw("widget constructor for '" + wtype + "' still not defined after loading widget script");
        return this_.call_ctor(ctor, node);
     };

     // alert("starting async loading of widget script with module_name " + module_name);

     // get the url to the script we have to load
     // for now, don't rely on the runtime to always have done_handler support.
     //var full_url = bu_ScriptLoader.resolveWidgetUrlByName(wtype);
     //return bu_Runtime.readEval(full_url, done_handler);
     return bu_ScriptLoader.requireModule('burst.webui.WidgetManager', module_name, done_handler);
  }
  return this.call_ctor(ctor, node);
}

/**
 * A convenience, equivalent to <code>this.processConstructorNodes(WidgetManager.findConstructorNodes())</code> .
*/
//:CLIMETHOD void findAndProcessConstructorNodes()
BU_WidgetManager.prototype.findAndProcessConstructorNodes = function() {
  var nodes = BU_WidgetManager.findConstructorNodes();
  // alert("found " + nodes.length + " nodes");
  this.processConstructorNodes(nodes);
}

/**
 * A convenience, equivalent to <code>this.processConstructorNodes(WidgetManager.getConstructorNodes(ids, widget_id_strict))</code> .
*/
//:CLIMETHOD void getAndProcessConstructorNodes(Array ids, Boolean widget_id_strict)
BU_WidgetManager.prototype.getAndProcessConstructorNodes = function(ids, widget_id_strict) {
//alert("in getAndProcessConstructorNodes");
   var nodes = BU_WidgetManager.getConstructorNodes(ids, widget_id_strict);
   this.processConstructorNodes(nodes);
}


// private method to do the actual call to the constructor
BU_WidgetManager.prototype.call_ctor = function(ctor, node) {
    if (typeof ctor !== 'function') { 
      var mess = "(WidgetManager.js) ctor not a function: " + ctor;
      bu_alert(mess);
      bu_throw(mess);
    }
    // alert("about to call constructor: " + ctor);
    // second arg says to parse_properties
    return new ctor(node, true);
}



// class constant 
BU_WidgetManager.MARKER_ATT = 'buType';
BU_WidgetManager.APPID_ATT = 'buAppId';
BU_WidgetManager.PROPERTIES_ATT = 'buProperties';
BU_WidgetManager.PROPERTY_ATT = 'buProperty';
BU_WidgetManager.VALUE_ATT = 'buValue';

/*
* Utility to set a property from an element like &lt;div buProperty="fred" buValue="17"/>
* If buProperty is found but buValue is not, then innerHTML is used (not any 'value' attribute).
*/
BU_WidgetManager.maybe_set_inline_prop = function(obj, node, marker_ok) {
  // marker_ok is false for children; it is so we skip child inline constructors when setting parent.
  if (!marker_ok && BU_WidgetManager.isInlineConstructor(node)) return false;
  var natt = burst.xml.DomUtil.getAttribute(node, BU_WidgetManager.PROPERTY_ATT);
  if (!natt) return false;
  var vatt = burst.xml.DomUtil.getAttribute(node, BU_WidgetManager.VALUE_ATT);
  if (vatt) {
    obj[natt] = vatt;
  }
  else {
    obj[natt] = burst.xml.DomUtil.getInnerHTML(node);
  }
  return true;
}



/**
* Extract the widget class name from an inline constructor. Must be found in the attribute unless missing_ok is true.
* The value is not checked, but just returned.
*/
//:CLCMETHOD String getInlineWidgetType(Node node, Boolean missing_ok)
BU_WidgetManager.getInlineWidgetType = function(node, missing_ok) {
  if (node.nodeType !== Node.ELEMENT_NODE) return BU_UNDEFINED;
  var butype = burst.xml.DomUtil.getAttribute(node, BU_WidgetManager.MARKER_ATT);
  if (!butype && (arguments.length < 2 || !missing_ok)) bu_throw("Node has no '" + BU_WidgetManager.MARKER_ATT + "' attribute to indicate widget type");
  return butype;
}

/**
* Returns whether the provided node is an inline constructor or not (by checking for the marker attribute). 
*/
//:CLCMETHOD String isInlineConstructor(Node node)
BU_WidgetManager.isInlineConstructor = function(node) {
  return BU_WidgetManager.getInlineWidgetType(node, true) ? true : false;
}

/**
* Iterates through the Nodes in the DOM document looking for ones with the marker attribute.
* Returns an Array of the Nodes.
* It does search recursively, but it does not descend the element tree underneath any inline
* constructor; it only finds top-level inline constructors.
* This function does not process the constructors, it just returns the Nodes.
* This function should not be called prior to the document onload event (since there may not
* be much of a tree yet).
* @param doc Optional, defaults to document.
*/
//:CLCMETHOD Array findConstructorNodes(Document doc)
BU_WidgetManager.findConstructorNodes = function(doc, node, arr) {
  if (arguments.length < 1) doc = document;
  if (arguments.length < 2) return arguments.callee(doc, doc.documentElement, []);
  if (BU_WidgetManager.isInlineConstructor(node)) {
     arr.push(node);
     return arr;
  }
  burst.xml.DomUtil.for_children(node, function(child) {
    BU_WidgetManager.findConstructorNodes(doc, child, arr);
  });
  return arr;
};

/**
* Unlike findConstructorNodes, which iterates through the whole DOM, this
* returns an Array of Node objects matching just the provided element ids.
*
* If any id does not exist as a DOM element, or if the found node has no marker attribute, the
* handling is according to be_strict. If be_strict, an Error is thrown.
* If !be_strict a warning is issued and the id is skipped.
* This function should not be called prior to the document onload event.
* @param ids The Array of id values, or String of comma-separated ids.
* @param be_strict Optional, defaults to false. Whether to throw an Error (versus just warn) about bad ids
* @param doc Optional, defaults to document.
*/
//:CLCMETHOD Array getConstructorNodes(Array ids, Boolean be_strict, Document doc)
BU_WidgetManager.getConstructorNodes = function(ids, be_strict, doc) {
  if (typeof be_strict == 'undefined') be_strict = false;
  if (typeof doc == 'undefined') doc = document;
  if (typeof ids == 'string') {
    var idarray = ids.split(/\s*,\s*/);
    ids = idarray;
  }
  var arr = [];
  burst.Alg.for_each(ids, function(id) {
    var node = doc.getElementById(id);
    var mess;
    if (!node) {
       mess = "(WidgetManager.js) There is no Node with ID '" + id + "'";
       if (be_strict) bu_throw(mess); else bu_warn(mess + '; skipping it');
       return;
    }
    if (!BU_WidgetManager.isInlineConstructor(node)) {
       mess = "The Node with ID '" + id + "' does not appear to be an inline constructor";
       if (be_strict) bu_throw(mess); else bu_warn(mess + '; skipping it');
       return;
    }
    arr.push(node);
  });
  return arr;
}



// class properties
BU_WidgetManager.PROP_DEFS = [
//:CPROPERTYDEFS
  new burst.reflect.PropertyDefBoolean({
    name: 'findWidgets',
    defaultValue: false,
    description: "Whether to search all elements on the page for inline constructors (elements with the marker attribute)"
  }),
  new burst.reflect.PropertyDefString({
    name: 'initWidgetIds',
    description: "If specified, a list of IDs of elements on the page which are inline constructors. Can be an Array or a comma-separated String."
  })
//:ENDPROPERTYDEFS
];

// in onDocumentLoad, not onConfig since we need the document to be loaded.
BU_WidgetManager.onDocumentLoad = function() {
   var props = bu_Config.setObjectValues({}, BU_WidgetManager.PROP_DEFS, 'burst.webui.WidgetManager', true);
   if (props.findWidgets) {
     bu_WidgetManager.findAndProcessConstructorNodes();
   }
   if (props.initWidgetIds) {
     var widget_id_strict = true;
     bu_WidgetManager.getAndProcessConstructorNodes(props.initWidgetIds, widget_id_strict);
   }
   else {
   //alert("props.initWidgetIds not set"); 
   }
};

//=java } // class WidgetManager

bu_loaded('burst.webui.WidgetManager', ['burst.Alg','burst.xml.DomUtil'], BU_WidgetManager);
