/**
* @file AbstractWidget.js
* Defines burst.webui.widgets.AbstractWidget
*
* @author Copyright 2003 Mark D. Anderson (mda@discerning.com)
* @author Licensed under the Academic Free License 1.2 http://www.opensource.org/licenses/academic.php
*/

//=java package burst.webui.widgets;

/** 
* Abstract base class for widgets. These instances are managed by a burst.webui.WidgetManager instance.
* 
* <h3>Identifying and Finding Widgets</h3>
* The provided node might be used for an inline constructor,
* and it will be replaced.
* Or it might be a node that will be left as is, but some
* behavior will be attached.
* Thus there might be 3 different objects that a person might want
* to relate:
* <ol>
* <li>The input Node (which might be dynamically created by the caller, or from the document).
* <li>The instance of a AbstractWidget subclass (hereafter, "widget instance").
* <li>Any top-level Node that the widget instance may create as part of its implementation.
* </ol>
* It is possible to recover any widget instance that was constructed using a given Node using
* the class method <code>findByNode</code>. It is conceivable that there may be multiple widgets
* using the same Node, but the function just returns the first one.
*
* From a widget instance, it is possible to determine its input Node by the
* public variable <var>inputNode</var>.
*
* The library maintains globally unique numerical ids of every widget instance.
* This can be obtained from an instance via the public variable <var>widgetId</var>.
* A widget instance from its widget id by <code>findByWidgetId</code>
*
* A widget instance also holds a public readwrite String variable called <var>appId</var>.
* The library does not care what it is, or if it is even set. It is available for
* use by the client programmer. A widget can be recovered from a given appId using
* <code>findByAppId</code>
*
* The <var>id</var> attribute of the input node, if any, is not used by us.
* Some widgets will remove their input node from the document, and id attributes are not
* copied over ito any replacement node(s).
*/
//:CLABEGIN AbstractWidget

// public member variables
/** The Node used in this widget instance's constructor. readonly. */
//:CLIVAR Node inputNode

/** The generated globally unique id of this widget instance. readonly. */
//:CLIVAR Number widgetId

/** Any id associated with the instance by the client programmer. readwrite and optional */
//:CLIVAR String appId

/**
* constructor.
* If parse_properties, call parseProperties(true, true).
* Otherwise, it calls <var>burst.reflect.PropertyDef.setDefaultEach</var>
*
* @param node The input node (either in the document, or just created by the caller).
* @param propdefs An Array of PropertyDef instances, used for configuration. 
* @param parse_properties Whether to parse the node for any configuration properties
*/
//:CLCONSTRUCT AbstractWidget(Node node, Array propdefs, Boolean parse_properties)
function BU_AbstractWidget(node, propdefs, parse_properties) {
	 if (arguments.length == 0) return;

  // superclass constructor
  //BU_Configurable.call(this, propdefs);
  this.propdefs_ = propdefs;

  // public member variables
  this.inputNode = node;
  this.widgetId = bu_WidgetManager.nextInstanceId();
  // don't set to undefined for IE5.0 PC
  this.appId = null;
  this.childWidgets = null;
  this.parentWidget = null;

  // set attribute of node with id of corresponding widget
  if (false) {
    dom_node.setAttribute('BU_compID', this.widgetId);
    dom_node.id = this.widgetId;
  }

  // register instance by id
  // TODO: note that we just assume which singleton we want; a better relationship with WidgetManager.call_ctor would be ideal.
  bu_WidgetManager.registerWidget(this, this.widgetId);

  // parse configuration
  if (parse_properties) this.parseProperties(true, true);
  else if (this.propdefs_) burst.reflect.PropertyDef.setDefaultEach(this, this.propdefs_);
}
burst.webui.widgets.AbstractWidget = BU_AbstractWidget;

//bu_inherits(BU_AbstractWidget,BU_Configurable);

/**
* Add the provided widget instance as a child of this widget instance parent.
* Returns the child.
*/
//:CLIMETHOD AbstractWidget addChildWidget(AbstractWidget child_widget)
BU_AbstractWidget.prototype.addChildWidget = function(child_widget) {
  child_widget.parentWidget = this;
  if (!this.childWidgets) this.childWidgets = [child_widget];
  else this.childWidgets.push(child_widget);
  return child_widget;
}

/**
* Initialize the widget instance from any "inline constructor" attributes
* as well as special immediate children.
* It uses the <var>inputNode</var> member variable (probably set by the constructor).
* It also uses the <var>propdefs_</var> member variable (set to some constant array
* when the subclass calls the AbstractWidget super class constructor).
*
* There are a few different ways property values are found from a node:
* <ul>
* <li>Any "buAppId" attribute becomes a value for the <var>appId</var> member.
* <li>Any declared property named "fooBar" will obtain its value from a attribute with the same name.
* <li>If the node has a "buProperties" attribute, its value is parsed to set multiple properties.
*     The value syntax is CSS-like, for example: <pre>"fooBar: 1; title: 'some title'"</pre>
* <li>Within the node, if it has a 'buProperty' attribute, and with
*     any immediate child which has a 'buProperty' attribute and no 'buType' attribute,
*     set the property whose name is the value of the 'buProperty' attribute.
*     The value of the property is the value of the 'buValue' property, if any, otherwise it is the innerHTML
*     of the node with the buProperty attribute.
* </ul>
*
* @todo also try xmlns attributes within some namespace uri, for example getAttributeNS(BURST_WIDGET_NS, unqualname)
*
* @param do_defaults as with burst.props.AbstractProperties.setObjectValues
* @param do_mandatory as with burst.props.AbstractProperties.setObjectValues
*/
//:CLIMETHOD void parseProperties(Boolean do_defaults, Boolean do_mandatory)
BU_AbstractWidget.prototype.parseProperties = function(do_defaults, do_mandatory) {
  bu_debug("in AbstractWidget.prototype.parseProperties("," do_defaults=",do_defaults," do_mandatory=",do_mandatory,")");
  var node = this.inputNode;
  // application-specified id
  var app_id = burst.xml.DomUtil.getAttribute(node, BU_WidgetManager.APPID_ATT);
  if (app_id) this.appId = app_id;

  // start an array of burst.props.AbstractProperties instances that we will use for property lookup
  var props_arr = [];

  // property values from individual node attributes
  var attprops = new burst.props.NodeAttributesProperties(node);
  props_arr.push(attprops);

  // property values from a AbstractProperties attribute value
  var prop_text = burst.xml.DomUtil.getAttribute(node, BU_WidgetManager.PROPERTIES_ATT);
  if (prop_text) {
    var prop_text_map = burst.xml.HtmlUtil.parseStyle(prop_text);
    props_arr.push(new burst.props.MapProperties(prop_txt_map));
  }

  // look for property values in node attributes and in child elements
  var propvals = {};
  // node attributes
  BU_WidgetManager.maybe_set_inline_prop(propvals, node, true);
  // child elements
  burst.xml.DomUtil.for_children(node, function(child) {BU_WidgetManager.maybe_set_inline_prop(propvals, child, false)}, Node.ELEMENT_NODE);
  props_arr.push(new burst.props.MapProperties(propvals));

  // combine them all and set values
  var group_props = new burst.props.GroupProperties(props_arr);
  var prefix = null;
  group_props.setObjectValues(this, this.propdefs_, prefix, do_defaults, do_mandatory);
}

//:CLEND AbstractWidget

/**
* Abstract subclass of AbstractWidget for widgets that participate as form controls.
* This adds two readonly variables: <var>controlName</var> and <var>controlValue</var>.
* 
* When parsing configuration:
* - controlName is the first of these attributes found: 'buControlName', 'name', 'id'
* - controlValue is the first of these: 'buControlValue', 'value', innerHTML
*/
//=java public abstract class ControlWidget extends AbstractWidget {

/** 
* Construct a ControlWidget. 
* Calls superclass, then figures out values for member variables <var>controlName</var> and <var>controlValue</var>
* if possible.
*/
//:CLCONSTRUCT ControlWidget(Node node, Array propdefs, Boolean parse_properties) : AbstractWidget(node, propdefs, parse_properties)
var BU_ControlWidget = function(node, propdefs, parse_properties) {
   BU_AbstractWidget.call(this, node, propdefs, parse_properties);

   this.controlName = null;
   this.controlValue = null;
   if (parse_properties) this.parseControlName();
   else {
   }
}

bu_inherits(BU_ControlWidget, BU_AbstractWidget);
burst.webui.widgets.ControlWidget = BU_ControlWidget;

BU_ControlWidget.NAME_ATTS = ['buControlName', 'name', 'id'];
BU_ControlWidget.VALUE_ATTS = ['buControlValue', 'value'];

BU_ControlWidget.prototype.parseControlName = function() {
  var node = this.inputNode;

  this.controlName = burst.Alg.find_if_value(BU_ControlWidget.NAME_ATTS, function(attname) {
    return burst.xml.DomUtil.getAttribute(node, attname);
  });
  if (!this.controlName) bu_throw("can't find controlName in node using any of the attributes: " + burst.Lang.uneval(BU_ControlWidget.NAME_ATTS));

  this.controlValue = burst.Alg.find_if_specified(BU_ControlWidget.VALUE_ATTS, function(attname) {
    return burst.xml.DomUtil.getAttribute(node, attname);
  });
  if (!this.controlValue) this.controlValue = burst.Text.trim(burst.xml.DomUtil.getInnerHTML(node));
}

/*
replace constructor node
set_n.parentNode.replaceChild(localNode, set_n);

*/

//:CLEND



