/**
* @file SortTable.js
* Defines burst.webui.widgets.SortTable, a widget which makes an HTML table sortable by clicking on its column heads.
*
* @author Copyright (C) 2003, Mark D. Anderson (mda@discerning.com).
* @author Released under the Academic Free License.
* @version $Id: SortTable.js,v 1.3 2004/05/16 02:45:25 mdaoh Exp $
*/

//=java package burst.webui.widgets;
/** 
Subclass of AbstractWidget providing sort behavior for an HTML table.

<h3>DOM implementation</h3>
We do not create a new DOM hierarchy for this widget, but instead just
use the Node provided in the constructor. That means all styling is
preserved.

<h3>DOM operations</h3>
The only DOM create/remove operations we do are:
- to re-order the tr rows
- to add/remove/change an arrow icon the head of the primary sorting column

It is required that there be a single head row, table/thead/tr.
It must contain either th or td children, from which SortHead child widgets are created.
If you change the structure of the thead or below by your own DOM calls
after creating this widget you must call the <var>initHeads()</var> method.

The data must be in <code>tbody/tr</code> rows.
You may add rows at any time.

<h3>Inline constructor</h3>

You may associate inline constructor properties either with the <code>table</code>
Node or with any th/td in a head. The available properties in each context are
overlapping, since you can provide defaults at the table level for most head properties.

@todo Set (add) an additional class whenever a sort head is clicked, so that css can be used to visually show sort state.

*/
//:CLBEGIN SortTable extends AbstractWidget

/** 
* Widget constructor.
*/
//:CLCONSTRUCT SortTable(Node dom_node, Boolean parse_properties) : AbstractWidget(dom_node, SortTable.PROP_DEFS, parse_properties)
function BU_SortTable(dom_node, parse_properties) {
  if (!BU_SortTable.PROP_DEFS) bu_throw("internal error: no SortTable.PROP_DEFS");
  burst.webui.widgets.AbstractWidget.call(this, dom_node, BU_SortTable.PROP_DEFS, parse_properties);

  // private state 
  this.lastSortedColumnIndex_ = -1;
  this.lastSortReversed_ = false;
  this.lastSortedImg_ = null;

  this.initFromNode(dom_node, parse_properties);
}
burst.webui.widgets.SortTable = BU_SortTable;

BU_SortTable.PROP_DEFS = [
//:IPROPERTYDEFS
  new burst.reflect.PropertyDefBoolean({
	name: 'sortOnLoad', 
	defaultValue: false, 
	description: "Whether to sort the table upon load."
	}),
  new burst.reflect.PropertyDefBoolean({
	name: 'sortOnLoadAscending',
	defaultValue: true, 
	description: "If false, reverses the sense of the onLoad sort."
	}),
  new burst.reflect.PropertyDefBoolean({
	name: 'sortEnabled', 
	defaultValue: true, 
	description: "Whether user-directed sorting is enabled (after load)."
	}),
  new burst.reflect.PropertyDefImageUrl({
	name: 'upImgSrc', 
	defaultValue: 'sortedUp.png', 
	description: "The url to the image for the up arrow (currently ascending)"
	}),
  new burst.reflect.PropertyDefImageUrl({
	name: 'downImgSrc',
	defaultValue: 'sortedDown.png',
	description: "The url to the image for the down arrow (currently descending)"
	})
//:ENDPROPERTYDEFS
/* unimplemented
  new burst.reflect.PropertyDefBoolean({
	name: 'staticRows', 
	defaultValue: true, 
	description: "Whether to assume that the rows in the table will not change after creation (for example by external DOM usage)."
	}),
  new burst.reflect.PropertyDefBoolean({
	name: 'staticHeads', 
	defaultValue: true, 
	description: "Whether to assume that the heads in the table will not change after creation (for example by external DOM usage)."
	})
*/
];

bu_inherits(BU_SortTable,burst.webui.widgets.AbstractWidget);

burst.webui.WidgetManager.registerWidgetClass('SortTable', BU_SortTable);

BU_SortTable.prototype.initHeads = function(parse_properties) {
  this.thead_node_ = burst.xml.DomUtil.firstChildWithTagName(this.inputNode, 'thead', true);
  this.staticHeads = true;

  // get the head nodes (first try 'th', and if there are none, try 'td')
  var thead_tr = burst.xml.DomUtil.firstChildWithTagName(this.thead_node_, 'tr', true);
  this.head_nodes_ = burst.xml.DomUtil.childrenNamed(thead_tr, 'th', this.staticHeads);
  if (this.head_nodes_.length == 0)
    this.head_nodes_ = burst.xml.DomUtil.childrenNamed(thead_tr, 'td', this.staticHeads);
  if (this.head_nodes_.length == 0) throw "No thead/tr/th or thead/tr/td children";
  bu_debug("found ", this.head_nodes_.length, " heads with this.staticHeads=", this.staticHeads);

  var this_ = this; // for anonymous function bodies

  //  parse out head sort specifications
  this.head_widgets_ = burst.Alg.transform(this.head_nodes_, function(head_node, col_ind) {
    var shead = new BU_SortHead(head_node, col_ind, parse_properties);
    this_.addChildWidget(shead);
    return shead;
  });

  // Add event listeners to head elements. 
  // We could put a single listener to click events in the thead, rather than
  // associating with each individual head. But browsers vary in which descendant
  // they might identify as the target (might not be the th).
  if (this.sortEnabled) {
    bu_debug("attaching listeners to ", this.head_nodes_.length, " heads");
    for(var i=0;i<this.head_nodes_.length;++i) {
       // bu_alert("attaching listener to head " + i + " with tagname " + this.head_nodes_[i].nodeName);
       burst.web.WindowEvent.addEventListener(this.head_nodes_[i], 'click', function(evt) {
          this_.sortColumnHead(evt.currentTarget);
       });
    }
  }
}

BU_SortTable.prototype.initFromNode = function(dom_node, parse_properties) {
  this.tbody_node_ = burst.xml.DomUtil.firstChildWithTagName(dom_node, 'tbody', true);
  this.ownerDocument_ = burst.xml.DomUtil.ownerDocument(dom_node);

  this.initHeads(parse_properties);

  // maybe do initial sort
  if (this.sortOnLoad) this.doInitialSort();
}

// have to pre-process all columns used, then use a combined sort comparator that 
// applies them according to sortOrder.
// TODO: "sortOnLoadAscending" (vs. old "onLoadSortAscending")
BU_SortTable.prototype.doInitialSort = function() {
    var sort_col_inds = [];

    // loop through head widgets finding ones with sortOrder set
    burst.Alg.for_each(this.head_widgets_, function(headwidget) {
      if (typeof headwidget.sortOrder == 'number') {
        if (typeof sort_col_inds[headwidget.sortOrder] != 'undefined') 
           bu_throw("multiple heads with the same sortOrder " + headwidget.sortOrder);
        sort_col_inds[headwidget.sortOrder] = headwidget.col_ind_;
      }
    });

    // check for holes
    for(var j=0;j<sort_col_inds.length;++j) {
      if (typeof sort_col_inds[j] == 'undefined')
         bu_throw("You specified " + sort_col_inds.length + " heads with sortOrder, but none for order number " + j);
    }

    // if no head has sortOrder is specified, use the first column.
    if (sort_col_inds.length == 0) {
       this.head_widgets_[0].sortOrder = 0;
       sort_col_inds.push(0);
    }
    // a trick to get correct direction
    this.lastSortedColumnIndex = sort_col_inds[0];
    this.lastSortReversed_ = this.sortOnLoadAscending;
    this.sortColumns(sort_col_inds);
}

BU_SortTable.prototype.sortColumns = function(sort_col_inds) {
  this.sortColumns_(sort_col_inds);
}

BU_SortTable.prototype.sortColumns_ = function(sort_col_inds) {
  //bu_alert("(SortTable.js) in sortColumns_");
  var primary_col_ind = sort_col_inds[0];

  var prev_col_ind = this.lastSortedColumnIndex_;
  if (primary_col_ind == this.lastSortedColumnIndex_) {
    this.lastSortReversed_ = !this.lastSortReversed_;
  }
  else {
    this.lastSortedColumnIndex_ = primary_col_ind;
    this.lastSortReversed_ = false;
  }

  //var row_nodes = burst.xml.DomUtil.childrenNamed(this.tbody_node_, 'tr', this.staticRows);
  // we want a snapshot because we don't want it changing while we do appendChild
  var row_nodes = burst.xml.DomUtil.getChildrenWithTagName(this.tbody_node_, 'tr');
  var nrows = row_nodes.length;
  var ctr=0;
  var index_array = burst.Alg.generate_n(new Array(nrows), nrows, function() {return ctr++});

  var func = sort_col_inds.length > 1 ? this.makeMultipleColumnSortFunction(sort_col_inds, row_nodes) :
       this.makeColumnSortFunction(primary_col_ind, row_nodes);

  // do the sort
  index_array.sort(func);

  //bu_alert("index array after sorting on columns " + sort_col_inds + ": " + index_array.join('|'));
  if (this.lastSortReversed_) index_array.reverse();

  // re-order rows
  var tbody = this.tbody_node_;
  burst.Alg.for_each(index_array, function(ind) {
    //bu_alert("appending row with index " + ind);
    tbody.appendChild(row_nodes[ind]);
  });

  // remove the old sort arrow, if there was a previous one (regardless of whether same
  // column or not, since direction has changed).
  // NOTE: It would probably be faster, if we already have a lastSortedImg_,
  // to simply change the src attribute, but that optimization only helps on same column.
  // @todo A better improvement would be to have member variables for the at most two image elements,
  // and do a combination replaceChild, removeChild, appendChild as appropriate, possibly
  // leaving one image as an orphan.
  if (this.upImgSrc) {
    if (this.lastSortedImg_) {
      this.head_nodes_[prev_col_ind].removeChild(this.lastSortedImg_);
    }
    this.lastSortedImg_ = this.lastSortReversed_ ? 
       burst.xml.HtmlUtil.createImg(this.ownerDocument_, this.downImgSrc, 'Sorted Down', 'Sorted Down') :
       burst.xml.HtmlUtil.createImg(this.ownerDocument_, this.upImgSrc, 'Sorted Up', 'Sorted Up');
    this.head_nodes_[this.lastSortedColumnIndex_].appendChild(this.lastSortedImg_);
  }
}


BU_SortTable.prototype.sortColumnHead = function(head_node) {
  //bu_alert("(SortTable.js) in sortColumnHead");
  var index = burst.Alg.find(this.head_nodes_, head_node);
  if (index == -1) bu_throw("could not find a table head matching the event node: " + head_node);  
  this.sortColumn(index);
}

// return an Array of post-processed values for a column
BU_SortTable.prototype.getValueArray = function(col_ind, row_nodes, convert, stats) {
  if (!convert) convert = burst.Alg.identity;
  var row_count = 0;
  var undefined_count = 0;
  var value_array = burst.Alg.transform(row_nodes, function(node) {
     //bu_debug("extracting value from row ", row_count, ": ", node);
     var cell_node = burst.xml.DomUtil.nthChildWithTagName(node, col_ind, 'td', false);
     var val;
     if (cell_node) {val = convert(burst.Text.ltrim(burst.xml.DomUtil.getInnerText(cell_node)));}
     else {bu_warn("no column td " + col_ind + " in row " + row_count);}
     var is_bad = false;
     if ((typeof val == 'undefined') || val === null || (typeof(val) == 'number' && isNaN(val)) || 
         typeof(val) == 'string' && val == '') {
        is_bad = true;
        ++undefined_count;
     }
     row_count++;
     if (is_bad) return BU_UNDEFINED; else return val;
  });
  //bu_alert("got value_array for column " + col_ind + ": " + value_array.join('|'));
  if (typeof stats != 'undefined') stats[0] = undefined_count;
  return value_array;
}

// loop through columns, composing sort functions
BU_SortTable.prototype.makeMultipleColumnSortFunction = function(sort_col_inds, row_nodes) {
  var func = null;
  var this_ = this;
  burst.Alg.for_each(sort_col_inds, function(col_ind) {
     var col_func = this_.makeColumnSortFunction(col_ind, row_nodes);
     func = func ? burst.Functional.compose_sort(col_func, func) : col_func;
  });
  return func;
}

// return a binary function of two array indexes.
BU_SortTable.prototype.makeColumnSortFunction = function(col_ind, row_nodes) {
  if (col_ind > this.head_widgets_.length) bu_throw("no such column index " + col_ind);
  var head_widget = this.head_widgets_[col_ind];
  var comparator = burst.Comparator.getComparator(head_widget.sortComparator);
  bu_debug("for column ", col_ind, " got comparator: ", comparator);
  // alert("sorting with column " + col_ind + " using comparator " + comparator.name_);

  var stats = [];
  var value_array = this.getValueArray(col_ind, row_nodes, comparator.convert, stats);
  var undefined_count = stats[0];
  
  var compare = comparator.compare || burst.Comparator.stringCompare;

  var func;

  // if there are no undefined values, things are simple
  if (undefined_count == 0) {
    bu_debug("no undefined values for column ", col_ind);
    func = function(ai,bi) {
      return compare(value_array[ai], value_array[bi]);
    }
  } 

  // otherwise we have to deal with incomparable values
  else {
    bu_debug("have ", undefined_count, " undefined values for column ", col_ind);
    var bad_result;
    switch(head_widget.sortIncomparable || this.sortIncomparable) {
      case 'first': bad_result = -1; break;
      case 'last': bad_result = 1; break;
      case 'lowest': bad_result = this.lastSortReversed_ ? 1 : -1;
      case 'highest': bad_result = this.lastSortReversed_ ? -1 : 1;
      default: bu_throw("unknown value for sortIncomparable: '" + this.sortIncomparable + "'");
    }
    func = function(ai,bi) {
      var a = value_array[ai];
      var b = value_array[bi];
      //bu_debug("compare(",ai,',',bi,") has a=", a, " b=", b);
      //if (a == undefined) return b == undefined ? (ai - bi) : bad_result;
      //if (b == undefined) return (- bad_result);
      if (typeof a == 'undefined') return ( (typeof b == 'undefined') ? (ai - bi) : bad_result );
      if (typeof b == 'undefined') return (- bad_result);
      return compare(a, b);
    };
  }
  return func;
}

/** 
* Sort by the column with the specified index (zero-based).
* If the table was just sorted on the same column, the direction of the sort is toggled.

<h3>Implementation Approach</h3>

Approach #1:
<ol>
<li>Make two arrays, of incomparables and comparables.
<li>Create a an associative array between every comparable value and the indexes that have it.
<li>Call the Array.sort directly on the value array of comparables.
<li>After sorting, iterate through values, looking up the indexes in the original order.
<li>Reorder the rows, dealing with the incomparables as well.
</ol>

Approach #2:
<ol>
<li>Make an array of indexes and an array of values.
<li>Sort the array of indexes, with the comparison function indirecting through indexes
to values and dealing with incomparables values.
<li>Reorder the rows.
</ol>

Approach #1 has the benefit of allowing a simpler sort comparator (because it sorts
values directly, and excludes the incomparables).
In fact, for a native String sort, it can call Array.sort without the second argument
at all.

Approach #2 has the benefit of simplicity, and avoids creating a reverse hash
from value to index. It also is simpler to extend to the case of sorting
by multiple columns at once (primary sort, secondary sort, etc.).

We use Approach #2.
*/
BU_SortTable.prototype.sortColumn = function(col_ind) {
  bu_debug("sortColumn(", col_ind, ")");
  this.sortColumns_([col_ind]);
}

//=java } // class SortTable



/**
* Subclass of AbstractWidget to hold per-column sorting configuration.
* It is a full widget on its own, since it can have html inline constructors.
*/
//:CLBEGIN SortHead extends AbstractWidget
/**
* Construct a SortHead from a DOM Node (th or td)
*/
//:CLCONSTRUCT SortHead(Node dom_node, Number col_ind, Boolean parse_properties) : AbstractWidget(dom_node, SortHead.PROP_DEFS, parse_properties)
function BU_SortHead(dom_node, col_ind, parse_properties) {
  bu_debug("SortHead(", dom_node, ',', col_ind, ',', parse_properties, ")");
  burst.webui.widgets.AbstractWidget.call(this, dom_node, BU_SortHead.PROP_DEFS, parse_properties);
  // alert("made sort head " + col_ind + " with comparator " + this.sortComparator + " getAttribute=" + dom_node.getAttribute('sortComparator') + " parse_properties=" + parse_properties);
  this.col_ind_ = col_ind;
}
burst.webui.widgets.SortHead = BU_SortHead;
bu_inherits(BU_SortHead,burst.webui.widgets.AbstractWidget);


BU_SortHead.PROP_DEFS = [
//:IPROPERTYDEFS
  new burst.reflect.PropertyDefString({
	name: 'sortComparator', 
	defaultValue: 'string:cs', 
	enumArray: burst.Lang.keys(burst.Comparator.ALL), 
	description: "The name of a sort comparator (from burst.Comparator)."
	}),
  new burst.reflect.PropertyDefBoolean({
	name: 'sortAscending',
	defaultValue: true,
	description: "Whether the initial sort should be in the ascending direction (the default for comparator) or the reverse."
	}),
  new burst.reflect.PropertyDefString({
	name: 'sortIncomparable', 
	defaultValue: 'last', 
	//enumMap: {first: 'first', last: 'last', lowest: 'lowest', highest: 'highest'},
	enumArray: ['first', 'last', 'lowest', 'highest'],
        description: "How incomparable values such as blank are ordered."
	}),
  new burst.reflect.PropertyDefNumber({
	name: 'sortOrder', 
	description: "The 0-based order of this column in onLoad sorting."
	})
//:ENDPROPERTYDEFS
];
//=java } // class SortHead



bu_loaded('burst.webui.widgets.SortTable');

