/**
* @file DomUtil.js
*
* Defines burst.xml.DomUtil which contains static generic DOM-related utility functions.
* There are other files that are also DOM-related:
* - burst/web/UserAgent.js :       browser identification.
* - burst/xml/XPath.js :           a simple XPath implementation.
* - burst/io/IframeURIRequest.js : iframe manipulation. 
* - burst/xml/XmlDoc.js :          xml dom object
*
* @author Copyright 2003 Mark D. Anderson (mda@discerning.com)
* @author Licensed under the Academic Free License 1.2 http://www.opensource.org/licenses/academic.php
*/

// turn on internal debugging
var BU_DBG_DOM = false;

function bu_debug_Dom() {
  if (BU_DBG_DOM) bu_alert.apply(null, arguments);
}

//=java package burst.xml;
/**
* Scoping class to hold static generic DOM-related utility functions; see also burst.xml.XmlDoc .
*/
//=java public class DomUtil {

burst.xml.DomUtil = {};


/**
* Replace: less-than, greater-than, and ampersand with their html parameters.
* @param s The string to escape.
* @param quot_too Optional. If true, also escape double-quote as quot.
*/
//:NSFUNCTION String escapeHtml(String s, Boolean quot_too)
burst.xml.DomUtil.escapeHtml = function(s, quot_too) {
  //bu_debug("escaping s=" + s);
  if (typeof s === 'number' || typeof s === 'boolean') return '' + s;
  // escape ampersands first, since we are about to add some...
  s = s.replace(/\&/g, '&amp;');
  s = s.replace(/</g, '&lt;');
  s = s.replace(/>/g, '&gt;');
  if (quot_too) s = s.replace(/\"/g, '&quot;');
  return s;
}



/**
On platforms that only support window.setTimeout of String, not Function, simulate Function support.

@param win The window to call setTimeout on
@param func Either a String or a Function (it'll figure out what to do).
@param millis The timeout delay.
*/
//:NSFUNCTION Object setTimeout(Window win, Function func, Number millis)
burst.xml.DomUtil.setTimeout = function(win, func, millis) {
   if (bu_UA.can_setTimeout_function() || typeof func == 'string') return win.setTimeout(func, millis);
   if (typeof func != 'function') bu_throw("func is not a function: " + func);
   var str = burst.Lang.unevalFunctionCall(func, win);
   return win.setTimeout(str, millis);
}

/**
 * Function to emulate node.ownerDocument in a non-compliant environment (such as IE5.0 Win).
 */
//:NSFUNCTION Node ownerDocument(Node node)
burst.xml.DomUtil.ownerDocument = function(node) {
  if (typeof node.ownerDocument != 'undefined') return node.ownerDocument;
  while (node.parentNode) node = node.parentNode;
  return node;
}

/**
* Returns true if the document (defaults to current) is complete (finished loading).
*
* It is very important to never call <code>document.write</code> after a document
* has finished loading.
* If it is, the DOM will typically re-open and replace the entire document, 
* which may also wipe out all functions previously defined on it.
*
* For now, this function only works with an html dom.
*
* @param doc Optional, defaults to document.
*/
//:NSFUNCTION Boolean isDocumentComplete(Document doc)
burst.xml.DomUtil.isDocumentComplete = function(doc) {
   if (arguments.length < 1 || !doc) doc = document;

   // In IE, doc.readyState == 'complete' (or 'loaded') for html, and doc.readyState == 4 for xmldoc.
   if (doc.readyState && doc.readyState == 'complete') return true;

   return (typeof doc.body != 'undefined' && doc.body != null);
}




/**
* On W3-compliant systems, just returns node.attributes (which supports
* both by-number and by-name indexing).
* On IE, it returns an Array of Attr nodes (which will only support by-number).
* See http://www.xs4all.nl/~ppk/js/w3c_core.html
*/
//:NSFUNCTION Array getAttributeNodes(Node node)
burst.xml.DomUtil.getAttributeNodes = function(node) {
  if (bu_UA.bug_ie_attributes()) {
    var a = [];
    var push_att = function(attn) {
      if (!attn) return;
      if (typeof attn.nodeValue == 'undefined' || attn.nodeValue == null) return;
      if (attn.nodeValue == '') return;
      a.push(attn);
    };
    for(var k in node.attributes) {
      var attnode = node.attributes[k];
      // we may have an array of 81 potential attributes, some entries might be null.
      // also, the Node might exist but have null nodeValue.
      // we also skip the many with empty string values
      if (!attnode) continue;
      if (typeof attnode == 'object') {  // IE6
        push_att(attnode);
      }
      else { // a value, IE5
        if (typeof node.getAttributeNode != 'undefined')  // case never hit?
           push_att(node.getAttributeNode(k));
        else if (typeof node.getAttribute != 'undefined' && typeof node.getAttribute(k) == 'object')
           push_att(node.getAttribute(k)); // take advantage of bad implementation
        else
           bu_unsupported("getAttributeNodes");
      }
    }
    return a;
  }
  else {
    return node.attributes;
  }
}








/**
innerText is a convenient node member, but not part of the standard DOM .
If available (as in IE), we use the native member. 
Otherwise we implement it ourselves.

Note that IE innerText and innerHTML are members of document.documentElement, but
not of document.

In IE, innerHTML can be roundtripped (what it returns can be set).
What it provides is normalized:
- upper case tag names
- normalized attribute quotes
- attributes may be ordered
- normalized ending tags (and inserting implicit ones)
- it converts href values to absolute urls
- it converts less-than and ampersand to character references when used in text and attributes nodes

Furthermore, IE normalizes it to its own DTD, which might include changing the doctype.

In the case of innerText, "what you see is what you get". But in the case of html,
you only get concatenated text nodes. You don't get any tags. You may get less-than and
ampersand, but they are literals (not markup).

See http://www.xs4all.nl/~ppk/js/w3c_html.html for which browsers implement these.

NS6 has innerHTML but does not have innerText or outerHTML.

IE5 Mac has innerHTML but not innerText.

Opera 7, has innerHTML, innerText, outerHTML, outerText.
Note that (at least in beta) in Opera, innerHTML has escaped angle brackets, whereas in IE they are not:
http://groups.google.com/groups?selm=65sih4uf.fsf%40hotpop.com

* @todo Is there a reliable distinction between returning '' and null if for example there are no children?
* @param node The DOM Node whose innerText should be returned as a String
* @return The innerText.
*/
//:NSFUNCTION String getInnerText(Node node)
burst.xml.DomUtil.getInnerText = function(node) {
  if (!node) bu_throw("getInnerText called with bad node: " + node);
  if (typeof node.innerText != 'undefined') return node.innerText;
  // in case we are passed in document, which at least in IE5.0 does not have documentElement as a firstChild
  // opera seems to sometimes have a null (vs. undefined) documentElement
  if (typeof node.documentElement != 'undefined' && node.documentElement !== null && typeof node.documentElement.outerText != 'undefined')
    return node.documentElement.outerText;
  return bu_DomUtil_getInnerText(node);
}

// recursive
function bu_DomUtil_getInnerText(node) {
  var texts = [];
  burst.xml.DomUtil.for_children(node, function(child) {
    if (child.nodeType == Node.TEXT_NODE) texts.push(child.nodeValue);
    else if (child.nodeType == Node.ELEMENT_NODE) texts.push(bu_DomUtil_getInnerText(child));
  });
  return texts.join('');
}

/**
* Setting innerText.
* This is native on IE, but not Moz.
*
* If it is not native, it is implemented in terms of innerHTML (which is native on Moz),
* by simply escaping html (less-than, greater-than, ampersand).
*/
//:NSFUNCTION void setInnerText(Node node, String text)
burst.xml.DomUtil.setInnerText = function(node, text) {
  if(typeof node.innerText != 'undefined') {node.innerText = text; return;}
  bu_DomUtil_setInnerText(node, text);
}

function bu_DomUtil_setInnerText(node, text) {
   var html = burst.xml.DomUtil.escapeHtml(text);
   burst.xml.DomUtil.setInnerHTML(node, html);
}

/**
* innerHTML is non-standard, but is available as a gettable and settable member
* variable in Mozilla, IE, and Opera 7.
*
* If it is not native, we implement by calling getOuterHTML on all children.
*/
//:NSFUNCTION String getInnerHTML(Node node)
burst.xml.DomUtil.getInnerHTML = function(node) {
  if (typeof node.innerHTML != 'undefined') return node.innerHTML;
  // in case we are passed in document, which at least in IE5.0 does not have documentElement as a firstChild
  if (typeof node.documentElement != 'undefined' && node.documentElement !== null && typeof node.documentElement.outerHTML != 'undefined')
    return node.documentElement.outerHTML;
  return bu_DomUtil_getInnerHTML(node);
}

// implementation in terms of combining the results of getOuterHTML on all children.
function bu_DomUtil_getInnerHTML(node) {
  var texts = [];
  burst.xml.DomUtil.for_children(node, function(child) {
    texts.push(burst.xml.DomUtil.getOuterHTML(child));
  });
  return texts.join('');
}

/**
* Set innerHTML. At the moment this just sets the member variable:
* we do not have an emulation for setting innerHTML if it is not native
* (which it is on IE and Moz and Opera).
*/
//:NSFUNCTION void setInnerHTML(Node node, String html)
burst.xml.DomUtil.setInnerHTML = function(node, html) {
  if (typeof node.innerHTML != 'undefined') {node.innerHTML = html; return;}
  bu_throw("No implementation of innerHTML");
}

/**
* IE has native outerHTML, but Moz does not.
*
* If getting outerHTML is not native, we emulating by writing
* the top level ourselves, then calling <var>getInnerHTML</var> (which may
* in turn recursively call us again).
*/
//:NSFUNCTION String getOuterHTML(Node node)
burst.xml.DomUtil.getOuterHTML = function(node) {
  if (typeof node.outerHTML != 'undefined') return node.outerHTML;
  return bu_DomUtil_getOuterHTML(node);
}

function bu_DomUtil_getOuterHTML(node) {
  return burst.xml.DomUtil.serialize(node, true);
}

/**
* If setting outerHTML is not native (as it is on IE),
* we emulate on Moz using its proprietary createContextualFragment (which is also available on recent KHTML browsers).
* 
* We do not have an emulation otherwise.
*
* (The technique of providing settable outerHTML using createContextualFragment may be seen by:
* - Martin Honnen, http://groups.google.com/groups?selm=38722529.72387901%40t-online.de
* - Erik Arvidsson, http://webfx.nu/dhtml/mozInnerHTML/mozInnerHtml.html
* )
*/
//:NSFUNCTION void setOuterHTML(Node node, String html)
burst.xml.DomUtil.setOuterHTML = function(node, html) {
  if (typeof node.outerHTML != 'undefined') {node.outerHTML = html; return;}
  var doc = burst.xml.DomUtil.ownerDocument(node);
  var range = doc.createRange();
  range.setStartBefore(node);
  if (typeof range.createContextualFragment == 'undefined')
     bu_unsupported("setOuterHTML without Range.createContextualFragment");
  var fragment = range.createContextualFragment(html);
  node.parentNode.replaceChild(fragment, node);
}

/* for the record, how Martin and Erik did it
// dunno the point of outerHTMLInput
function bu_DomUtil_setOuterHTML_honnen(node, html) {
  node.outerHTMLInput = html;
  var range = document.createRange();
  range.setStartBefore(node);
  var docFrag = range.createContextualFragment(html);
  node.parentNode.replaceChild(docFrag, node);
}

// dunno the significance of using ownerDocument vs. global document
function bu_DomUtil_setOuterHTML_erik(node, html) {
   var r = node.ownerDocument.createRange();
   r.setStartBefore(node);
   var df = r.createContextualFragment(html);
   node.parentNode.replaceChild(df, node);
}
*/

/**
* insertAdjacentElement, insertAdjacentHTML and insertAdjacentText are IE-specific, but convenient.
*
* See 
* http://msdn.microsoft.com/workshop/author/dhtml/reference/methods/insertadjacentelement.asp
*
* Opera 7 implements them. Mozilla does not.
*
* The emulation for insertAdjacentHTML relies on the Mozilla non-standard createContextualFragment
* function (also implemented in recent KHTML browsers). We do not have an implementation otherwise.
*
* Our emulation for the other two functions, insertAdjacentElement and insertAdjacentText,
* does not rely on Mozilla-specific functionality.
*
* There are available implementations of insertAdjacent* by:
* - Thor Larholm, http://www.faqts.com/knowledge_base/view.phtml/aid/5756
* - Erik Arvidsson, http://webfx.nu/dhtml/ieemu/htmlmodel.html
*
* We are currently using Thor's.
*/
//:NSFUNCTION void insertAdjacentElement(Node node, String where, Node el)
burst.xml.DomUtil.insertAdjacentElement = function(where, node) {
  if (typeof node.insertAdjacentElement != 'undefined') // IE and Opera
    node.insertAdjacentElement(where, el);
  else
    bu_DomUtil_insertAdjacentElement_thor(node, where, el); // otherwise
}

function bu_DomUtil_insertAdjacentElement_thor(node, where, el) {
  switch (where){
  case 'beforeBegin':
    node.parentNode.insertBefore(el, node);
    break;
  case 'afterBegin':
    node.insertBefore(el, node.firstChild);
    break;
  case 'beforeEnd':
    node.appendChild(el);
    break;
  case 'afterEnd':
    if (node.nextSibling) 
      node.parentNode.insertBefore(el, node.nextSibling);
    else
      node.parentNode.appendChild(el);
    break;
  default:
    bu_throw("bad where '" + where + "'");
  }
}

/** 
* Calls native insertAdjacentText if available. Otherwise emulates it.
*/
//:NSFUNCTION void insertAdjacentText(Node node, String where, Node el)
burst.xml.DomUtil.insertAdjacentText = function(where, node) {
  if (typeof node.insertAdjacentText != 'undefined') 
    node.insertAdjacentText(where, el);
  else
    bu_DomUtil_insertAdjacentText_thor(node, where, el);
}

function bu_DomUtil_insertAdjacentText_thor(node, where, txt) {
  var parsedText = document.createTextNode(txt);
  burst.xml.DomUtil.insertAdjacentElement(node, where, parsedText);
}

/** 
* Calls native insertAdjacentHTML if available. Otherwise emulates it if 
* Mozilla-style createContextualFragment is available.
* Else it throws.
*/
//:NSFUNCTION void insertAdjacentHTML(Node node, String where, Node el)
burst.xml.DomUtil.insertAdjacentHTML = function(where, node) {
  if (typeof node.insertAdjacentHTML != 'undefined')  // IE and Opera
    node.insertAdjacentHTML(where, el);
  else
    bu_DomUtil_insertAdjacentHTML_thor(node, where, el);
}

function bu_DomUtil_insertAdjacentHTML_thor(node, where, html) {
  if (typeof node.ownerDocument == 'undefined' || typeof node.ownerDocument.createRange == 'undefined')
      bu_unsupported("insertAdjacentHTML without Document.createRange");
  var r = node.ownerDocument.createRange();
  r.setStartBefore(node);
  if (typeof r.createContextualFragment == 'undefined')
     bu_unsupported("insertAdjacentHTML without Range.createContextualFragment");
  var parsedHTML = r.createContextualFragment(html);
  burst.xml.DomUtil.insertAdjacentElement(node, where, html);
}



/**
* Serialize the provided node to a String. Works with a node from either an HTML or XML DOM.
*
* Note that in some browsers (such as IE5.0 PC), document has no children. You have to pass in document.documentElement,
* which is a real Node.
*
* @todo: offer control over putting \n after things like PIs, and possibly indenting. Ultimately, we'd need a "Writer" interface.
*
* @param node The node to start with. It is included in the result, not just its children.
* @param isHtml Optional. If true, then we may attempt to use innerHTML when available, for efficiency.
*/
//:NSFUNCTION String serialize(Node node, Boolean isHtml)
burst.xml.DomUtil.serialize = function(node, isHtml, depth) {
  if (typeof node == 'undefined' || !node) bu_throw("(burst.xml.DomUtil.js) bad node to serialize: " + node);

  // deal with IE5.0 and passing in document, which has no nodeType.
  if (typeof node.nodeType == 'undefined' && typeof node.documentElement != 'undefined' && node.documentElement !== null)
     return burst.xml.DomUtil.serialize(node.documentElement);

  if (arguments.length < 3) depth = 0;
  //bu_debug_Dom("(burst.xml.DomUtil.js) serialize nodeType " + node.nodeType + " isHtml=" + isHtml + " at depth " + depth + " Node.DOCUMENT_NODE=" + Node.DOCUMENT_NODE);
  var texts = [];
  switch(node.nodeType) {

  // todo: offer to call toLowerCase on node and attribute names
  case Node.ELEMENT_NODE: { // 1
    texts.push('<'); 
    texts.push(node.nodeName);
    burst.Alg.for_each(burst.xml.DomUtil.getAttributeNodes(node), function(att) {
       texts.push(' '); 
       texts.push(att.nodeName); 
       texts.push('="');
        // note that in IE, sometimes typeof nodeValue is 'boolean' (for example, att "hideFocus")
       texts.push(burst.xml.DomUtil.escapeHtml(att.nodeValue,true));
       texts.push('"');
    });
    if (node.firstChild) {
       texts.push('>');
       if (isHtml) {
         texts.push(burst.xml.DomUtil.getInnerHTML(node));
       }
       else {
        //ENVBUG: Konqueror 3.04 has "TypeError: Expression is no object. Cannot be called"
        // if we attempt to call as bu_DomUtil_serialize instead of burst.xml.DomUtil.serialize
        // because it just throws away function names, so doesn't know about bu_DomUtil_serialize
        burst.xml.DomUtil.for_children(node, function(child) {texts.push(burst.xml.DomUtil.serialize(child, isHtml, ++depth))});
       }
       texts.push('</'); texts.push(node.nodeName); texts.push('>');
    }
    // no children, self-closing
    else {
       texts.push('/>');
    }
    return texts.join('');
  }

  // case Node.ATTRIBUTE_NODE: { // 2 

  case Node.TEXT_NODE: { // 3
    // Note we assume the implementation is properly returning character references for lt, gt, ampersand.
    // This is as per http://www.w3.org/TR/2000/REC-xml-20001006#syntax
    // only CDATA, PI, or comment has literals.
    //
    // We also assume that unicode is dealt with correctly.
    // In the simplest case, we might otherwise map all chars above 127 to an escape sequence, for example
    // byte 160 to '\u00A0' or '&nbsp;'.
    // That is, the charCodeAt result might need to be expanded.
    return node.nodeValue;
  }

  case Node.CDATA_SECTION_NODE: { // 4
     // node.data vs. node.nodeValue?
    return "<![CDATA[" + node.nodeValue + "]]>"; // same as node.data
  }

  // not that we expect to see this.
  // entities are resolved before we get here, typically.
  case Node.ENTITY_REFERENCE_NODE: { // 5
    return "&" + node.nodeName + ";"
  }

  // case Node.ENTITY_NODE: { // 6 

  case Node.PROCESSING_INSTRUCTION_NODE: { // 7
    return "<?" + node.target + " " + node.data + "?>" /* + "\n"*/;
  }

  case Node.COMMENT_NODE: { // 8
   return "<!--" + node.nodeValue + "-->" /* + "\"*/;
  }

  case Node.DOCUMENT_NODE: { // 9
       if (isHtml) {
         texts.push(burst.xml.DomUtil.getInnerHTML(node));
       }
       else {
         burst.xml.DomUtil.for_children(node, function(child) {texts.push(burst.xml.DomUtil.serialize(child,isHtml,++depth))});
       }
       return texts.join('');
  }

  case Node.DOCUMENT_TYPE_NODE: { // 10
    texts.push('<!DOCTYPE ');
    texts.push(node.nodeName);
    if (node.publicId) {
      texts.push(' PUBLIC \"'); texts.push(node.publicId); texts.push('\"');
      if (node.systemId) {
        texts.push(' \"'); texts.push(node.systemId); texts.push('\"');
      }
    }
    else if (node.systemId) {
      texts.push(' SYSTEM \"'); texts.push(node.systemId); texts.push('\"');
    }
    if (node.internalSubset) {texts.push(" "); texts.push(node.internalSubset);}
    texts.push(">\n");
    return texts.join('');
  } 

  // case Node.DOCUMENT_FRAGMENT_NODE: { // 11 
  // case Node.NOTATION_NODE: { // 12 


  default:
    if (typeof node.nodeType == 'undefined') {
      bu_alert("(burst.xml.DomUtil.js) skipping node with undefined nodeType at depth " + depth);
    }
    else bu_throw("(burst.xml.DomUtil.js) unsupported node.nodeType=" + node.nodeType);
    return '';
  }
}





/**
* Set the attributes in attmap in the provided node, and call appendChild on all children.
*
* @todo: IE will blindly setAttribute('onclick', str) to a string, which doesn't work.
* @param node The node in question.
* @param attmap Optional. If specified, an associative array from name to value. 
* @param child1 Optionally any number of children (varargs). Each can be an Array of Node, or a Node.
*/
//:NSFUNCTION Node nd(Node node, Object attmap, Node child1, ...)
burst.xml.DomUtil.nd = function(node, attmap, children) {
  if (attmap) {
    for(var attname in attmap) {
      node.setAttribute(attname, attmap[attname]);
    }
  }
  if (children) {
    for(var i=2;i<arguments.length;++i) {
      var child = arguments[i];
      if (burst.Lang.isArray(child)) {
        for(var j=0;j<child.length;++j) {node.appendChild(child[j])}
      }
      else node.appendChild(child);
    };
  }
  return node;
}


/**
* Convert a hierarchical structure to a Node.
* The structure is of the form:
<pre>
  [tagname, {attname1: attval1, ...}, [childname1, ...], 'some text2', [childname3, ...], ...]
</pre>
*
* The returned Node is not attached anywhere in the document by this function.
* That would be done afterwards by the caller.
*
* If tagname is '#document-fragment', then createDocumentFragment is called instead of createElement.
* (Not that you have to use a DocumentFragment, as w3 allows appendChild on orphaned parents.)
*/
//:NSFUNCTION Element structToNode(Object st, Document doc)
burst.xml.DomUtil.structToNode = function(st, doc, depth, child_index) {
  if (!doc) doc = document;
  if (arguments.length <= 2) {depth = 0; child_index = -1;}

  bu_debug("structToNode(",st,")");
  if (typeof st === 'number') st = '' + st;
  if (typeof st === 'string') {
    return doc.createTextNode(st);
  }
  if (!burst.Lang.isArray(st)) bu_throw("(burst.xml.DomUtil.js) unexpected structure type: " + st + "\n typeof=" + (typeof st) + " st.constructor=\n" + st.constructor + "\nisArray=" + (st.constructor==Array) + " name=" + st.constructor.name + " has push=" + (typeof st.push) + "\n at depth " + depth + " and child index " + child_index);

  var tagname = st[0] || bu_throw("no tag name");
  var el;

  switch(tagname) {
  case '#document-fragment': el = doc.createDocumentFragment(); break;
  default: el = doc.createElement(tagname);
  }
  var attmap = st[1];
  if (attmap) {
    for(var attname in attmap) {
      el.setAttribute(attname, attmap[attname]);
    }
  }
  // recursive
  for(var i=2;i<st.length;++i) {
    var child = burst.xml.DomUtil.structToNode(st[i], doc, ++depth, i - 2);
    el.appendChild(child);
  }
  return el;
}


/**
* Return a String of HTML (or XML) for the element with provided name, attributes
* and children.
* For attribute values, we do value quoting and escaping; the caller is not expected to.
* We skip attributes whose value is null. We throw on attributes whose value is undefined.
*
* @todo have an XMLWriter object that supports options such as controllable indenting, doctype, etc.
*
* @param name The element name. Case is preserved.
* @param attmap Optional. If specified, an associative array from name to value. 
* @param children Optional. If specified, it is an (already serialized) String or an Array of String. Any following arguments are concatenated.
*/
//:NSFUNCTION String el(String name, Object attmap, String children, ...)
burst.xml.DomUtil.el = function(name, attmap, children) {
  if (typeof name != 'string') bu_throw('(burst.xml.DomUtil.js) bad name argument to el: ' + name);
  var indent = false;
  var a = [];
  a.push('<'); a.push(name);
  if (attmap) {
    for(var attname in attmap) {
      a.push(' '); a.push(attname); a.push('="');
bu_debug("attname -> " + attmap[attname]);
      var attval = attmap[attname];
      if (typeof attval == 'undefined') {
         var mess = '(burst.xml.DomUtil.js) attribute "' + attname + '" has bad typeof ' + (typeof attval) + 
            ' and value |' + attval + '|' +
            ' el so far is "' + a.join('') + '"';
         bu_alert(mess);
	 // bu_throw(mess);
      }
      else if (attval == null) {}
      else 
         a.push(burst.xml.DomUtil.escapeHtml(attval, true));
      a.push('"');
    }
  }
  a.push('>');
  if (children) {
    function add_child(s) {
      if (indent) {a.push("\n  "); s = s.replace(/\n/g, "\n  ");}
      a.push(s);
    }
    for(var i=2;i<arguments.length;++i) {
      var str = arguments[i];
      if (burst.Lang.isArray(str)) {
        for(var j=0;j<str.length;++j) {add_child(str[j])}
      }
      else add_child(str);
    };
  }
  a.push('</'); a.push(name); a.push('>');
  return a.join('');
}



/**
* Iterate through the node's child nodes calling the provided function with each child.
*
* Implementation note: This is implemented with firstChild and nextSibling because
* NodeList Node.childNodes isn't a true Array but a collection and isn't terribly efficient.
*
* @param node A DOM Node instance.
* @param unary_func A function taking one argument.
* @param node_type Optional. One of the Node.*_NODE constants. If specified, only nodes with that type are passed to unary_func.
*/
//:NSFUNCTION void for_children(Node node, Function unary_func, Number node_type)
burst.xml.DomUtil.for_children = function(node, unary_func, node_type) {
  //bu_alert("(burst.xml.DomUtil.js) in for_children");
  if (!node) bu_throw("no node");
  if (!unary_func) bu_throw("no unary_func");
  if (typeof unary_func != 'function') bu_throw("unary_func not a function");

  for(var child=node.firstChild;child;child=child.nextSibling) {
    //bu_alert("(burst.xml.DomUtil.js) in loop"); 
     //bu_alert("(burst.xml.DomUtil.js) at child " + child);
     //bu_alert("(burst.xml.DomUtil.js) with child.nodeType " + child.nodeType);
    if (!node_type || child.nodeType == node_type) {
       //bu_alert("about to call: " + unary_func); 
       unary_func(child);
    }
    //if (!node_type || child.nodeType == node_type) unary_func.call(null, child);
  }
}

// see http://www.xs4all.nl/~ppk/js/w3c_core.html
// NamedNodeMap Node.attributes can be used as an array (length and item)
// this calls the function on the Attr objects.
/* commented out since ppk make clear it is hopeless. we could use our burst.xml.DomUtil.getAttributes() implementation I suppose.
burst.xml.DomUtil.for_attributes = function(node, func) {
  var atts = node.attributes;
  for(var i=0;i<atts.length;++i) {func(atts[i]);}
}
*/



/**
* Find the first child of node satisfying the predicate.
* @param node
* @param unary_pred
* @param node_type Optional. One of the Node.*_NODE constants. If specified, only nodes with that type are passed to unary_pred.
* @return The satisfying child node, or null if none.
*/
//:NSFUNCTION Node find_child(Node node, Function unary_pred, Number node_type)
burst.xml.DomUtil.find_child = function(node, unary_pred, node_type) {
  if (!node) throw "no node";
  for(var child=node.firstChild;child;child=child.nextSibling) {
    //bu_debug("(burst.xml.DomUtil.js) child.nodeName=" + child.nodeName);
    if (!node_type || child.nodeType == node_type) {
      if (unary_pred(child)) return child;
    }
  }
  return null;
}

/**
* Return the first child with the specified tag name, or null.
* Note that this only iterates through immediate children (unlike DOM getElementsByTagName,
* which goes through all descendants).
* It will only return an Element child, of an Element parent.
*
* By default (cs is false), name is uppercased prior to iteration.
*
* @param node the parent node
* @param name the tag name of the desired child
* @param must_exist Optional. If true the function will throw instead of returning null.
* @param cs Optional. If true, us a case-sensitive comparison.
* @return The satisfying child node, or null if none and !must_exist.
*/
//:NSFUNCTION Node firstChildWithTagName(Node node, String name, Boolean must_exist, Boolean cs)
burst.xml.DomUtil.firstChildWithTagName = function(node, name, must_exist, cs) {
  if (!cs) name = name.toUpperCase();
  var func = function(n) {
    // Opera has some Element children with no name
    //if (!n.nodeName) {bu_alert("(burst.xml.DomUtil.js) element node with no nodeName: " + n + " nodeType=" + n.nodeType + " nodeValue=" + n.nodeValue); return false;}
    return name == n.nodeName;
  };
  var child = burst.xml.DomUtil.find_child(node, func, Node.ELEMENT_NODE);
  if (!child && must_exist) bu_throw("node has no '" + name + "' child");
  return child;
}

/**
* Like firstChildWithTagName, but returns the nth (zero-based) child element with the specified tag name.
*/
//:NSFUNCTION Node nthChildWithTagName(Node node, Number n, String name, Boolean must_exist, Boolean cs)
burst.xml.DomUtil.nthChildWithTagName = function(node, n, name, must_exist, cs) {
   if (!cs) name = name.toUpperCase();
   var seen = 0;
   for(var child=node.firstChild; child; child=child.nextSibling) {
     //bu_debug("testing child ", child);
     if (child.nodeType == Node.ELEMENT_NODE && name == child.nodeName) {
       if (n == seen) break;
       seen++; 
     }
   }
   if (!child && must_exist) bu_throw("found " + seen + " element children with tag name " + name + " but not " + n);
   return child;
}

/* alternative implementation of firstChildWithTagName:
  var children = node.getElementsByTagName(name);
  if (!children || children.length == 0) {
     if (must_exist) throw "node has no '" + name + "' child";
     return; //undefined
  }
  return children[0];
*/


/**
* Returns a snapshot Array, unlike getElementsByTagName(name).
* Has special treatment for name '*', to return all Element children.
*/
//:NSFUNCTION Array getChildrenWithTagName(Node node, String name, Boolean cs)
burst.xml.DomUtil.getChildrenWithTagName = function(node, name, cs) {
   if (!cs) name = name.toUpperCase();
   var children = [];
   for(var child=node.firstChild; child; child=child.nextSibling) {
     bu_debug("(burst.xml.DomUtil.js) getChildrenWithTagName at child with name " + child.nodeName);
     if (child.nodeType == Node.ELEMENT_NODE && (name == '*' || name == child.nodeName)) {
       children.push(child);
     }
   }
   return children;
}

/**
A convenience function. If is_snapshot, then calls burst.xml.DomUtil.getChildrenWithTagName.
Otherwise uses Node.getElementsByTagName (which differs semantically in that it
returns all descendants not just immediate children).
*/
//:NSFUNCTION Array childrenNamed(Node node, String name, Boolean is_snapshot)
burst.xml.DomUtil.childrenNamed = function(node, name, is_snapshot) {
  return is_snapshot ?
    burst.xml.DomUtil.getChildrenWithTagName(node, name) :
    node.getElementsByTagName(name);
}

/**
* Perform what Node.getAttribute should do, but doesn't always.
*
* PPK has much discussion of this, for example at: http://www.xs4all.nl/~ppk/js/w3c_core.html
* He suggests that Node.getAttribute in IE5 and IE6 is not always correct,
* saying the function is only trustworthy in Moz, Safari, IE-mac.
* However, I think it does actually work in IE5/IE6 except for the special cases noted below.
* In fact, see PPK's own tests at http://www.xs4all.nl/~ppk/js/w3c/attributes.html
*
* 
* <h3>Attributes that are ECMAScript reserved names</h3>
* Apparently getAttribute('class') doesn't work in IE (must use getAttribute('className')).
* Apparently opera does not support getAttribute('className').
*
* <br><b>Disposition</b>: Resolved. caller should pass in 'class'. We try both 'class' and 'className'.
* For set, we do node.className = 'whatever';
*
* 
* <h3>Attribute name case-sensitivity</h3>
* In IE, Node.getAttribute takes a second argument that is 0 for case-insensitive, and 1 for case-insensitive.
* The default is 0, for case-insensitive.
* However, reputedly in practice IE5+ still acts case-sensitively for special attributes (e.g.,
* requiring 'bgColor' or 'colSpan' vs. allowing 'bgcolor' or 'colspan'), see:
* http://www.pxl8.com/setAttribute_test.html
*
* The W3 DOM has no such second argument.
* Moz and IEMac are case-insensitive at least for HTML.
*
* <br><b>Disposition</b>: We do nothing.
*
*
* <h3>Non-Standard Attributes</h3>
* Apparently, only IE preserves non-standard attributes in "expando" properties.
* But for getAttribute and setAttribute, non-standard attributes seem to be generally supported.
*
* <br><b>Disposition</b>: We do nothing.
*
*
* <h3>Attributes for event handlers</h3>
* In IE, attributes for event handlers (such as 'onclick') may be queried with getAttribute,
* but not set with setAttribute, but will allow a set with "expando" properties.
* If they are queried, IE returns a Function object, not a String.
* To set in IE, you have to say <code>node[attname] = new Function(attval)</code>
* instead of node.setAttribute(attname) = attval.
* It would be safest to use lowercase event handler names (like 'onclick', not 'onClick' or 'ONCLICK').
* Alternatively, <code>node.attachEvent(attname, new Function(attval))</code>.
*
* Moz returns a String from getAttribute, and supports setAttribute with String too.
*
*
* <h3>The 'style' attribute</h3>
* For the 'style' attribute, IE6 getAttribute('style') returns an object.
* Moz returns a String for getAttribute('style').
*
* Note that object accessor "node.style" is an Object in IE.
* Note however that it has no value unless the document has an inline style attribute
* (from the document or from setAttribute). 
*
* To set a 'style' attribute, node.setAttribute(style, 'top: 5px') works in Moz,
* but not in IE, where it is necessary to do something like node.style.top = '5px'
* or node.style.cssText = 'top: 5px' or node.style.setProperty('top', '5px').
*
* (Btw, 'float' may appear in css text, but the name for setProperty and for rule attribute
* is 'cssFloat'. See http://www.w3.org/TR/DOM-Level-2-Style/css.html .)
*
* The W3 standard way to get the effective style on a element is 
* document.defaultView.getComputedStyle(node,null).
* You can get particular properties with getPropertyValue('top')
* This is implemented in Mozilla 1.2 but not IE6, which instead has "node.currentStyle".
*
* Note that the choice between setting style by:
* - changing the class (or other attribute affecting CSS selectors)
* - setting style on the node
* - changing the CSS rules
* Is beyond the scope here. For an intro, see:
* - http://www.xs4all.nl/~ppk/js/index.html?getstyles.html
* - http://www.xs4all.nl/~ppk/js/w3c_css.html
* - http://www.xs4all.nl/~ppk/js/changess.html (particularly for changing css style rules)
* - http://developer.apple.com/internet/javascript/styles.html 
*
* <br><b>Disposition</b>: When getting, if the value is an object (and not null or string),
* we return value.cssText. When setting, we similarly set cssText instead.
*
*
* <h3>Missing attributes</h3>
* The W3 standard says that the empty string '' should be returns for attribute that aren't set.
* Allegedly, in NN6 getAttribute of a non-existent attribute returns '' instead of null or undefined.
* Moz and NN7 instead return something that tests as false.
* See http://www.pxl8.com/setAttribute_test.html
*
* <br><b>Disposition</b>: We do nothing.
*
*
* <h3>White space in values</h3>
* Apparent IE does not preserve white space in attribute values: http://www.zvon.org/xxl/DOM2reference/Output/examples/Element_getAttribute.html
*
* <br><b>Disposition</b>: We do nothing.
*
*
* <h3>XML namespaces and prefixes</h3>
* We don't do anything about xml namespaces here; we are dealing with getAttribute here,
* not level 2 getAttributeNS. 
*
*
* @todo A similar function with xmlns support. 
* @todo Determine whether attribute global namespace is done with getAttributeNS('', n) or getAttributeNS(null,n)
*
* @param node The node which possibly contains the attribute in question.
* @param name The name of the attribute.
* @return The String value of the attribute. null or undefined is returned if none.
*/
//:NSFUNCTION String getAttribute(Element node, String name)
burst.xml.DomUtil.getAttribute = function(node, name) {
  // IE will throw a runtime error if you attempt to getAttribute on a Text object
  if (node.nodeType != Node.ELEMENT_NODE) return BU_UNDEFINED;
  if (!node || !node.getAttribute) throw Error("(DomUtil.js) attempt to call getAttribute(" + node + ", " + name + ")"); 
  var val = node.getAttribute(name);
  if (name == 'class') {return val ? val : burst.xml.DomUtil.getAttribute(node, 'className');}
  if (name == 'for') {return val ? val : burst.xml.DomUtil.getAttribute(node, 'htmlFor');}
  if (name == 'style') {
    if (!val || typeof val == 'string') return val;
    return val.cssText;
  }
  // IE5.0PC?
  if (val && typeof val == 'object') {alert("getAttribute returned object"); return val.value;}
  return val;
}

/**
* Correct for any problems with native Node.setAttribute.
* See burst.xml.DomUtil.getAttribute for more information.
*/
//:NSFUNCTION void setAttribute(Element node, String name, String val)
burst.xml.DomUtil.setAttribute = function(node, name, val) {
  if (name == 'class') {node.className = val; return;}
  if (name == 'style' && bu_UA.bug_set_style_attribute()) {
     node.style.cssText = val;
     return;
  }
  node.setAttribute(name, val);
}


/* Unused implementation
function bu_DomUtil_getAttribute(node, name) {
  if (node.getAttributeNode) {
    var v = node.getAttributeNode(name);
    // IE mac returns value not object
    if (typeof v == 'string') return v;
    return v ? v.value : null;
  }
  var val = node.attributes[name];
  if (typeof val == 'string') return val;
  return val ? val.value : null;
}
*/


//:NSEND

bu_loaded('burst.xml.DomUtil', ['burst.BurstError', 'burst.Lang', 'burst.web.UserAgent', 'burst.URI', 'burst.Alg']);
