/**
* @file HtmlBox.js
*
* Defines burst.xml.HtmlBox which contains static functions for manipulating html size and position.
*
* @author Copyright 2003 Mark D. Anderson (mda@discerning.com)
* @author Licensed under the Academic Free License 1.2 http://www.opensource.org/licenses/academic.php
*/

//=java package burst.xml;

/**
* Scoping class holding static functions for manipulating html element size and position.
*/
//:NSBEGIN HtmlBox

burst.xml.HtmlBox = {};
var BU_HtmlBox = burst.xml.HtmlBox;

/*
Determining window size.

See:
   http://www.xs4all.nl/~ppk/js/winprop.html
   http://www.xs4all.nl/~ppk/js/doctypes.html
   http://www.evolt.org/article/document_body_doctype_switching_and_more/17/30655/index.html

IE/BackCompat:
   amount scrolled                     = body.scrollLeft/scrollTop
   viewport excluding scrollbar        = body.clientWidth/clientHeight
   viewport including scrollbar        = de.scrollWidth/scrollHeight
   canvas width/height                 = body.scrollWidth/scrollHeight

IE/CSS1Compat:
   amount scrolled                     = de.scrollLeft/scrollTop
   viewport excluding scrollbar        = de.clientWidth/clientHeight
   viewport including scrollbar        = de.offsetWidth/offsetHeight
   canvas width/height                 = de.scrollWidth/scrollHeight (body.scrollWidth/scrollHeight is smaller)

Mozilla:
   amount scrolled                     = body.scrollLeft/scrollTop = window.pageXOffset/pageYOffset
   viewport excluding scrollbar        = body.clientWidth/Height
   viewport including scrollbar        = window.innerWidth/innerHeight
   canvas width/height                 = body.scrollWidth/scrollHeight = de.scrollWidth/scrollHeight

Opera7:
   amount scrolled                     = body.scrollLeft/scrollTop = de.scrollLeft/scrollTop = window.pageXOffset/pageYOffset
   viewport width excluding scrollbar  = {body,de}.{client,scroll}Width
   viewport height excluding scrollbar = body.clientHeight
   viewport including scrollbar        = window.innerWidth/innerHeight
   canvas height                       = de.clientHeight = de.scrollHeight
   canvas width                        = ?
   garbage?                            = body.scrollHeight

Safari:
   amount scrolled                     = body.scrollLeft/scrollTop
   viewport width excluding scrollbar  = window.innerWidth=de.{clientWidth,scrollWidth}
   viewport height excluding scrollbar = window.innerHeight
   viewport including scrollbar        = ?
   canvas width/height                 = body.scrollWidth/Height
   canvas height                       = de.{clientHeight,scrollHeight}

TODO: can the document canvas be smaller than the window viewport?
Apparently in IE6/CSS1Compat, the HTML element always has size >= the viewport.
*/

/**
pixel width of entire document, regardless of scrolling. Unsupported on Opera7 (returns 0).
*/
//:NSFUNCTION Number getDocumentWidth(DomDocument doc)
BU_HtmlBox.getDocumentWidth = function(doc) {
   if (!doc) doc = document;
   return bu_UA.isIE ? burst.xml.HtmlUtil.getDocumentElementIE(doc).scrollWidth : (bu_UA.isOpera ? 0 : doc.body.scrollWidth);
}

// pixel height of entire document, regardless of scrolling
BU_HtmlBox.getDocumentHeight = function(doc) {
   if (!doc) doc = document;
   return bu_UA.isIE ? burst.xml.HtmlUtil.getDocumentElementIE(doc).scrollHeight : (bu_UA.isOpera ? doc.documentElement.scrollHeight : doc.body.scrollHeight);
}

// offset within canvas of left-most pixel that can be seen in viewport
BU_HtmlBox.getDocumentScrollLeft = function(doc) {
   if (!doc) doc = document;
   return bu_UA.isIE ? burst.xml.HtmlUtil.getDocumentElementIE(doc).scrollLeft : doc.body.scrollLeft;
}

// offset within canvas of top-most pixel that can be seen in viewport
BU_HtmlBox.getDocumentScrollTop = function(doc) {
   if (!doc) doc = document;
   return bu_UA.isIE ? burst.xml.HtmlUtil.getDocumentElementIE(doc).scrollTop : doc.body.scrollTop;
}

// pixel width of viewport excluding scrollbar
BU_HtmlBox.getDocumentViewWidth = function(doc) {
   if (!doc) doc = document;
   return bu_UA.isIE ? burst.xml.HtmlUtil.getDocumentElementIE(doc).clientWidth : (bu_UA.isKHTML ? doc.documentElement.clientWidth : doc.body.clientWidth);
}

// pixel width of viewport excluding scrollbar
// on Safari, only works on current window
BU_HtmlBox.getDocumentViewHeight = function(doc) {
   if (!doc) doc = document;
   if (bu_UA.isKHTML) return doc === document ? window.innerHeight : bu_throw("unimplemented");
   return bu_UA.isIE ? burst.xml.HtmlUtil.getDocumentElementIE(doc).clientHeight : doc.body.clientWidth;
}

// pixel width of viewport including scrollbar
// unsupported on Safari (returns width excluding scrollbar, if any).
BU_HtmlBox.getWindowWidth = function(win) {
   if (!win) win = window;
   //if (bu_UA.isKHTML) return 0;
   return bu_UA.isIE ? (bu_UA.isIEBox ? win.document.documentElement.scrollWidth : win.document.documentElement.offsetWidth) :
   win.innerWidth;
}

// pixel width of viewport including scrollbar.
// unsupported on Safari (returns height excluding scrollbar, if any).
BU_HtmlBox.getWindowHeight = function(win) {
   if (!win) win = window;
   //if (bu_UA.isKHTML) return 0;
   return bu_UA.isIE ? (bu_UA.isIEBox ? win.document.documentElement.scrollHeight : win.document.documentElement.offsetHeight) :
   win.innerHeight;
}

/*
Get screen position.
IE:
   window.screenLeft/screenTop is offset from screen top-left to document top-left (first potential content pixel).
Mozilla, Safari:
   window.screenX/screenY is offset from screen top-left to top-left of application OS-level window for browser. 
   (Safari also has window.screenLeft/screenTop, but is is the same thing)
*/


/*
Determining element location and size.

Note that the issue here at hand is not what the element location and size *should* be
according to standards, but determining from script what the browser actually did.

Readings:

   http://www.w3.org/TR/REC-CSS2/box.html
   http://www.w3.org/TR/REC-CSS2/visudet.html
   http://msdn.microsoft.com/workshop/author/om/measuring.asp
   http://msdn.microsoft.com/library/en-us/dnie60/html/cssenhancements.asp
   http://msdn.microsoft.com/workshop/author/css/overview/measurementandlocation.asp
   http://msdn.microsoft.com/workshop/author/dhtml/reference/properties/offsetwidth.asp
   http://www.xs4all.nl/~ppk/js/findpos.html
   http://www.xs4all.nl/~ppk/js/doctypes.html
   http://web.mit.edu/bzbarsky/www/testcases/getComputedStyle/ (mozilla only)
   http://www.webreference.com/dhtml/diner/realpos4/
   http://www.washington.edu/webinfo/case/flyout/flyout-ann.html

There are a number of properties that give some information about position and size.

The basic challenge is that it is not possible to determine position and size
using solely examination of style.
  - IE has node.currentStyle, but it does not convert values to pixels
    (so you may find strings such as '3', '17em', 'auto', 'medium', and so on).
  - Opera7 and Safari do not support currentStyle nor getComputedStyle
  - Mozilla supports getComputedStyle, but 'left' and 'top' are not among the
    special attributes that it calculates (it does calculate 'width' and 'height')

All browsers seem to support offset* node properties, although
they are not dictated by a standard:

  offsetLeft    left offset relative to node.offsetParent (see calculation below)
  offsetTop     top offset relative to node.offsetParent (see calculation below)
  offsetWidth   the border box width.
  offsetHeight  the border box height.
These are numbers, in pixels. They are readonly.

There are also these special readonly properties on the node (not node.style):
  clientLeft     same as border-left-width
  clientTop      same as border-top-width
  clientWidth    offsetWidth - border left width - border right width - scroll bar width (if any)
  clientHeight   offsetWidth - border top width - border bottom width - scroll bar width (if any)
IE and Opera support all four; Mozilla and Safari support only clientWidth/clientHeight:
  http://bugzilla.mozilla.org/show_bug.cgi?id=111207
  http://bugzilla.mozilla.org/show_bug.cgi?id=159365

The formula above for clientWidth/clientHeight does not apply to BODY and HTML, just other elements,
such as DIVs.

If there are scroll bars, clientWidth/clientHeight measures the visible space inside them.
The "padding box" resides within the "viewport", so that any configured padding is scrolled.
Borders go outside of scrollbars.
If there is a "width" or "height" specified in style, and scroll bars are necessary,
the space for the scroll bars is "taken" from visible area available to the content.
This is so that the border box edge can equal the value of (specified style + padding + border width).

IE has some non-standard properties in that are explicitly numbers, in units of pixels.
So even though IE does not convert style values for things like 'left', there is a way in some
cases to retrieve the calculated locations. These mappings are:
  left   pixelLeft 
  top    pixelTop
  width  pixelWidth
  height pixelHeight
  bottom pixelBottom
  right  pixelRight
These values are readwrite. Note that these are present only in style, not in currentStyle.

In the CSS standard, each element has four boxes associated with it: 
content box, padding box, border box, and margin box.
Generally speaking, each of these is contained in the next, although this is not the case
if for example there are negative margin values.
Margins are always transparent. Border style determines the border; the "background" style determines
the style for padding and content.
The content box is the one that clipping and overflow apply to.

According to the standard, the 'width' and 'height' CSS style attributes are supposed
to be interpreted as the width and height of the "content box".
However, IE5, and IE6 without a doctype, interpret them as the width and height of the "border box".
This is called the "IE box model". When IE6 is given a page with a doctype,
then document.compatMode == 'CSS1Compat', and it uses the same "CSS box model" that other browsers do.
The difference is in whether the border and padding widths are included in the element "width".

This is how offsetWidth/offsetHeight may be interpreted:

IE/BackCompat:
   offsetWidth = border box width = style width 
   (except that if the element contains nested children that would exceed the box due to position and
   width, its offsetWidth will be larger).
    
IE/CSS1Compat and non-IE:
   offsetWidth = border box width = style width (content width) + padding + borders

When there is scrolling, offsetWidth and offsetHeight are the border box of the viewport. 


This is how offsetLeft/offsetTop are interpreted:

IE/BackCompat: 
    offsetLeft/offsetTop is the distance between the "border box" edge of the node, 
    and the "padding box" edge of its offsetParent.
    That means n.offsetLeft = n.offsetParent.currentStyle.paddingLeft + 
                              n.currentStyle.marginLeft +
                              n.currentStyle.left
    To get to the "border box" edge of the offsetParent, it is necessary to also add in 
    offsetParent.clientLeft (which is its border left width); similarly for clientTop.
    The top-level offsetParent is BODY, in BackCompat.
    For nested elements with position 'static', the offsetParent is the parent element.
    Unless otherwise specified, the BODY element has a marginLeft of 10px and a marginTop of 15px.
    This appears to be incorporated into the offsetLeft/offsetTop of children, so it is
    not explicitly needed in the calculation.
    The BODY element also has a clientLeft = clientTop = 2.
    This comes from the default border of "medium none" (the style of "none" means it is invisible).
    So to calculate document-relative coordinates of the top-level corner of the border box:
      canvasX = n.offsetLeft + n.offsetParent.offsetLeft + n.offsetParent.clientLeft + ...
      canvasY = n.offsetTop + n.offsetParent.offsetTop + n.offsetParent.clientTop + ... 
    Thus typically a first top-level div with no margin would have this for top-level of border box:
      canvasX = 10 + 0 + 2 = 12
      canvasY = 15 + 0 + 2 = 17
    If that same first top-level div had a 1px border, then the coordinate of the most top-level pixel
    in the padding border is [13,18].

IE/CSS1Compat:
    When position is 'relative' or 'absolute', the offsetParent is HTML, not BODY.
    When position is 'static', immediate or ancestor offsetParent is BODY.
    HTML has clientLeft=clientTop=2 and offsetLeft=offsetTop=0
    BODY has clientLeft=clientTop=0 and offsetleft=10, offsetTop=15.
    Thus in the position static case, the (10px,15px) margin is found in the
    offsetParent.offsetLeft and offsetParent.offsetTop,
    versus being incorporated in the offsetLeft/offsetTop values of the top-level children,
    as is the case in the BackCompat case.
    From currentStyle, both BODY and HTML are reported as having a 'medium' border,
    but there are of course not actually two different borders for layout.
    Nevertheless, it still appears necessary to assume the 2px border to get a correct value.
    So we have the same formula as IE/BackCompat for position 'relative' and 'absolute'.
    For position 'static' we have:
      canvasX = n.offsetLeft + n.offsetParent.offsetLeft +  ... + 2
      canvasY = n.offsetTop + n.offsetParent.offsetTop + ... + 2

Mozilla:
   offsetLeft/offsetTop is the distance between the "border box" edges of it and its offsetParent.
   That means n.offsetLeft = n.offsetParent.currentStyle.paddingLeft + 
                              n.currentStyle.marginLeft +
                              n.currentStyle.left
   It appears that the offsetParent is always BODY in all cases (no chaining).
   It has offsetLeft=offsetTop=0, and no clientLeft/clientTop.
   By default, Mozilla BODY has a marginLeft = marginTop = 8px, which is incorporated
   into the child node.offsetLeft calculation.
   Mozilla BODY has 0px border width.
   So in summary, the document-relative top-level corner pixel of the border box is:
      canvasX = n.offsetLeft
      canvasY = n.offsetTop

Opera7:
   Like Mozilla, it appears that offsetParent is always BODY, which always has
   offsetLeft=offsetTop=0.
   Like Mozilla, the default BODY margin is (8px,8px).
   Opera does have clientLeft/clientTop as border width (like IE), but they
   are 0 for the BODY element.
   So in summary, the document-relative top-level corner pixel of the border box is:
     canvasX = n.offsetLeft
     canvasY = n.offsetTop

Safari:
   Like Mozilla, offsetParent is always BODY. It has no clientLeft/clientTop.
   It has a 8px default BODY margin.
   Like IE/CSS1Compat, Safari incorporates the margin into the offsetParent.offsetLeft value, rather
   than inside the offsetLeft values of its top-level children.
   Thus:
     canvasX = n.offsetLeft + n.offsetParent.offsetLeft
     canvasY = n.offsetTop + n.offsetParent.offsetTop

All:
    For position 'absolute', canvasX = node style margin + node style left = border box edge
    This is consistent with offsetLeft being border box edge, and style left being margin box edge.

According to the standard, the interpretation of the 'left' and 'top' CSS style attributes
differs according to the 'position' value:
- 'static', left and top are ignored.
- 'relative', left and top are relative to where it would be in static
(which is to the content edges of the first block ancestor).
- 'absolute', left and top are relative to the padding edges
of the first non-static positioned block ancestor, if any, and otherwise are relative
to the content edge of the root element.
- 'fixed', like 'absolute', but viewport-relative

In all cases, style 'left' and 'top' indicate the distance to the "margin box" edge of the node
(it is where that distance measurment starts that varies.)

Notice that style 'left' and 'top' do *not* lead to the pixel where style 'width' and 'height' begin.
The 'left' and 'top' concern the "margin box", while 'width' and 'height' concern the content
box (or the border box in the IE box model).
  
TODO:
    Deal with table cellspacing (adds to td margin?) and cellpadding (adds to td padding?).
      cellspacing and cellpadding not deprecated: http://tom.me.uk/html-to-css/cellspacing.html
      cellpadding -> td padding
      http://groups.google.com/groups?hl=en&lr=&ie=UTF-8&oe=UTF-8&safe=off&threadm=3cac9203.34716199%40news.surfeu.fi&rnum=9
      cellspacing html attribute is like CSS border-spacing (but border-spacing not supported by IE?)
         http://www.w3.org/TR/REC-html40/struct/tables.html#adef-cellspacing
	 it should also do the spacing between left of leftmost column, etc.
         http://www.w3.org/TR/CSS2/tables.html#q1
	 Apparently Opera and Mozilla do border-spacing that way.
     td ignores margin (standard and actual):
             http://groups.google.com/groups?hl=en&lr=&ie=UTF-8&oe=UTF-8&safe=off&threadm=eric-CCABDC.14113911062002%40forums.macromedia.com&rnum=12
             http://groups.google.com/groups?hl=en&lr=&ie=UTF-8&oe=UTF-8&safe=off&threadm=slrnbahvqh.nqo.Salagir%40choam.efrei.fr&rnum=10
      vertical margins collapse and horiz don't ?
      And see "border-collapse: collapse"
      http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnwdn/html/ch10_tables.asp
    What if an ancestor is scrolled?
    How does offsetWidth change if child content won't fit?
    Anything else PPK mentions at http://www.xs4all.nl/~ppk/js/findpos.html
*/

//    http://groups.google.com/groups?hl=en&lr=&ie=UTF-8&oe=UTF-8&safe=off&threadm=87isscefdx.fsf%40dinopsis.dur.ac.uk&rnum=3
//    table {
//        border-collapse: collapse; /* tells it to use collapsed border model on the cells within it. */
//        border-spacing: 0; /* the above should make this unnecessary, but it doesn't hurt to add it in case of browser bugs */
//        margin: 0; /* stops the table itself having a margin. */
//        padding: 0; /* stops there being padding between the table and the table rows */
//     }td {
//        padding: 0; /* effectively sets cellpadding="0" */ }


/**
* Document-relative coordinate of border edge. Returns a number, in pixels.
*/
//:NSFUNCTION Number getBorderBoxLeft(Node el)
BU_HtmlBox.getBorderBoxLeft = function(el) {
    var x = 0;
    var lastel = el;
    while(el) {
	x += el.offsetLeft || 0; 
	lastel = el;
	el = el.offsetParent; 
	if (el && bu_UA.isIE) x += el.clientLeft;
    }
    // add 2 for CSS1Compat when statically positioned, from HTML clientLeft
    // which is missing when BODY is last.
    if (bu_UA.isIE && !bu_UA.isIEBox && lastel.tagName == 'BODY') {
	//var pos = burst.xml.HtmlUtil.getComputedStyle(el, 'position');
	//if (!pos || pos == 'static') 
	x += 2;
    }
    return x;
}

//:NSFUNCTION Number getBorderBoxTop(Node el)
BU_HtmlBox.getBorderBoxTop = function(el) {
    var y = 0; 
    var lastel = el;
    while(el) {
	y += el.offsetTop || 0; 
	lastel = el;
	el = el.offsetParent;
	if (el && bu_UA.isIE) y += el.clientTop;
    }
    // add 2 for CSS1Compat when statically positioned, from HTML clientTop 
    // which is missing when BODY is last.
    if (bu_UA.isIE && !bu_UA.isIEBox && lastel.tagName == 'BODY') {
	//var pos = burst.xml.HtmlUtil.getComputedStyle(el, 'position');
	//if (!pos || pos == 'static') 
	y += 2;
    }
    return y;
}

/**
* Width of the border box (a number in pixels).
*/
//:NSFUNCTION Number getBorderBoxWidth(Node el)
BU_HtmlBox.getBorderBoxWidth = function(el) {
   return el.offsetWidth;		       
}

/**
* Height of the border box (a number in pixels).
*/
//:NSFUNCTION Number getBorderBoxHeight(Node el)
BU_HtmlBox.getBorderBoxHeight = function(el) {
   return el.offsetWidth;	       
}

BU_HtmlBox.getBorderBoxRect = function(el) {
    return [BU_HtmlBox.getBorderBoxLeft(el), BU_HtmlBox.getBorderBoxTop(el),
	    BU_HtmlBox.getBorderBoxWidth(el), BU_HtmlBox.getBorderBoxHeight(el)];
}

/**
Width in pixels of the left, top, right, or bottom border.

On IE, Mozilla, and Opera, getBorderRightWidth and getBorderBottomWidth include the width of any scrollbar.

Platforms:
Works on IE, Opera, and Mozilla. Not on Safari (returns 0 if not specified via inline style; fails if not in px units.)
Note that in IE, currentStyle.borderLeftWidth could be 'thin', '0.5cm', etc., but we use el.clientLeft instead.
*/
//:NSFUNCTION Number getBorderLeftWidth(Node el)
BU_HtmlBox.getBorderLeftWidth = function(el) {
    return (bu_UA.isIE || bu_UA.isOpera) ? el.clientLeft : burst.xml.HtmlUtil.getComputedStylePixels(el, 'border-left-width');
}

//:NSFUNCTION Number getBorderTopWidth(Node el)
BU_HtmlBox.getBorderTopWidth = function(el) {
    return (bu_UA.isIE || bu_UA.isOpera) ? el.clientTop : burst.xml.HtmlUtil.getComputedStylePixels(el, 'border-top-width');
}

//:NSFUNCTION Number getBorderRightWidth(Node el)
BU_HtmlBox.getBorderRightWidth = function(el) {
    if (bu_UA.isIE || bu_UA.isOpera) 
	return el.offsetWidth - el.clientWidth - el.clientLeft;
    else
       return burst.xml.HtmlUtil.getComputedStylePixels(el, 'border-right-width');
}

//:NSFUNCTION Number getBorderBottomWidth(Node el)
BU_HtmlBox.getBorderBottomWidth = function(el) {
    if (bu_UA.isIE || bu_UA.isOpera) 
	return el.offsetHeight - el.clientHeight - el.clientTop;
    else
       return burst.xml.HtmlUtil.getComputedStylePixels(el, 'border-bottom-width');
}

/**
Width in pixels of the left, top, right, or bottom padding.
Platforms:
Returns 0 on Safari and Opera if not specified via inline style. 
Fails on IE/Safari/Opera if not in px units.
 */
//:NSFUNCTION Number getPaddingLeftWidth(Node el)
BU_HtmlBox.getPaddingLeftWidth = function(el) {
    var pl = burst.xml.HtmlUtil.getComputedStylePixels(el, 'padding-left');
    return pl;
}

//:NSFUNCTION Number getPaddingTopWidth(Node el)
BU_HtmlBox.getPaddingTopWidth = function(el) {
    var pt = burst.xml.HtmlUtil.getComputedStylePixels(el, 'padding-top');
    return pt;
}

//:NSFUNCTION Number getPaddingRightWidth(Node el)
BU_HtmlBox.getPaddingRightWidth = function(el) {
    var pr = burst.xml.HtmlUtil.getComputedStylePixels(el, 'padding-right');
    return pr;
}

//:NSFUNCTION Number getPaddingBottomWidth(Node el)
BU_HtmlBox.getPaddingBottomWidth = function(el) {
    var pb = burst.xml.HtmlUtil.getComputedStylePixels(el, 'padding-bottom');
    return pb;
}

/**
If the element has no scrolling, or if it has scrolling but has not scrolled anywhere,
the document-relative coordinate of the padding box edge.

Platforms: see getBorderLeftWidth.
*/
//:NSFUNCTION Number getPaddingBoxLeft(Node el)
BU_HtmlBox.getPaddingBoxLeft = function(el) {
    return BU_HtmlBox.getBorderBoxLeft(el) + BU_HtmlBox.getBorderLeftWidth(el);
}

//:NSFUNCTION Number getPaddingBoxTop(Node el)
BU_HtmlBox.getPaddingBoxTop = function(el) {
    return BU_HtmlBox.getBorderBoxTop(el) + BU_HtmlBox.getBorderTopWidth(el);
}

/**
If the element has no scrolling, the pixel width of the padding box.
If the element has scrolling, the pixel width of the visible area inside the viewport.

Platforms: see getBorderLeftWidth.
*/
//:NSFUNCTION Number getPaddingBoxWidth(Node el)
BU_HtmlBox.getPaddingBoxWidth = function(el) {
    if (bu_UA.isIE || bu_UA.isOpera) return el.clientWidth; // faster
    return BU_HtmlBox.getBorderBoxWidth(el) -
    BU_HtmlBox.getBorderLeftWidth(el) -
    BU_HtmlBox.getBorderRightWidth(el);
}

//:NSFUNCTION Number getPaddingBoxHeight(Node el)
BU_HtmlBox.getPaddingBoxHeight = function(el) {
    if (bu_UA.isIE || bu_UA.isOpera) return el.clientHeight; // faster
    return BU_HtmlBox.getBorderBoxHeight(el) -
    BU_HtmlBox.getBorderTopWidth(el) -
    BU_HtmlBox.getBorderBottomWidth(el);
}

/**
If the element has no scrolling, or if it has scrolling but has not scrolled anywhere,
the document-relative coordinate of the content box edge.

Platforms: see getPaddingLeftWidth.
*/
//:NSFUNCTION Number getContentBoxLeft(Node el)
BU_HtmlBox.getContentBoxLeft = function(el) {
    var padleft = BU_HtmlBox.getPaddingBoxLeft(el);
    var padwidth = BU_HtmlBox.getPaddingLeftWidth(el);
    return padleft + padwidth;
}

//:NSFUNCTION Number getContentBoxTop(Node el)
BU_HtmlBox.getContentBoxTop = function(el) {
    var padtop = BU_HtmlBox.getPaddingBoxTop(el);
    var padwidth = BU_HtmlBox.getPaddingTopWidth(el);
    return padtop + padwidth;
}


/**
* If the element has no scrolling, the pixel width of the content box.
* If the element has scrolling, this isn't very useful (it isn't even the visible content area, since
* the right and bottom padding may be hidden).
*
* Implementation note: 
* If getComputedStyle/currentStyle is available, we can use that.
* Otherwise, if style.width is set, we can trust that at least for position 'absolute' and CSS box model.
* (When not 'absolute', layout may not always choose to obey the style.)
*/
//:NSFUNCTION Number getContentBoxWidth(Node el)
BU_HtmlBox.getContentBoxWidth = function(el) {
    // gecko has getComputedStyle, and 'width' and 'height' actually are calculated values, not original strings
    if (bu_UA.isGecko) {
	return burst.xml.HtmlUtil.getComputedStylePixels(el, 'width');
    }

    // IE can figure out from offsetWidth because it has currentStyle to determine padding widths
    var cw = BU_HtmlBox.getPaddingBoxWidth(el) - 
    BU_HtmlBox.getPaddingLeftWidth(el) -
    BU_HtmlBox.getPaddingRightWidth(el);

    if (bu_UA.isIE) return cw;

    // this browser doesn't support computed style. try for style.width

    if (bu_UA.isIE) { // never run
	if (!bu_UA.isIEBox) {
	    var pw = el.style.pixelWidth;
	    if (pw > 0) return pw;
	}
    }

    if (!bu_UA.isIEBox) {
	var pos = burst.xml.HtmlUtil.getComputedStyle(el, 'position');
	var try_style = (pos && pos == 'absolute');
	if (try_style) {
	    var sw = burst.xml.HtmlUtil.getComputedStylePixels(el, 'width', 0);
	    if (sw > 0) return sw;
	}
    }

    // just return the cw we calculated before
    return cw;
}

//:NSFUNCTION Number getContentBoxHeight(Node el)
BU_HtmlBox.getContentBoxHeight = function(el) {
    if (bu_UA.isGecko) {
	return burst.xml.HtmlUtil.getComputedStylePixels(el, 'height');
    }

    var ch = BU_HtmlBox.getPaddingBoxHeight(el) - 
    BU_HtmlBox.getPaddingTopWidth(el) -
    BU_HtmlBox.getPaddingBottomWidth(el);

    if (bu_UA.isIE) return ch;

    // this browser doesn't support computed style. try for style.width

    if (bu_UA.isIE) { // never run
	if (!bu_UA.isIEBox) {
	    var ph = el.style.pixelHeight;
	    if (ph > 0) return ph;
	}
    }

    if (!bu_UA.isIEBox) {
	var pos = burst.xml.HtmlUtil.getComputedStyle(el, 'position');
	var try_style = (pos && pos == 'absolute');
	if (try_style) {
	    var sh = burst.xml.HtmlUtil.getComputedStylePixels(el, 'height', 0);
	    if (sh > 0) return sh;
	}
    }

    // just return the ch we calculated before
    return ch;
}

/** Return [left, top, width, height] */
//:NSFUNCTION Array getContentBoxRect(Node el)
BU_HtmlBox.getContentBoxRect = function(el) {
    return [BU_HtmlBox.getContentBoxLeft(el), BU_HtmlBox.getContentBoxTop(el),
	    BU_HtmlBox.getContentBoxWidth(el), BU_HtmlBox.getContentBoxHeight(el)];
}

/**
* The document-relative position of the margin box edge.
*/
//:NSFUNCTION Number getMarginBoxLeft(Node el)
BU_HtmlBox.getMarginBoxLeft = function(el) {
    return BU_HtmlBox.getBorderBoxLeft(el) - BU_HtmlBox.getMarginLeftWidth(el);
}

//:NSFUNCTION Number getMarginBoxTop(Node el)
BU_HtmlBox.getMarginBoxTop = function(el) {
    return BU_HtmlBox.getBorderBoxTop(el) - BU_HtmlBox.getMarginTopWidth(el);
}

//:NSFUNCTION Number getMarginBoxWidth(Node el)
BU_HtmlBox.getMarginBoxWidth = function(el) {
    return BU_HtmlBox.getBorderBoxWidth(el) +
    BU_HtmlBox.getMarginLeftWidth(el) +
    BU_HtmlBox.getMarginRightWidth(el);
}

//:NSFUNCTION Number getMarginBoxHeight(Node el)
BU_HtmlBox.getMarginBoxHeight = function(el) {
    return BU_HtmlBox.getBorderBoxHeight(el) +
    BU_HtmlBox.getMarginTopWidth(el) +
    BU_HtmlBox.getMarginBottomWidth(el);
}

/*
Get scroll position.

Scrolling behavior is controlled with the CSS style "overflow", in particular values "auto" and "scroll".

The scrolling cases include:
 - div
 - tbody
 - textarea
 - iframe
 - select/option
 - document
About tbody scrolling:
   http://groups.google.com/groups?hl=en&lr=&ie=UTF-8&oe=UTF-8&safe=off&threadm=Xns916DAC826BDCEtryandguessithehecom%40128.32.206.40&rnum=3

All browsers seem to support scroll* node properties:
  scrollLeft   pixel distance from canvas padding left of content, to inside of viewport border (0 if no scrolling)
  scrollTop    pixel distance from canvas padding left of content, to inside of viewport border (0 if no scrolling)
  scrollWidth  canvas padding box width
  scrollHeight canvas padding box height
They are readwrite in IE. 
There are several issues in Mozilla concerning scroll*:
   http://bugzilla.mozilla.org/show_bug.cgi?id=62536
     Available only on DIVs and BODY?
   http://groups.google.com/groups?hl=en&lr=&ie=UTF-8&oe=UTF-8&safe=off&threadm=akstbr%24hgq4%40ripley.netscape.com&rnum=10
     Writable only on BODY?
   http://groups.google.com/groups?hl=en&lr=&ie=UTF-8&oe=UTF-8&safe=off&threadm=an599l%24j3m1%40ripley.netscape.com&rnum=26
     scrollLeft/scrollTop are writable on DIVs?
In Safari, scrollLeft/scrollTop is apparently not available on DIVs, only BODY.

For a div with no scrollbars, scrollWidth =
                  IE:             clientWidth
		  Mozilla:        offsetWidth
                  Opera7:         offsetWidth
		  Safari:         clientWidth
For a div with no scrollbars, scrollHeight =
                  IE:             clientHeight
		  Mozilla:        offsetHeight
                  Opera7:         offsetHeight - top-border (=clientTop) - top-padding
		  Safari:         clientHeight
This makes it difficult to determine the canvas dimensions (i.e. not viewport dimensions)
of the element in a way that works regardless of whether it is scrolled or not.
On IE and Safari, we can simply return scrollWidth/scrollHeight for the canvas "padding box" width,
regardless of scrolling.
On Mozilla and Opera it is necessary to determine whether the div is scrolling; if it is
scrolling, use scrollWidth/scrollHeight, otherwise call getPaddingBoxWidth/getPaddingBoxHeight.

TODO: Address the cross-browser variability as to whether rendered content dimensions
can ever be found to be smaller than the viewport upon them.
*/

//:NSFUNCTION Number getScrollLeft(Node el)
BU_HtmlBox.getScrollLeft = function(el) {
  return el.scrollLeft;
}

//:NSFUNCTION Number getScrollTop(Node el)
BU_HtmlBox.getScrollTop = function(el) {
  return el.scrollTop;
}

//:NSFUNCTION Number getCanvasPaddingBoxWidth(Node el)
BU_HtmlBox.getCanvasPaddingBoxWidth = function(el) {
  if (bu_UA.isIE || bu_UA.isKHTML) return el.scrollWidth;
  // no scrolling
  if (el.scrollWidth == el.offsetWidth) return BU_HtmlBox.getPaddingBoxWidth(el);
  return el.scrollWidth;
}

// Unsupported on Opera7 when there is no scrolling (returns el.scrollHeight regardless).
//:NSFUNCTION Number getCanvasPaddingBoxHeight(Node el)
BU_HtmlBox.getCanvasPaddingBoxHeight = function(el) {
  if (bu_UA.isIE || bu_UA.isKHTML) return el.scrollHeight;
  // In mozilla, scrollHeight == offsetHeight when no scrolling.
  // In opera, no easy way to tell.
  if (bu_UA.isOpera && el.scrollHeight > el.clientHeight) return el.scrollHeight; // won't cover all cases.
  if (!bu_UA.isOpera && el.scrollHeight == el.offsetHeight) return BU_HtmlBox.getPaddingBoxHeight(el);
  return el.scrollHeight;
}


/*
Setting element position or size.

About style.left and style.top:
    - nothing set initially (empty string), unless the element has an inline style attribute setting it
    - readwrite
    - ignored if position:static
    - relative to static if position:relative
    - includes only content (not border/padding), except for the IE box model

If "position: relative" css, then style.left and style.top are offsets from what its static position would be.
If "position: static", then style.left and style.top are ignored.
If "position: absolute" and has a parent (also absolute), then style.left and style.right are relative to parent.

Browser may or may not create scroll bars: http://www.webreference.com/js/tips/011015.html
*/
//:NSFUNCTION void setContentBoxLeft(Node el, Number contentleft)
BU_HtmlBox.setContentBoxLeft = function(el, contentleft) {
    // in IE box model, style.left is box left so have to subtract.
    if (bu_UA.isIEBox) {
	contentleft = contentleft - BU_HtmlBox.getBorderBoxLeft(el) - BU_HtmlBox.getPaddingBoxLeft(el);
    }
    el.style.left = contentleft + 'px';
}

//:NSFUNCTION void setBorderBoxLeft(Node el, Number boxleft)
BU_HtmlBox.setBorderBoxLeft = function(el, boxleft) {
    // style.left is content left, so ordinarily have to add border and padding to it, except
    // in IE box model, style.left is already box left
    if (!bu_UA.isIEBox) {
	boxleft += BU_HtmlBox.getBorderBoxLeft(el) + BU_HtmlBox.getPaddingBoxLeft(el);
    }
    el.style.left = boxleft + 'px';
}

//:NSFUNCTION void setContentBoxTop(Node el, Number contenttop)
BU_HtmlBox.setContentBoxTop = function(el, contenttop) {
    // in IE box model, style.top is box top so have to subtract.
    if (bu_UA.isIEBox) {
	contenttop = contenttop - BU_HtmlBox.getBorderBoxTop(el) - BU_HtmlBox.getPaddingBoxTop(el);
    }
    el.style.top = contenttop + 'px';
}

//:NSFUNCTION void setBorderBoxTop(Node el, Number boxtop)
BU_HtmlBox.setBorderBoxTop = function(el, boxtop) {
    // style.top is content top, so ordinarily have to add border and padding to it, except
    // in IE box model, style.top is already box top
    if (!bu_UA.isIEBox) {
	boxtop += BU_HtmlBox.getBorderBoxTop(el) + BU_HtmlBox.getPaddingBoxTop(el);
    }
    el.style.top = boxtop + 'px';
}

//:NSFUNCTION void setContentBoxWidth(Node el, Number contentwidth)
BU_HtmlBox.setContentBoxWidth = function(el, contentwidth) {
    // in IE box model, style.width is box width so have to add.
    if (bu_UA.isIEBox) {
	contentwidth = contentwidth + BU_HtmlBox.getBorderLeftWidth(el) + BU_HtmlBox.getBorderRightWidth(el) +
	BU_HtmlBox.getPaddingLeftWidth(el) + BU_HtmlBox.getPaddingRightWidth(el);
    }
    el.style.width = contentwidth + 'px';
}

//:NSFUNCTION void setBorderBoxWidth(Node el, Number boxwidth)
BU_HtmlBox.setBorderBoxWidth = function(el, boxwidth) {
    // style.width is content width, so ordinarily have to add border and padding to it, except
    // in IE box model, style.width is already box width
    if (!bu_UA.isIEBox) {
	boxwidth = boxwidth - (BU_HtmlBox.getBorderLeftWidth(el) + BU_HtmlBox.getBorderRightWidth(el) +
	BU_HtmlBox.getPaddingLeftWidth(el) + BU_HtmlBox.getPaddingRightWidth(el));
    }
    el.style.width = boxwidth + 'px';
}

//:NSFUNCTION void setContentBoxHeight(Node el, Number contentheight)
BU_HtmlBox.setContentBoxHeight = function(el, contentheight) {
    // in IE box model, style.height is box height so have to add.
    if (bu_UA.isIEBox) {
	contentheight = contentheight + BU_HtmlBox.getBorderTopWidth(el) + BU_HtmlBox.getBorderBottomWidth(el) +
	BU_HtmlBox.getPaddingTopWidth(el) + BU_HtmlBox.getPaddingBottomWidth(el);
    }
    el.style.height = contentheight + 'px';
}

//:NSFUNCTION void setBorderBoxHeight(Node el, Number boxheight)
BU_HtmlBox.setBorderBoxHeight = function(el, boxheight) {
    // style.height is content height, so ordinarily have to add border and padding to it, except
    // in IE box model, style.height is already box height
    if (!bu_UA.isIEBox) {
	boxheight = boxheight - (BU_HtmlBox.getBorderTopWidth(el) + BU_HtmlBox.getBorderBottomWidth(el) +
	BU_HtmlBox.getPaddingTopWidth(el) + BU_HtmlBox.getPaddingBottomWidth(el));
    }
    el.style.height = boxheight + 'px';
}

/**
move to absolute document-relative position specified by numbers left and top (in pixels).
This function does not attempt to adjust for whether IE box model or not.
@todo could set runtimeStyle in IE instead?
@todo Opera want the 'px' or not?
*/
//:NSFUNCTION void setStylePos(Node el, Number left, Number top)
BU_HtmlBox.setStylePos = function(el, left, top) {
    el.style.left = left + 'px';
    el.style.top = top + 'px';
}

/**
increment existing values of style left and top by the amount specified.
assumes element already has a numerical or number + 'px' value in el.style.left or computed style left.
*/
//:NSFUNCTION void incStylePos(Node el, Number left_delta, Number top_delta)
BU_HtmlBox.incStylePos = function(el, left_delta, top_delta) {
    var curleft = burst.xml.HtmlUtil.getComputedStylePixels(el, 'left');
    el.style.left = (currleft + left_delta) + 'px';
    var curtop = burst.xml.HtmlUtil.getComputedStylePixels(el, 'top');
    el.style.top = (curtop + top_delta) + 'px';
}

/**
increments style width and height by indicated number of pixels (negative for shrink).
assumes there is already a value.
*/
//:NSFUNCTION void incStyleDims(Node el, Number width_delta, Number height_delta)
BU_HtmlBox.incStyleDims = function(node, width_delta, height_delta) {
  var curwidth = burst.xml.HtmlUtil.getComputedStylePixels(el, 'width');
  node.style.width = (curwidth + width_delta) + 'px';
  var curheight = burst.xml.HtmlUtil.getComputedStylePixels(el, 'height');
  node.style.height = (curheight + height_delta) + 'px';
}


/*
@todo: Workarounds for the IE box model.

Readings:
   http://www.tantek.com/CSS/Examples/boxmodelhack.html
      takes advantage of an IE-specific CSS parsing bug to have two widths specified
   http://www.webmasterworld.com/forum83/1376.htm
      mentions Tantek hack does not fix IE6 box model without doctype
   http://www.webmasterworld.com/forum83/652.htm
      Avoiding Tantek Celik's Box Model Hack 
      suggests divs with no padding or border (use additional wrapper div)
      suggests using <!--[if IE 5]> ...<![endif]-->
   http://jscript.dk/Util/BoxModelFix/
      a fix iterating over stylesheet rules altering width/height based on accessing padding and border values
      (assumes parseable px values)
   http://webfx.eae.net/dhtml/xmenu/xmenu.html
      a fix using the IE 'expression' attribute
   http://webfx.eae.net/dhtml/boxsizing/boxsizing.html
      a fix by implementing an IE behavior that supports the CSS3 box-sizing property
*/
BU_HtmlBox.fixBoxModel = function() {
}

//:NSEND HtmlBox
