/**
* @file HtmlUtil.js
*
* Defines burst.xml.HtmlUtil which contains static HTML DOM-related utility functions; see also burst.xml.HtmlBox .
*
* @author Copyright 2003 Mark D. Anderson (mda@discerning.com)
* @author Licensed under the Academic Free License 1.2 http://www.opensource.org/licenses/academic.php
*/

//=java package burst.xml;

/**
* Scoping class holding static HTML DOM-related utility functions.
* The scoping class burst.xml.HtmlBox has utilities specifically about element size and location.
*/
//:NSBEGIN HtmlUtil

burst.xml.HtmlUtil = {};

/**
* Extract the full 'src' attribute value of the 'script' element
* in this document which matches the provided regexp or string.
* It checks the type of its first argument).
* <p>
* This is useful when generating a script "src" value, because
* it will be interpreted relative to the base url of the document
* which included this file (location object), *not* relative to this file.
* 
* @param matching Either a Regexp to match against script, or a String to test substring
* @param mustFind If true, the function throws and exception if no match is returned
* @return The String value of the 'src' attribute of the matching 'script' element.
*/
//:NSFUNCTION String getDocumentScriptSrc(String matching, Boolean mustFind)
burst.xml.HtmlUtil.getDocumentScriptSrc = function(matching, mustFind) {
  var scripts = document.getElementsByTagName('script');
  if (typeof matching == 'string') {
  for(var i=0;i<scripts.length;++i) {
    if (scripts[i].src && scripts[i].src.indexOf(matching)!=-1) return scripts[i].src;
  }
  }
  else {
  for(/*var*/ i=0;i<scripts.length;++i) {
    if (scripts[i].src && scripts[i].src.match(matching)) return scripts[i].src;
  } 
  }
  if (mustFind) throw Error("none of the " + scripts.length + " script elements have a src matching " + matching);
  return null;
}

/**
* Add the provided value to any existing value for the "class" attribute of the provided Element.
* If there no "class" already or it has the empty value, it is set to this.
* Otherwise if first is true, it is prepended, otherwise appended,
* always preserving space separation. Also, it checks to see if it
* is already present, and if so, does nothing.
* @param node the element in question
* @param val the string id of a CSS class
* @param first (Optional) whether to prepend instead of append the val.
* @return whether it was added
*/
//:NSFUNCTION void addClassValue(Element node, String val, Boolean first)
burst.xml.HtmlUtil.addClassValue = function(el, val, first) {
  var s = burst.xml.DomUtil.getAttribute(el, 'class'); 
  if (!s || s.length == 0) {
    burst.xml.DomUtil.setAttribute(el, val);
    return true;
  }
  var vals = s.split(' ');
  if (burst.Alg.find(vals, val) != -1) return false;
  if (first) vals.shift(val); else vals.push(val);
  burst.xml.DomUtil.setAttribute(node, 'class', vals.join(' '));
  return true;
}

/**
* Remove the id val from the "class" attribute of the provided Element
* @param node the element in question
* @param val the string id of a CSS class
* @return whether it was found
*/
//:NSFUNCTION Boolean removeClassValue(Node node, String val)
burst.xml.HtmlUtil.removeClassValue = function(el, val) {
  var s = burst.xml.DomUtil.getAttribute(el, 'class'); 
  if (!s || s.length == 0) {return false;}
  var vals = s.split(' ');
  var ind = burst.Alg.find(vals, val);
  if (ind == -1) return false;
  vals.splice(ind, 1);
  burst.xml.DomUtil.setAttribute(node, 'class', vals.join(' '));
  return true;
}

/**
* Get a single style property. This returns the computed (aka 'current') style.
* @see http://www.xs4all.nl/~ppk/js/w3c_css.html
*/
//:NSFUNCTION String getStyleProperty(Node node, String propName)
burst.xml.HtmlUtil.getStylePropertyValue = function(node, propName) {
   if (propName == 'float') propName = 'cssFloat';
   var v;
   if (document.defaultView)
      v = document.defaultView.getComputedStyle(node,null).getPropertyValue(propName);
   else if (x.currentStyle)
      v = eval('node.currentStyle.' + propName);
   else bu_throw("no way to get style properties");
   return v;
}


/**
* Set a single style property (in node.style).
*/
//:NSFUNCTION void setStyleProperty(Node node, String propName, String propVal)
burst.xml.HtmlUtil.setStyleProperty = function(node, propName, propVal) {
  // in case our caller is confused. The DOM API spec says use 'cssFloat'. 'float' is just for full cssText.
  if (propName == 'float') propName = 'cssFloat';
  if (bu_UA.bug_set_style_attribute()) {
    node.style.setProperty(propName, propVal);
  }
  else {
    node.style[propName] = propVal;
  }
}

/** finds the maximum z-index set in the document. returns null if none set. */
//:NSFUNCTION Number maxZIndex(DomDocument doc)
burst.xml.HtmlUtil.maxZIndex = function(doc) {
  var maxz = null;
  // document.getElementsByTagName("*")
  burst.xml.DomUtil.for_children(doc.documentElement, function(n) {
				     var nz = burst.xml.HtmlUtil.getComputedStyle(n, 'z-index');
				     var nzi = parseInt(nz);
				     if (nz && !isNaN(nzi) && (maxz == null || nz > maxz)) maxz = nz;
				 });
  return maxz;
}

/**
Get computed style value.

See:
  http://www.xs4all.nl/~ppk/js/w3c_css.html 
  http://www.w3.org/TR/1998/REC-CSS2-19980512/cascade.html#computed-value

The getPropertyValue function takes a CSS hyphenated "style attribute" name like 'font-size', 'z-index', 'background-color'.
When using members of a style object, such as node.style['color'], it is necessary to have "style property" member names like
'fontSize', 'zIndex', 'backgroundColor'.

Only Mozilla currently supports getComputedStyle.
In Mozilla, both node.style and getComputedStyle support access both by getPropertyValue(attname) 
and by propname members.

The IE style and currentStyle objects support access only by propname, both as expando and using getAttribute.
IE also has a runtimeStyle object, which can be used for overrides.

Opera 7 supports neither currentStyle nor getComputedStyle.
Its style object does expand out 'border' and 'padding' settings to their individual property values,
though still preserving the original units:
if it was 'padding: 0.25in', then there will be a 'padding-left' value of '0.25in'.
But it otherwise does not have a way to retrieve a computed style attribute.

The return value is always a String. An empty string means no value. (This is how the style object works
in all browsers).

The standard says that computed values should in fact be computed numbers.
This is not always the case; some browsers return the actually set strings, or return nothing at all.

Mozilla will return a number + 'px' value for some attributes regardless of whether they were explicitly set or not,
and regardless of whether it was set by an expression such 'auto' or '37%' or '5in'.
These reliable attribute names include: 
   'width', 
   'padding-left', 'padding-right'
   'border-left-width', 'border-right-width',
   'margin-left', 'margin-right', 
   'height',
   'padding-top', 'padding-bottom'
   'border-top-width', 'border-bottom-width',
   'margin-top', 'margin-bottom'
This does not include:
   'left', 'right',
   'top', 'bottom',
Furthermore, if just 'border' or 'padding' is set explicitly in css, 
Mozilla getComputedStyle will return appropriate values for the particular
measurements on each side such as 'border-left-width' (the reverse is not true).

IE's currentStyle object does not generally do conversion. If a node style is 'width: 37%', then currentStyle['width'] == '37%'.
Like Opera, it does expand border and padding settings to the individual ones, preserving original units.

Note that the CSS rules may be accessed via:
document.styleSheets[0].cssRules[0].style.borderWidth
*/
//:NSFUNCTION String getComputedStyle(Node node, String attname)
burst.xml.HtmlUtil.getComputedStyle = function(node, attname) {
    // Mozilla
    var doc = node.ownerDocument; // document. ownerDocument not in IE5.5
    if (doc && doc.defaultView && doc.defaultView.getComputedStyle) {
      // second arg should be null. Some claim that some version of Netscape requires ''?
	var s = doc.defaultView.getComputedStyle(node, null).getPropertyValue(attname);
      return s;
    }
	
    var propname = burst.xml.HtmlUtil.stylePropertyName(attname);

    // IE
    if (bu_UA.isIE) {
	// No need to do eval('node.currentStyle.' + propname), can do node.currentStyle[propname]
	return node.currentStyle[propname];
	// Still need to convert to property name if use getAttribute
	//return node.currentStyle.getAttribute(attname);
    }

    // Opera 7 has no currentStyle or getComputedStyle(), so only works with inline style attributes
    // http://groups.google.com/groups?hl=en&lr=&ie=UTF-8&oe=UTF-8&safe=off&selm=oprjvykih2iaf507%40news.opera.no
    // http://www.opera.com/docs/specs/js/dom/css/index.dml
    var val = node.style[propname];
    // alert("no way to getComputedStyle: " + document.defaultView.getComputedStyle + " or " + node.currentStyle + " so style[" + propname + "]=" + val);
    return val;
}

/**
Convert a string like '30' or '30px' into a number like 30.
*/
//:NSFUNCTION Number parseStylePixels(String s, String attname, Number fallback)
burst.xml.HtmlUtil.parseStylePixels = function(s, attname, fallback) {
    // make sure it is entirely digits, or ends in 'px'
    if (s && (/px$/i.test(s) || /^\d+$/.test(s)) ) {
	var v = parseInt(s);
	if (!isNaN(v)) {
	    return v;
	}
    }
    //throw Error("unexpected value '" + s + "' from style value of '" + attname + "'");
    if (typeof fallback != 'undefined') return fallback;
    // TEMPORARY
    alert("unexpected value '" + s + "' from style value of '" + attname + "'");
    return 0;
}

/**
returns a number
if fallback is not defined, and it finds a non-empty string it can't parse as a number, it throws.
*/
//:NSFUNCTION Number getComputedStylePixels(Node node, String attname, Number fallback)
burst.xml.HtmlUtil.getComputedStylePixels = function(node, attname, fallback) {
    //if (!node) alert("no node, for attname=" + attname);
    var s = burst.xml.HtmlUtil.getComputedStyle(node, attname);

    // if failed, try normal style
    if (!s || s == '') {
	var propname = burst.xml.HtmlUtil.stylePropertyName(attname);
	var inlineval = node.style[propname];
	//alert("no value for attname '" + attname + "': " + (typeof s) + " style='" + inlineval + "'");
	return 0;
    }

    return burst.xml.HtmlUtil.parseStylePixels(s, attname, fallback);
}

/**
convert an "attribute name" (such as "font-size") to a "property name" (such as "fontSize").
*/
//:NSFUNCTION String stylePropertyName(String attname)
burst.xml.HtmlUtil.stylePropertyName = function(attname) {
  var words = attname.split('-');
  if (words.length == 1) return attname;
  var a = [words[0]];
  for(var i=1;i<words.length;++i) {a.push(burst.Text.ucfirst(words[i]))}
  return a.join('');
}

/**
Return doc.body or doc.documentElement according to IE compatMode. 
Only for use with Internet Explorer.
*/
//:NSFUNCTION Node getDocumentElementIE(DomDocument doc)
burst.xml.HtmlUtil.getDocumentElementIE = function(doc) {
    if (!doc) doc = document;
    return doc.compatMode == 'CSS1Compat' ? doc.documentElement : doc.body;
}





/**
* Parse a string which follows the conventions of a CSS style declaration.
* Any example might be <code>a: 1; foobar: 'some text'</code>
*
* Note that this does not parse a CSS <i>rule</i> (which would include
* selectors and {}). Just the body declaration, as might appear in a
* style attribute.
*
* This is not a perfect emulation. If you have a <em>real</em> style,
* you should use the style object and call the official 
* <code>style.getPropertyValue</code> or <code>style.getPropertyCSSValue</code>
* methods. This function is intended just for use if you have a string
* which mimics the CSS syntax.
* 
* Returns an associative array with keys being property names.
* Both names and values have surrounding white space stripped.
*
* Quoted values are unquoted, if the entire value string (between ':' and ';')
* is quoted. If the value string contains an internal quoted string, it
* is preserved.
*/
//:NSFUNCTION Object parseStyle(String text)
burst.xml.HtmlUtil.parseStyle = function(text) {
  if (!text || text.length == 0) return {};
  var a = burst.Text.splitTerms(text, '[\\w,#%+-]+|;|:');
  var map = {};
  if (a.length == 0) return map;
  var i=0;
  while(true) {
    var name = a[i++];
    if (a[i++] != ':') bu_throw("found '" + a[i-1] + "' expected ':'");
    var vals = [];
    for(;i<a.length;i++) {
      if(a[i] == ';') {i++; break;}
      // unquote at the whole value level
      vals.push(a[i]);
      //vals.push(burst.Text.unquote(a[i]));
    }
    var valstring = burst.Text.unquote(vals.join(' '));
    //bu_debug("name='" + name + "', val='" + valstring + "'");
    map[name] = valstring;
    //map[name] = vals.join(' ');
    if(i==a.length) break;
  }
  //bu_debug("converted '" + text + "' to: " + burst.Lang.uneval(map));
  return map;
}


/*
Convenience to get the body contents of a window as html.
*/
/*
burst.xml.HtmlUtil.getWindowBodyHtml = function(win) {
  // body is second child of html, after head
  return win.document.documentElement.childNodes[1].innerHTML;
}
*/




/**
Return the base URI for the provided window.

It determines this from these sources:
- any 'base' element with a 'href' attribute
- win.location.href

(Note that Content-Base is not in the HTTP 1.1 RFC (it is not in 2616, only 2068),
so 'META http-equiv="Content-Base" content="http://w.com/"' is a non-issue.)

In an xml document, the "xml:base" attribute is relevant: http://www.w3.org/TR/xmlbase/
However, as that could appear anywhere in the hierarchy, we don't
deal with it.

In general, the string this function returns might be absolute or relative.

See http://www.w3.org/TR/REC-html40/struct/links.html#resolving-relative-uris
If a base element is provided, it must have an href attribute, which must
be an absolute URI.

@param win The window.
@param recurse_parent If this windows base url is relative, whether it should recurse to parent. 
*/
//:NSFUNCTION String getWindowBaseUrl(Window win, Boolean recurse_parent)
burst.xml.HtmlUtil.getWindowBaseUrl = function(win, recurse_parent) {
  var base_url;

  // Want '/html/head/base'. documentElement=html, firstChild=head.
  var head_el = win.document.documentElement.firstChild;
  var base_el
  if (head_el)
    base_el = burst.xml.DomUtil.firstChildWithTagName(head_el, 'base');
  if (base_el) {
    base_url = base_el.getAttribute('href');
  }
  if (!base_url) base_url = win.location.href;

  if (recurse_parent && !burst.URI.isAbsolute(url) && win.parent && win != win.parent) {
     return burst.xml.HtmlUtil.resolveWindowUrl(base_url, win.parent);
  }
  return base_url;
}

/**
Resolve the uri relative to the provided window.

The logic is:
- If url is already absolute, it is returned.
- Determine the base url of window (which may require a recursive call through window parents)
- Relativize url to that base url

The result may still not be an absolute url.
*/
//:NSFUNCTION String resolveWindowUrl(String url, Window win)
burst.xml.HtmlUtil.resolveWindowUrl = function(url, win) {
  if (burst.URI.isAbsolute(url)) return url;

  var base_url = burst.xml.HtmlUtil.getWindowBaseUrl(win);
  return burst.URI.resolveUrl(url, base_url, true);
}

burst.xml.HtmlUtil.createImg = function(doc, src, alt, title) {
    /*
    if (false && bu_UA.isIE) {
	var s = '<img src="' + src + '"';
	if (typeof alt !== 'undefined' && alt) s += ' alt="' + alt + '"';
	if (typeof title !== 'undefined' && title) s += ' title="' + title + '"';
	s += ' width="9px" height="9px"';
	s += ' style="display: inline;"';
	s += '>';
	alert("creating element with: " + s);
	return doc.createElement(s);
    }
    else {
    */
    return burst.xml.DomUtil.structToNode(['img', {src: src, alt: alt, title: title}], doc); 
    //}
}


/**
* Create a hidden iframe element child in the current document.
*
* @todo what browsers require Function vs. String for onload?
* @todo For browsers that support iframe onload, do the events bubble up? does returning false help?
*
* @param iframe_id The value for both the 'name' and 'id' attributes.
* @param srcval Optional. The value for the 'src' attribute. If undefined, defaults to burst.io.IframeURIRequest.NO_SRC
* @param onloadval Optional. The value for the 'onload' attribute. can be a Function or String.
* @param create_cb Optional. Called with the newly created element (prior to attaching to parent, if possible).
* @param as_visible Optional. If true, do not create the iframe to be initially invisible (it won't set size either though).
* @return the created Element
*/
//:NSFUNCTION Element iframeCreate(String iframe_id, String srcval, String onloadval, Function create_cb, Boolean as_visible)
burst.xml.HtmlUtil.iframeCreate = function(iframe_id, srcval, onloadval, create_cb, as_visible) {
  if (typeof srcval == 'undefined' || srcval == null) {
     srcval = burst.io.IframeURIRequest.NO_SRC;
  }

  if (!document.body) 
      bu_throw("(burst.xml.HtmlUtil.js) no document.body to append the iframe element child to, window.location.href=" +
            (window.location ? window.location.href : 'undefined'));
  var iframe_el_parent = document.body;

  var needs_onload = true;
  if (typeof onloadval == 'undefined' || onloadval == null) {
    needs_onload = false;
    // just for debugging
    onloadval = "alert('(HtmlUtil.js) in onload from initial iframe create of url " + srcval + "')";
  }

  var iframe_el;

  if (bu_UA.bug_create_iframe() || // IE5.0
      (needs_onload && bu_UA.can_iframe_onload_static() && !bu_UA.can_iframe_onload_dyn()) // IE5.5 and IE6
  ) {
    var visstyle = bu_UA.bug_iframe_display_none() ? 
       'visibility: hidden; border: 0px; height: 0px; width: 0px;' :
       'display: none;';
    // visstyle = '';
    if (typeof onloadval == 'undefined') onloadval = null; // so burst.xml.DomUtil.el will ignore it without complaint
    if (onloadval && typeof onloadval != 'string') {
      if (typeof onloadval != 'function') bu_throw("(burst.xml.HtmlUtil.js) onloadval neither string nor function: " + onloadval);
      onloadval = 'parent.' + burst.Lang.unevalFunctionCall(onloadval);
    }
    bu_debug_Dom("(burst.xml.HtmlUtil.js) bug_create_iframe() || needs_onload=true. using html to create iframe with id " + iframe_id + " and onload=" + onloadval);

    var html = burst.xml.DomUtil.el('IFRAME', {
	ID:     iframe_id,
	NAME:   iframe_id,
	STYLE:  visstyle,
        // Not called: Opera.
	// Is called:  IE6.
        ONLOAD: onloadval,
	SRC:    srcval
	}, null);

    bu_debug("(burst.xml.HtmlUtil.js) inserting html: ", html);
    // iframe_el_parent.innerHTML += html;
    iframe_el_parent.insertAdjacentHTML('beforeEnd', html);

    // todo: IE5.0 PC shouldn't have to use document.all[iframe_id] because we didn't do it with createElement
    /*
http://msdn.microsoft.com/workshop/author/dhtml/reference/properties/name_2.asp
"The NAME attribute cannot be set at run time on elements dynamically
created with the createElement method. To create an element with a name
attribute, include the attribute and value when using the createElement
method."
    */
    iframe_el = document.getElementById(iframe_id) || bu_throw("can't get the iframe I just created");
    if (typeof create_cb == 'function') create_cb(iframe_el);
  }

  else {
    bu_debug_Dom("(burst.xml.HtmlUtil.js) bug_create_iframe()=false. using createElement to create iframe with id " + iframe_id + " bug_iframe_display_none=" + bu_UA.bug_iframe_display_none() + " src=" + srcval);
    iframe_el = document.createElement('iframe');

    if (create_cb) create_cb(iframe_el);

    iframe_el.setAttribute('id', iframe_id);
    iframe_el.setAttribute('name', iframe_id);

    // TEMPORARY
    if (!as_visible) {
      if (bu_UA.bug_iframe_display_none()) {
        iframe_el.style.visibility = 'hidden';
        iframe_el.style.border = '0px';
        iframe_el.style.height = '0px';
        iframe_el.style.width = '0px';
      }
      else
        iframe_el.style.display = 'none';
    }

    // done this way before Safari 
    //iframe_el.src = srcval;
    iframe_el.setAttribute('src', srcval);

    // done this way:  
    //   setAttribute onload called by:     Moz1.3
    //   setAttribute onload not called by: IE6 Opera
    //iframe_el.setAttribute('onload', onloadval);
    if (onloadval) {
       // bu_debug_Dom('(burst.xml.HtmlUtil.js) setting iframe_el.onload=' + onloadval);
       if (typeof onloadval == 'function') iframe_el.onload = onloadval;
       else iframe_el.onload = new Function(onloadval);
    }
    iframe_el_parent.appendChild(iframe_el);
  }

  // getAttribute('name') after html creation works for: IE6, Opera
  // getAttribute('name') after createElement works for: IE6, not Opera
  if (!iframe_el.getAttribute('name')) 
    bu_debug_Dom("(HtmlUtil.js) the iframe element I just made does not have a value for getAttribute('name')");

/*
  bu_alert("(HtmlUtil.js) has window.frames[" + iframe_id + "]=" + (window.frames[iframe_id]?true:false));
  //bu_alert("(HtmlUtil.js) has window.frames[" + iframe_id + "]=" + window.frames[iframe_id]);
  bu_alert("(HtmlUtil.js) window.frames.length=" + window.frames.length);
*/
/*
  if (!window.frames[iframe_id])
    setTimeout(function() {
         alert("(HtmlUtil.js) after timeout, window.frames[" + iframe_id + "]=" + (window.frames[iframe_id]?true:false))
   }, 1000);
*/

  // opera won't have one here if created via innerHTML, but will for createElement, at least after a timeout
  if (typeof window != 'undefined' && !window.frames[iframe_id]) {
     //bu_throw("The iframe I just created is not in window.frames");
     if (BU_DBG_DOM || !bu_UA.bug_iframe_delayed_create())
       bu_alert("(burst.xml.HtmlUtil.js) The iframe I just created is not in window.frames");
  }

  return iframe_el;
}


/**
Go from iframe element (in the parent document) to the iframe window object.

Current W3 standards define contentDocument but do not require a contentWindow method.

Here a bit of the sad story we work around here:
<ul>
<li>NS6 iframe window has contentDocument, not contentWindow.
<li>Gecko 1.0.1 has contentWindow.
<li>IE5 Mac and IE5.0 PC has no contentWindow, but do have a document child.
<li>IE5.5+ PC has contentWindow but no contentDocument.
<li>IE4+ has document.parentWindow but Mozilla does not. Mozilla and Opera have document.defaultView
<li>Mozilla has __parent__
</ul>
*/
//:NSFUNCTION Window iframeContentWindow(Element iframe_el)
burst.xml.HtmlUtil.iframeContentWindow = function(iframe_el) {
  bu_assertArgs('iframeContentWindow',1,arguments);
  var win = 
     iframe_el.contentWindow || // IE
     burst.xml.HtmlUtil.iframeContentDocument(iframe_el).defaultView || // Moz, opera
     burst.xml.HtmlUtil.iframeContentDocument(iframe_el).__parent__ || // Moz. TODO: is this available when defaultView isn't?
     (iframe_el.name && document.frames[iframe_el.name]) ||
     bu_unsupported('iframeContentWindow');
  return win;
}

/**
Get the Document object for the iframe.

Note: iframe_el.document, which is set in Opera and IE6, is the containing document, not the
document of the iframe. It should *not* be used.
*/
//:NSFUNCTION Document iframeContentDocument(Element iframe_el)
burst.xml.HtmlUtil.iframeContentDocument = function(iframe_el) {
  bu_assertArgs('iframeContentDocument',1,arguments);
/*
alert("contentDocument=" + iframe_el.contentDocument);
alert("contentWindow=" + iframe_el.contentWindow);
if (iframe_el.contentWindow) alert("contentWindow.document=" + iframe_el.contentWindow.document);
alert("iframe_el.document=" + iframe_el.document);
*/
  var doc =
    iframe_el.contentDocument || // W3
    (iframe_el.contentWindow && iframe_el.contentWindow.document) ||  // IE
    (iframe_el.name && document.frames[iframe_el.name] && document.frames[iframe_el.name].document) ||
    bu_unsupported('iframeContentDocument');
//alert("returning iframeContentDocument=" + doc);
  return doc;
}

/**
Get the Location object for the iframe element.
*/
//:NSFUNCTION Location iframeLocation(Element iframe_el)
burst.xml.HtmlUtil.iframeLocation = function(iframe_el, fallback_name) {
  bu_assertArgs('iframeLocation',1,arguments);
  var loc =
    // Internet Explorer
    (iframe_el.contentWindow && iframe_el.contentWindow.location) || 

    // Mozilla and W3
    (iframe_el.contentDocument &&                                    
         (iframe_el.contentDocument.location ||
          (iframe_el.contentDocument.defaultView && iframe_el.contentDocument.defaultView.location)));

  // Opera and general fallback, see http://www.xs4all.nl/~ppk/js/iframe.html
  if (!loc) {
    bu_alert("could not find frame window using expected methods");
    var name = iframe_el.getAttribute('name') || fallback_name;
    loc = frames && frames[name] && frames[name].location;
    if (loc) bu_alert("found in frames");
    if (!loc) {
      loc = document.frames && document.frames[name] && document.frames[name].location;
      if (loc) bu_alert("found in document.frames");
    }
/*
    if (!loc)
      loc = iframe_el.document && iframe_el.document.location;
*/
    if (iframe_el.location) {
       bu_alert("iframe element itself has a location " + iframe_el.location + " ===window.location:" + (window.location===iframe_el.location));
       loc = iframe_el.location;
    }
    if (!loc)
      bu_unsupported("iframeLocation, for iframe with getAttribute('name')='" + 
            iframe_el.getAttribute('name') + "' and fallback_name='" + fallback_name + "'");
  }
  return loc;
}

/** 
Replaces the contents of the iframe with the provided html.
Calls <code>iframeContentDocument(iframe_el)</code> and performes a Document.write on it with the html.
*/
//:NSFUNCTION void iframeReplaceContents(Element iframe_el, String html)
burst.xml.HtmlUtil.iframeReplaceContents = function(iframe_el, html) {
    var idoc = burst.xml.HtmlUtil.iframeContentDocument(iframe_el);
    if (idoc === document) bu_throw("(burst.xml.HtmlUtil.js) attempt to write over my own document");

	//bu_alert('(burst.xml.HtmlUtil.js) idoc=' + idoc);
    idoc.close(); // or idoc.clear() ?
	//bu_alert('(burst.xml.HtmlUtil.js) after idoc close' + idoc);
    idoc.open(/*'text/html'*/); // w3 has no parameters. IE does not support it. text/html is default anyway.
	//bu_alert("(burst.xml.HtmlUtil.js) about to write html:\n" + html);
    idoc.write(html);
	//bu_alert("(burst.xml.HtmlUtil.js) about to close");
    idoc.close();
	//bu_alert("(burst.xml.HtmlUtil.js) idoc is closed");
}


// useful for debugging
burst.xml.HtmlUtil.iframeSetVisible = function(iframe_el, isvis) {
  if (bu_UA.bug_iframe_display_none()) {
    if (isvis) {
      iframe_el.style.visibility = 'visible';
      iframe_el.style.width = '300px';
      iframe_el.style.height = '100px';
      //iframe_el.width = 300;
      //iframe_el.height = 100;
    }
    else {
      iframe_el.style.visibility = 'hidden';
      iframe_el.style.width = '0px';
      iframe_el.style.height = '0px';
    }
  }
  else {
    iframe_el.style.display = isvis ? 'block' : 'none';
  }
}

//:NSEND HtmlUtil


bu_loaded('burst.xml.HtmlUtil', ['burst.xml.DomUtil', 'burst.Alg', 'burst.web.UserAgent', 'burst.Text', 'burst.io.IframeURIRequest']);
