/**
* @file XPath.js
*
* A simplified XPath implementation, including class burst.xml.XPath.
*
* @author Copyright 2003 Mark D. Anderson (mda@discerning.com)
* @author Licensed under the Academic Free License 1.2 http://www.opensource.org/licenses/academic.php
*/

//=java package burst.xml;

/**
Class representing a complied XPath expression.

* We support a simplified xpath syntax. The path contains 1 or more of the following,
* separated by '/', such as:
* <pre>
*   para     children named 'para'
*   para[5]  just the 5th child named 'para' (1-based)
*   *        all element children
*   text()   all text node children
* </pre>

The beginning of a path can be:
<pre>
           nothing, just start with the provided context node 
  .        same as nothing (start with context node), but useful to prefix //
  /        start from document and ignore context node
  //       look for the next path segment to begin any number of levels below context node
</pre>

In the middle of the path there can be:
<pre>
  //       go to all descendants        
</pre>

Not supported:
- anything to do with attributes
- any predicate expressions besides a single position
- path combination with |

@todo When there is a fast real XPath implementation available, use that instead.

<h3>See Also</h3>

<h4>Simplified XPath Languages</h4>

For other attempts at simplifying XPath for various purposes, see for example:
- Streamed XPath by http://www.pms.informatik.uni-muenchen.de/lehre/projekt-diplom-arbeit/streamedxpath.html
- SXPath (Sequential XPath) by Arpan Desai

<h4>Less than XPath in browsers</h4>

Starting with MSXML 2 (IE5.5+), IE has Node.selectNodes(expr) and Node.selectSingleNode(expr) which can be used on XML documents for
example from XMLDOM or XMLHTTP. There are installation/version complexities as to whether the expression language
is true XSLT or not; generally speaking, MSXML2 (IE5.5) is Microsoft's language, and MSXML3 (IE6) is standard.

Mozilla implements part of DOM Level 2 Traversal: http://www.w3.org/TR/DOM-Level-2-Traversal-Range/traversal.html
It implements TreeWalker and NodeFilter, but no NodeIterator so far.

<h4>Full XPath in browsers</h4>
In .NET, Microsoft there is a full XPath implementation including a family of interfaces such as XPathNodeIterator, XPathNavigator, XPathDocument, XPathExpression.
(Note that XPathDocument (XPathDocumentIterator) and XmlDocument (DocumentXPathNavigator) both implement
IXPathNavigable but XPathDocument has a more efficient implementation for XPath queries. 
Also re-using XPathNodeIterator instances avoids redundant expression compilation.)
An example is:
<pre>
  XPathNavigator nav = new XPathDocument(file).CreateNavigator();
  XPathExpression expr = nav.Compile('//bar[\@name="id17"]');
  XPathNodeIterator nodes = nav.Select(expr);
</pre>

In Mozilla (including Netscape 7) there is an implementation of the W3 DOM Level 3 XPath: http://www.w3.org/TR/DOM-Level-3-XPath/
This module has among other functions:
   <code>doc.evaluate(DOMString expression, Node contextNode, XPathNSResolver resolver,  unsigned short type, DOMObject resultexpr)</code>
This can be used if you have a DomDocument (from XMLHttpRequest responseXML or createDocument).
Other interfaces include XPathException, XPathEvaluator, XPathExpression, XPathNSResolver, XPathResult, XPathNamespace.
An example is:
<pre>
    var some_bar_node = xmlDocument.evaluate('//bar[\@name="id17"]', xmlDocument, null, 9, null).singleNodeValue;
    var idval = some_bar_node.getAttribute("name"); // idval should now be == 'id17'
</pre>

*/

//=java class XPath {


/**
* A function to return a burst.xml.XPath object.
* Internally it can be used as a constructor, but it should be
* called as a function so that we can cache compiled XPath instances
* based on the arguments.
*
* The flags supported are:
* - 'h', assume an HTML DOM, which means that all names returned by the DOM are in all uppercase
*
* Note that we assume that the HTML DOM tagName and nodeName always return uppercase;
* we leave their return values unchanged, and just uppercase our strings before comparing.
*
* If flags are not provided (undefined), we default to 'h'. 
*
* @todo do all browsers return uc for tagName and nodeName? do they do it even on xhtml doctype, or mime type text/xml?
*/
//=java public XPath(String pathstr, String flags) {}
function BU_XPath(pathstr, flags) {
  // not called as constructor, so leverage cache
  if (!(this instanceof BU_XPath)) {
    return burst.MOP.memoizeNew1(BU_XPath.instance_cache, BU_XPath, pathstr);
  }
  bu_debug("burst.xml.XPath(" + pathstr + ")");

  // default flags are 'h', and if 'h', we upper-case the entire path
  if (!flags) flags = 'h';

  // Properly, we should replace '//' with /descendant-or-self::node()/
  // Instead we replace '//' with '/descendant::' and hope there is another
  // step and that it doesn't have an axis.
  // Note that the XPath standard says '//para[1]' means the first descendant para,
  // while '/descendant::para[1]' means all para elements that are the first within their parent.
  // Regardless of whether that is correct, we always implement the latter semantic:
  // we do not collect counts per step (i.e. node set), but per iteration descent level.
  // Note also that currently we perform an optimization that at most 1 child can match
  // a predicate (because our predicate support is so limited).
  pathstr = pathstr.replace(/\/\//g,'/descendant::');

  // leading '/' means that we should start with document, not the passed-in node
  if (pathstr.charAt(0) === '/') {
    this.is_absolute_ = true;
    pathstr = pathstr.substring(1);
  }
  else {
    this.is_absolute_ = false;
  }

  // split on '/'
  var stepstrs = pathstr.split(/\//);

  bu_debug("pathstr='" + pathstr + "' stepstrs=" + stepstrs.join('|'));

  this.steps_ = burst.Alg.transform(stepstrs, function(stepstr) {return BU_XPathStep(stepstr,flags)});
  this.str_ = pathstr;
  return this; // silence warnings
}

burst.xml.XPath = BU_XPath;
BU_XPath.instance_cache = {};



// some day these might do things, but for now these are just used
// as pointers, and we use an external switch statement.
function BU_XPath_Axis_descendant_or_self() {}
function BU_XPath_Axis_descendant() {}
function BU_XPath_Axis_self() {}
function BU_XPath_Axis_child() {}

BU_XPath.AXES = {
  descendant_or_self: BU_XPath_Axis_descendant_or_self,
  descendant: BU_XPath_Axis_descendant,
  self: BU_XPath_Axis_self,
  child: BU_XPath_Axis_child
};

BU_XPath.compilePredicate = function(pred_str) {
  if (!pred_str.match(/^\d+$/)) bu_throw("predicate '" + pred_str + "' not supported");
  match_index = parseInt(pred);
  return function(node, pos) {return pos == match_index;}
}

function BU_XPath_NodeTest_text(node)    {return node.nodeType == Node.TEXT_NODE;}
function BU_XPath_NodeTest_node(node)    {return true;}
function BU_XPath_NodeTest_element(node) {return node.nodeType == Node.ELEMENT_NODE;}

BU_XPath.compileNodeTest = function(node_test_str, flags) {
  switch(node_test_str) {
    case '':        bu_throw("empty NodeTest");
    case 'text()' : return BU_XPath_NodeTest_text;
    case 'node()' : return BU_XPath_NodeTest_node;
    case '*' :      return BU_XPath_NodeTest_element;
    default:
      // qname or prefix:*  
      if (typeof node_test_str != 'string') alert("node_test_str " + node_test_str + " is a " + (typeof node_test_str));
      if (/^[\w\.\-\_\:]+$/.test(node_test_str)) {
        if (flags.indexOf('h') != -1) {node_test_str = node_test_str.toUpperCase();}
//bu_debug("compiling node test string '" + node_test_str + "'");
        return function(node) {return node.nodeType == Node.ELEMENT_NODE && node.nodeName == node_test_str};
      }
      bu_throw("unsupported NodeTest '" + node_test_str + "'");
  }
  throw("never reached"); // silence warnings
}

var BU_XPathStep_instance_cache = {};

// function or constructor
// parses axis:: nodetext predicate*
function BU_XPathStep(step_str,flags) {
  if (!(this instanceof BU_XPathStep)) {
    return burst.MOP.memoizeNew2(BU_XPathStep_instance_cache, BU_XPathStep, step_str, flags);
  }

  bu_debug("BU_XPathStep(" + step_str + ")");
  // original string
  this.str_ = step_str;
  this.preds_ = null;

  if (step_str == '.') {
    this.axis_ = BU_XPath_Axis_self;
    this.node_test_ = BU_XPath_NodeTest_node;
  }
  else if (step_str == 'descendant-or-self::node()') {
    this.axis_ = BU_XPath_Axis_descendant_or_self;
    this.node_test_ = BU_XPath_NodeTest_node;
  }
  else {
    // parse axis
    var rest = step_str;
    var axis_ind = step_str.indexOf('::');
    if (axis_ind != -1) {
      var axis_str = step_str.substring(0,axis_ind);
      rest = step_str.substring(axis_ind + 2);
      if (rest.indexOf('::')!=-1) bu_throw("multiple axes in step '" + step_str + "'");
      this.axis_ = BU_XPath.AXES[axis_str] || bu_throw("unknown axis '" + axis_str + "'");
    }
    else {
      this.axis_ = BU_XPath_Axis_child;
    }

    // parse node test
    var node_test_str;
    var pred_ind = rest.indexOf('['); // ']'
    if (pred_ind != -1) {
      node_test_str = rest.substring(0,pred_ind);
      rest = rest.substring(pred_ind + 1);
    }
    else {
      node_test_str = rest;
    }
    this.node_test_ = BU_XPath.compileNodeTest(node_test_str, flags);

    // parse any predicates
    if (pred_ind != -1) {
      this.preds_ = [];
      while(true) { // '['
        var end_pred = rest.indexOf(']');
        if (end_pred == -1) bu_throw("no end to predicate in '" + step_str + "'");
        var pred_str = rest.substring(0,end_pred);
        this.preds_.push(BU_XPath.compilePredicate(pred_str));
        if (rest.length == end_pred + 1) break;
        if (rest.charAt(end_pred + 1) != '[') //']'
           bu_throw("expected [ at '" + rest.substring(end_pred+1) + "' in '" + step_str + "'");
        rest = rest.substring(end_pred + 2);
      }
    }
  }
  return this; //silence warnings
}

BU_XPathStep.prototype.isMatch = function(node, match_counts) {
  var is_match = this.node_test_(node);

  // check any predicates
  if (is_match && this.preds_) {
    if (!match_counts) bu_throw("no match_counts");
    for(var i=0;i<this.preds_.length;++i) {
       var pred = step.preds_[i];
       var m = match_counts[i] || 0;
       if (pred(node, m)) {
         match_counts[i] = m + 1;
       }
       else {
         is_match = false;
         break;
       }
    }
  }

  return is_match;
}

/**
* Call the function for each match to the path, using the provided node as the
* context node.
* If the function returns true, iteration stops.
*
* If the path is absolute, then the node argument is irrelevant and can be null.
* @param node The context node
* @param func The callback function
*/
//:CLIMETHOD void iterate(Node node, Function func)
BU_XPath.prototype.iterate = function(node, func) {
  var cur_node = this.is_absolute_ ? document : node;
  // deal with IE5.0 and passing in document, which has no nodeType and no children.
  if (typeof cur_node.nodeType == 'undefined' && typeof cur_node.documentElement != 'undefined')
     cur_node = cur_node.documentElement;
  this.iterate_(node, func, cur_node, 0, 0);
}

/**
* Returns the first matching node, or undefined if none.
*
* See http://msdn.microsoft.com/library/en-us/xmlsdk30/htm/xmmthselectsinglenode.asp
* @param node The context node
*/
//:CLIMETHOD Node selectSingleNode(Node node)
BU_XPath.prototype.selectSingleNode = function(node) {
  var match;
  this.iterate(node, function(n) {match = n; return true;});
  return match;
}


/**
* Collect all matches and return them as an Array.
*
* See selectNodes: http://msdn.microsoft.com/library/en-us/xmlsdk30/htm/xmmthselectnodes.asp
* Unlike Microsoft selectNodes, our result is not a "live" collection, but a static snapshot.
*
* @param node The context node
*/
//:CLIMETHOD Array selectNodes(Node node)
BU_XPath.prototype.selectNodes = function(node) {
  var matches = [];
  this.iterate(node, function(n) {matches.push(n); return false;});
  return matches;
}

// recursive
BU_XPath.prototype.iterate_ = function(start_node, func, cur_node, tree_depth, step_index) {
  var step = this.steps_[step_index] || bu_throw("no step number '" + step_index + "'");
  var axis = step.axis_;
  var is_descend = (axis === BU_XPath_Axis_descendant_or_self || axis === BU_XPath_Axis_descendant);
  var is_last_step = (step_index == this.steps_.length - 1);
  var node_test = step.node_test_;

  // if the 'self' axis, we only support a NodeTest of 'node()' and no predicates
  if (axis === BU_XPath_Axis_self || axis === BU_XPath_Axis_descendant_or_self) {
    if (node_test !== BU_XPath_NodeTest_node) bu_throw("unsupported: self axis without node()");
    if (step.preds_) bu_throw("unsupported: self axis with predicates");
    if (is_last_step) {
      if (func(cur_node)) return true;
      return false;
    }
    return this.iterate_(start_node, func, cur_node, tree_depth, step_index + 1);
  }

  // now we are 'descendant_or_self' or 'child' axis  
  if (!is_descend && axis !== BU_XPath_Axis_child) bu_throw("unsupported axis: " + axis);

  // we support multiple predicates in a step, which requires an array of match counters
  var match_counts = step.preds_ ? new Array(step.preds_.length): null;

  bu_debug("step=" + step.str_ + " tree_depth=" + tree_depth + " step_index=" + step_index + " is_last_step=" + is_last_step + " is_descend=" + is_descend);

  // loop through children
  for(var child=cur_node.firstChild;child;child=child.nextSibling) {
     var is_match = step.isMatch(child, match_counts);

     bu_debug("  child=" + child.nodeName + " is_match=" + is_match);

     // call function if this is the last step
     if (is_match && is_last_step) {
       if (func(child)) return true;
     }
    
     // maybe descend.
     if ((is_match && !is_last_step) || is_descend) {
       // the rules are:
       // - if !is_descend (is_match and !is_last_step):  increment step_index.
       // - if is_descend and !is_match:                  do not increment step_index
       // - if is_descend and is_match and is_last_step:  do not increment step_index
       // - if is_descend and is_match and !is_last_step: call both ways
       //   At node /A we have to descend with no increment to find /A/A/B for the path //A/B
       //   At node /A we have to descend with increment to find /A/B for the path //A/B
       //   The latter case is only if !is_last_step
       if (!is_descend) ++step_index;
       if (this.iterate_(start_node, func, child, tree_depth+1, step_index)) return true;
       if (is_descend && is_match && !is_last_step) {
         bu_debug("descending again");
         if (this.iterate_(start_node, func, child, tree_depth+1, step_index + 1)) return true;
       }
     }

     // as an optimization, if we have a predicate, for now we happen to know
     // that there is at most one match at this level.
     if (is_match && step.preds_) break;
  }
  return false;
}

//:CLEND burst.xml.XPath

bu_loaded('burst.xml.XPath');
