/**
* @file XmlDoc.js
*
* Defines burst.xml.XmlDoc which contains static utility functions for XML DOM objects.
*
* @author Copyright 2003 Mark D. Anderson (mda@discerning.com)
* @author Licensed under the Academic Free License 1.2 http://www.opensource.org/licenses/academic.php
*/

//=java package burst.xml;
/**
Scoping class to hold static functions related to objects compliant with the XML (non-HTML) DOM.

Such an object has standard DOM methods and properties (such as documentElement, etc.).
While W3 does specify <var>createDocument()</var>, string parsing and url loading are not standardized
abilities, and so there are other methods here to do those operations.

In this library, we refer to the type of such builtin objects as "DomDocument", even though
the W3 specifications use just "Document". We find that overly confusing, because despite
what may be intended, a browser Document (the type of <code>window.document</code> ) 
is neither a superset nor subset of a XML DOM object in capabilities.

Note that this currently is not instantiable. It has some static methods that return
a DomDocument, and other functions that take a DomDocument as a first argument.

In addition to the methods here for creating a DomDocument, one may also be created
via XMLHTTP, in those browsers that support it. Because of its finer control over errors, it
may be preferable over createFromUrl.
*/
//:NSBEGIN XmlDoc
burst.xml.XmlDoc = {};

/**
ProgID values to try for Internet Explorer XMLDOM.

<h3>About ProgID's for Microsoft XML components</h3>

First, note that ProgID values are case-insensitive.

Here is a summary of Microsoft MSXML releases:
<pre>
version   dll         IE                 Comment
1.0       msxml.dll   IE4
2.0,2.0a  msxml.dll   IE4.01, IE5.0      Microsoft's incompatible WD-xsl
2.5,2.5a  msxml.dll   IE5.01             XPath and SAX2
2.5 SP1   msxml.dll   IE5.01 SP1, IE5.5
2.6       msxml2.dll                     download only
3.0 SP2   msxml3.dll  IE6.0              XSLT added, still supports WD-xsl
4.0       msxml4.dll                     WD-xsl dropped
</pre>

Some references that were used to create the above table:
- http://msdn.microsoft.com/library/en-us/dnmsxml/html/msxmlcabfile.asp
- http://www.webreference.com/xml/column20/
- http://www.perfectxml.com/MSXMLVersion.asp
- http://www.perfectxml.com/msxmlRef.asp
- http://www.topxml.com/parsers/
- http://msdn.microsoft.com/library/en-us/xmlsdk/htm/sdk_installregister_4r76.asp
- http://msdn.microsoft.com/library/en-us/xmlsdk/htm/sdk_guidprogid_1b3n.asp

Here is a summary of ProgID values for XML DOM:
<pre>
"Microsoft.XMLDom"
   Used starting with msxml.dll 1.0.
   Used as a name up until 2.5.
   Not version-independent (if you use the ProgID, you only get msxml.dll).
   For interface "IXMLDOMDocument", which is DOM Level-1

"MSXML.DOMDocument"
   same ProgID value and behavior as "Microsoft.XMLDom", started being used after it.

"Microsoft.FreeThreadedXMLDOM"
   1.0 ProgID for server-side component.

"Msxml2.DOMDocument"
   A different (extended) interface "IXMLDOMDocument2", which has extensions.
   Without a version suffix, will get the most recent version prior to (excluding) 4.0.
   With a suffix of ".2.6", ".3.0", ".4.0" will force MSXML 2.6, MSXML 3.0, or MSXML 4.0

"Msxml2.FreeThreadedDOMDocument"
   With suffix ".3.0" or ".4.0".
   For server-side use.
</pre>

Here is a summary of ProgID values for XML HTTP Request:
<pre>
"Microsoft.XMLHTTP"
   1.0, not version-independent.

"Msxml2.XMLHTTP"
   With ".2.6", ".3.0", or "4.0" suffix ("4.0" mandatory for MSXML 4.0)

"Msxml2.ServerXMLHTTP"
   For use on server-side.

</pre>

As an extra complication, up until MSXML 4.0, MSXML could be installed in
"replace" mode (take over the unversioned ProgID)
or "side-by-side" mode (require versioned ProgID).

Starting with MSXML 4.0, version-independent ProgID support (and "replace" mode) was removed.
To use 4.0, the programmer must ask for a ProgID with a particular version, such as:
- "Msxml2.DOMDocument.4.0"
- "Msxml2.FreeThreadedDOMDocument.4.0"
- "Msxml2.XMLHTTP.4.0"

(As a point of trivia, MSXML release 4.0 also renamed the SDK as
"Microsoft XML Core Services" from "Microsoft XML Parser".)

For the XSLT morass, see http://www.perfectxml.com/articles/xml/XSLTInMSXML.asp
*/

//:NSVAR Array XMLDOM_PROGID = ['Msxml2.DOMDocument.4.0', 'Msxml2.DOMDocument', 'Microsoft.XMLDom']
burst.xml.XmlDoc.XMLDOM_PROGID = ['Msxml2.DOMDocument.4.0', 'Msxml2.DOMDocument', 'Microsoft.XMLDom'];



/**
Create an empty DomDocument object.

The W3 specifies createDocument: http://www.w3.org/TR/DOM-Level-2-Core/core.html

If the environment supports the W3 createDocument, that is used for creation.
Otherwise, on IE, an ActiveX DomDocument is created:
http://msdn.microsoft.com/library/en-us/xmlsdk30/htm/xmobjpmexmldomdocument.asp
*/
//:NSFUNCTION DomDocument createEmpty()
burst.xml.XmlDoc.createEmpty = function() {
  var domdoc;
  if (document && document.implementation && document.implementation.createDocument) {
    // arguments are: nsuri, qname, doctype
    // since we are going to load from a uri or string, we don't have much
    //  interest setting the root tag name, etc.
    domdoc = document.implementation.createDocument('', '', null);
    //domdoc = bu_fix_readystate(domdoc);
  }
  else if (bu_Runtime.has_msxml()) {
    domdoc = burst.Lang.createActiveXObject(burst.xml.XmlDoc, 'XMLDOM_PROGID', 'DomDocument');
  }
  else {
    bu_unsupported('burst.xml.XmlDoc.createEmpty');
  }
  return domdoc;
}

/**
A convenience which combines createEmpty and loadUrl.
It returns the DomDocument object (though not much can be done with it until the handler is called).
*/
//:NSFUNCTION DomDocument createFromUrl(String url, Function handler)
burst.xml.XmlDoc.createFromUrl = function(url, handler) {
  var domdoc = burst.xml.XmlDoc.createEmpty();
  burst.xml.XmlDoc.loadUrl(domdoc, url, handler);
  return domdoc;
}

/**
Return a Dom Document initialized from the provided string of serialized XML.

On IE, this is implemented by Document.loadXML(String):
http://msdn.microsoft.com/library/en-us/xmlsdk30/htm/xmmthloadxml.asp

On Mozilla, DOMParser is used:
http://unstable.elemental.com/mozilla/build/latest/mozilla/extensions/dox/interfacensIDOMParser.html

The caller should expect the possibility of a thrown exception, and should
also check the value of burst.xml.XmlDoc.getParseError(domdoc).

@todo For IE, as with load(), is doc.loadXML() false if and only if doc.parseError is set?
*/
//:NSFUNCTION DomDocument createFromString(String str)
burst.xml.XmlDoc.createFromString = function(str) {
  var domdoc;
  if (bu_Runtime.has_msxml()) { // IE.
    domdoc = burst.xml.XmlDoc.createEmpty();
    var is_success = domdoc.loadXML(str);
    if (!is_success) bu_throw("loadXML('" + burst.Text.ellipsis(str,100) + "') failed, parseError: " + burst.xml.XmlDoc.getParseError(domdoc));
  }
  else if (DOMParser) {
    var parser = new DOMParser();
    domdoc = parser.parseFromString(str, 'text/xml');
  }
  return domdoc;
}


/**
Initialize the provided DomDocument object from the contents at the specified url.

There is currently no W3 specification for DomDocument.load(), only for createDocument.
(Note that http://www.w3.org/TR/DOM-Level-3-LS is a draft spec for "XML Load and Save".)

For DomDocument.load(), Mozilla attempts to follow Microsoft:
 - http://www.mozilla.org/newlayout/xml/
 - http://msdn.microsoft.com/library/en-us/xmlsdk30/htm/xmmthload.asp

For both Mozilla and IE, by default load() is async.
Mozilla only supports async, and we do not expose the ability to ask for synchronous
loading.

When the load is complete, the provided handler is called with two arguments: 
<code>handler(domdoc, url)</code> .
Note that this differs from a standard DOM onload event handler, which just takes an event argument.

Mozilla has no support for progress notification. 
IE has an API (via readystatechange), but in IE6 it is not helpful (it doesn't always
call the handler very frequently, and when it does, the object is not in a queriable state).
So we do not expose the ability to register a progress handler.

The error handling behavior in these environments is not totally understood.
For now, the caller should expect the possibility of a synchronous throw exception, 
as well as the need to check
for a non-null value from <code>getParseError()</code> when the onload_handler is called.
Errors could be caused by a security error, an IO error, or a parse error. 

There is no useful return value from this function.

This url to be loaded is subject to the usual "same origin" policy with respect to the current document.

For a background on these objects, see:
- http://www.webreference.com/programming/javascript/domwrapper/2.html

@todo In IE, does a parsing error make load() return false or just IO errors? if false is returned, is parseError always set?

@todo Does mozilla have a useful return value from DomDocument.load()?

@todo What does either IE or Moz do with a "file:" url? Is it still async?

@param xmldoc The DomDocument object to use for the load.
@param url The url to load. If it is a file, it should have a "file:" scheme.
@param handler The handler to call when it is done.
*/
//:NSFUNCTION void loadUrl(DomDocument xmldoc, String url, Function handler)
burst.xml.XmlDoc.loadUrl = function(domdoc, url, handler) {
  // create a no-args closure
  var onload2 = function() {handler(domdoc,url)};

  var assert_success = false;

  // W3
  // todo: same in Moz as DomDocument.create()?
  if (document && document.implementation && document.implementation.createDocument) {
    assert_success = true;
    domdoc.addEventListener("load", onload2, false);
    // same as: domdoc.onload = onload2;
    // not yet supported:
    // domdoc.setAsync(true);
  }

  // IE
  else {
    assert_success = true;
    // for good measure; async true is the default
    domdoc.async = true;
    // default is false
    domdoc.resolveExternals = false;
    domdoc.onreadystatechange = function() {if (domdoc.readyState == 4) onload2()};
  }

  // Mozilla may throw here? if for example attempting to load from a different server
  // IE returns a boolean, true on success. 
  var is_success = domdoc.load(url);
  if (!is_success && assert_success) bu_throw("load(" + url + ") failed, parseError: " + burst.xml.XmlDoc.getParseError(domdoc));
}

/**
Retrieve any parse error of a xml document.

When creating a xml document and it is initialized from a xml string or url,
the caller needs to expect either a synchronous exception to be thrown (such as
argument or security errors), or an error to be recorded in the xml document itself
(such as parse errors).

In IE, it has a member variable <var>parseError</var>.
In Mozilla, the document structure itself is altered to
have a single child with tagName 'parseerror':
<pre>
 &lt;parsererror xmlns="http://www.w3.org/1999/xhtml">
    XML Parsing Error: mismatched tag. Expected: &lt;/root>. 
    Location: file:///C:/DOCUME~1/Nicholas/MYDOCU~1/MYARTI~1/USINGT~1/Example/Invalid.xml 
    Line Number 3, Column 37:
    &lt;sourcetext>
       &lt;root image="root.gif">My Root&lt;/boot>
       ------------------------------------^
    &lt;/sourcetext>
&lt;/parsererror>
</pre>

See:
- http://www.webreference.com/programming/javascript/domwrapper/5.html
- Heikki Toivonen, http://groups.google.com/groups?selm=3D7F84DA.2080804%40netscape.com&rnum=4
- Bob Clary, http://groups.google.com/groups?threadm=8t4cm3%24o92%40secnews.netscape.com
- http://bugzilla.mozilla.org/show_bug.cgi?id=45566
- http://bugzilla.mozilla.org/show_bug.cgi?id=129608

@todo Mozilla uses parsererror or ParserError or PARSERERROR?
@todo When will IE throw a synchronous error to load()
@todo Any escaping of the string in case it has xml in it?

@param domdoc The DomDocument instance that might have an error.
@return a String suitable for presentation of an error message, or undefined if there is no error.
*/
//:NSFUNCTION String getParseError(DomDocument domdoc)
burst.xml.XmlDoc.getParseError = function(domdoc) {
  if (typeof domdoc.parseError != 'undefined') return domdoc.parseError;
  //if (domdoc.documentElement && domdoc.documentElement.tagName.toUpperCase() == 'PARSEERROR')
  if (domdoc.documentElement && domdoc.documentElement.localName == "parsererror") {
    return burst.xml.XmlDoc.toXmlString(domdoc);
  }
  return BU_UNDEFINED;
}





/**
Serializes the DomDocument into a single String.

If available (IE), uses the 'xml' member.
Note that in IE, when it loads an XML document into a window and there is no XSL sheet,
it uses a default one and the document that you find by serializing is HTML made
from that default XSL.
In IE6 for example this default html includes various "contentEditable" elements.

So for IE we also check for domdoc.XMLDocument.xml



Otherwise attempts to use XMLSerializer (in Mozilla, see http://www.mozilla.org/xmlextras/ ).

Otherwise it uses burst.xml.DomUtil.serialize
*/
//:NSFUNCTION String toXmlString(DomDocument domdoc)
burst.xml.XmlDoc.toXmlString = function(domdoc) {
  if (!domdoc) bu_throw("(burst.xml.XmlDocjs) no domdoc argument");
  //TEMPORARY: eventually this might be true
  if (domdoc === document) bu_throw("(burst.xml.XmlDocjs) domdoc === document");

  var must_msxml = false;

  var str = null;
  if (typeof domdoc.xml != 'undefined') {
    str = domdoc.xml;
  }
  else if (typeof domdoc.XMLDocument != 'undefined' && typeof domdoc.XMLDocument.xml != 'undefined') {
    str = domdoc.XMLDocument.xml;
  }
  else if (bu_Runtime.has_msxml()) {
    // innerXml and outerXml is .NET.
    var mess = "(burst.xml.XmlDoc.js) have msxml but domdoc has no .xml member. .innerXml=" + domdoc.innerXml + 
    " documentElement.xml=" + (domdoc.documentElement ? domdoc.documentElement.xml : 'no documentElement') +
    " XMLDocument=" + domdoc.XMLDocument ;
    if (must_msxml) bu_throw(mess);
    else bu_alert(mess);
  }

  if (!str) {
    if (typeof XMLSerializer != 'undefined') {
      var serializer = new XMLSerializer;
      str = serializer.serializeToString(domdoc);
    }
    else {

      //bu_alert("(burst.xml.XmlDoc.js) no native toXmlString, so trying serialize");
      // isHtml = false, so won't try to use innerHTML
      str = burst.xml.DomUtil.serialize(domdoc);
      // happens with a document from a text/xml window, with or without an xml declaration or doctype
      if (/contentEditable/.test(str)) {
        var mess2 = '(burst.xml.XmlDoc.js) tried to serialize xml document but got contentEditable crap'; // + str;
	bu_alert(mess2);
	//bu_throw(mess2);
      }
      //bu_unsupported('burst.xml.XmlDoc.toXmlString');
    }
  }
  return str;
}




// until Mozilla 0.9.4, Mozilla did not support readyState, see:
//    http://bugzilla.mozilla.org/show_bug.cgi?id=96459 
// The fix here is from http://webfx.eae.net/dhtml/xmlextras/xmlextras.html
// We aren't using it since this is only relevant for such an old mozilla.
function bu_fix_readystate(obj) {
   if (obj.readyState == null) {
       obj.readyState = 1;
       obj.addEventListener("load", function () {
          obj.readyState = 4;
          if (typeof obj.onreadystatechange == "function")
                  obj.onreadystatechange();
       }, false);
   }
}

//:NSEND DomDocument

bu_loaded('burst.xml.XmlDoc', ['burst.xml.DomUtil', 'burst.Text', 'burst.runtime.AbstractRuntime', 'burst.Lang']);